unit FieldDescDialogUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  sesDataSetUnit,StdCtrls, Buttons, ExtCtrls, Grids, DBGrids, Db, ComCtrls,
  sesSystem,sesEditUnit, ActnList, DBActns, ImgList;

type

TStruStatus  = record
 Empty:Boolean;
 Count:Integer;
 Restructured:Boolean;
 RecNo:Integer;
 dsState:Integer; // 0 - browse; 1 - Edit; 2 - Append; 3 - Insert
 end;

  TFieldDescDialog = class(TForm)
    Panel1: TPanel;
    SaveButton: TBitBtn;
    CancelButton: TBitBtn;
    GroupBox1: TGroupBox;
    Memo: TMemo;
    GroupBox2: TGroupBox;
    DataSource1: TDataSource;
    DBGrid1: TDBGrid;
    Splitter1: TSplitter;
    ActionList1: TActionList;
    stBar: TStatusBar;
    GroupBox3: TGroupBox;
    Panel2: TPanel;
    fName: TsesEdit;
    fAlignment: TsesEdit;
    fLabel: TsesEdit;
    fRequired: TsesEdit;
    fSize: TsesEdit;
    fWidth: TsesEdit;
    fType: TsesEdit;
    sesLogger1: TsesLogger;
    ButtonIcons: TImageList;
    EditAction: TAction;
    InsertAction: TAction;
    AppendAction: TAction;
    PostAction: TAction;
    CancelEditAction: TAction;
    EditButton: TSpeedButton;
    SpeedButton1: TSpeedButton;
    SpeedButton2: TSpeedButton;
    SpeedButton3: TSpeedButton;
    CheckBox1: TCheckBox;
    SpeedButton4: TSpeedButton;
    SpeedButton5: TSpeedButton;
    UpAction: TAction;
    DownAction: TAction;
    PostButton: TBitBtn;
    SpeedButton6: TSpeedButton;
    DeleteAction: TAction;
    sesDataSet1: TsesDataSet;
    AutoAdd: TCheckBox;
    procedure FormHide(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure sesDataSet1AfterScroll(DataSet: TDataSet);
    procedure DBGrid1ColEnter(Sender: TObject);
    procedure fNameEnter(Sender: TObject);
    procedure fNameExit(Sender: TObject);
    procedure ActionList1Update(Action: TBasicAction;
      var Handled: Boolean);
    procedure CheckBox1Click(Sender: TObject);
    procedure AppendActionExecute(Sender: TObject);
    procedure PostActionExecute(Sender: TObject);
    procedure EditActionExecute(Sender: TObject);
    procedure InsertActionExecute(Sender: TObject);
    procedure CancelEditActionExecute(Sender: TObject);
    procedure DeleteActionExecute(Sender: TObject);
    procedure SaveButtonClick(Sender: TObject);
    procedure UpActionExecute(Sender: TObject);
    procedure DownActionExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure fNameCheckText(Sender: TObject; var TextIsValid: Boolean);
    procedure fNameUserErrorProcess(Sender: TObject;
      var fModalResult: TModalResult);
    procedure fTypeChange(Sender: TObject);

  private
    { Private declarations }
    function  AddNewField(aFieldName:String):Boolean;
    function  IsFieldNameUnique(aFieldName:String):Boolean;
    function  IsFieldExisted(aFieldName:String):Boolean;
    procedure AddRecord(Append:Boolean);
    procedure TypeChanged;
    procedure MemoToDataSet;
    procedure OutInfo;
    procedure OutToMemo;
//    function GetsesType(fType:String):String;
//    function GetsesAlignment(fType:String):String;
    procedure ReadValuesFromDataSet;
    procedure WriteValuesToDataSet;
//    function VerifyField:Boolean;
    procedure ReOrder;
  public
    { Public declarations }
    sesDataSet:TsesDataSet;
    EnterStatus,Status:TStruStatus;
    FieldNameList:TStringList;
  end;

var
  FieldDescDialog: TFieldDescDialog=nil;

implementation

{$R *.DFM}

procedure TFieldDescDialog.FormHide(Sender: TObject);
begin
sesDataSet1.Close;
end;


procedure TFieldDescDialog.FormShow(Sender: TObject);
begin
FieldNameList.Clear;
MemoToDataSet;
If sesDataSet1.Active then
   begin
   ReOrder;
   sesDataSet1.First;
   ReadValuesFromDataSet;
   sesLogger1.Complete;
   end;
DBGrid1.SetFocus;
end;


procedure TFieldDescDialog.OutInfo;
var
Field:TField;
S:String;
dt:TFieldType;
begin
With stBar do
  begin
  Panels[0].Text :=IntToStr(sesDataSet1.RecNo) + ':'+
                   IntToStr(sesDataSet1.RecordCount);
  Field:=DBGrid1.SelectedField;
  If Field = nil then
     begin
     Panels[2].Text:= 'Field is not selected...';
     Exit;
     end;
  With Field do
  begin
  If FieldName='fOrder' then
     begin
     Panels[2].Text:= 'The field order in the structure(not editable)...';
     end
     else
  If FieldName='fName' then
     begin
     Panels[2].Text:= 'Name of the field: up 25 alphanumeric symbols, the first must be alpha...';
     end
     else
  If FieldName='fType' then
     begin
     dt := TFieldType(Field.AsInteger);
     Case dt of
     ftString: S:='String';
     ftInteger: S:='Integer';
     ftWord: S:='Word';
     ftFloat: S:='Float';
     ftDate: S:='Date';
     ftTime: S:='Time';
     else
     S:='Unknown';
     end; // case dt of

     Panels[2].Text:= 'DataType of the field: '+S+'; Pick up from the possible values list...';
     end
     else
  If FieldName='fLabel' then
     begin
     Panels[2].Text:= 'Display label for data-aware controls.';
     end
     else
  If FieldName='fWidth' then
     begin
     Panels[2].Text:= 'Controls the field display width, given and calculated by characters...';
     end
     else
  If FieldName='fSize' then
     begin
     Panels[2].Text:= 'Data size for String type only(bytes)...';
     end
     else
  If FieldName='fRequired' then
     begin
     Panels[2].Text:= 'The field is required: 0 or 1 must be used(pick up from list)...';
     end
     else
  If FieldName='fAlignment' then
     begin
     Panels[2].Text:= 'The text alignment of the field for data-aware controls(pick up from list)...';
     end;
  end; // with Field
  If Status.Restructured then
     Panels[1].Text := 'Modified'
     else
     Panels[1].Text := '';
  end; // with stBar
end;
procedure TFieldDescDialog.sesDataSet1AfterScroll(DataSet: TDataSet);
begin

If (Status.dsState = 0) and (DataSet.Active) then
    If (not DataSet.ControlsDisabled) then ReadValuesFromDataSet;
If sesDataSet1.Active then
OutInfo;

end;

procedure TFieldDescDialog.DBGrid1ColEnter(Sender: TObject);
begin
OutInfo;
end;

procedure TFieldDescDialog.OutToMemo;
var
S,S1,S3:String;
begin
With sesdataSet1 do
begin
First;
FieldNameList.Clear;
If Eof then
   begin
   If not Status.Empty then
      begin
      If Application.MessageBox('The existing structure will be destroyed! Continue?',
         PChar(Caption), MB_ICONWARNING+MB_YESNO) = IDNO then Exit;
      Status.Empty := True;
      Status.Restructured := True;
      end;
   Status.Count := 0;
   Memo.Clear;
   DBGrid1.SetFocus;
   end
   else
   begin
   Memo.Clear;
   While not Eof do
      begin
      S:='';
      S3 := FieldByName('fName').AsString;
      If S3 = '' then
         begin
         Application.MessageBox(PChar('Empty field name...'),
         PChar(Caption), MB_ICONSTOP);
         DBGrid1.SetFocus;
         Exit;
         end;
      FieldNameList.Add(ANSIUpperCase(S3));
      S:=S+S3 + sesFieldSeparator;

      S1 := FieldByName('fType').AsString;
      S:=S+S1 + sesFieldSeparator;
      S:=S+FieldByName('fLabel').AsString + sesFieldSeparator;
      S:=S+FieldByName('fWidth').AsString + sesFieldSeparator;
      S:=S+FieldByName('fSize').AsString + sesFieldSeparator;
      S:=S+FieldByName('fRequired').AsString + sesFieldSeparator;

      S1 := FieldByName('fAlignment').AsString;
      S:=S+S1;

      Memo.Lines.Add(S);
      Next;
      end;
   end;

end; // with sesdataSet1
end;

{
function TFieldDescDialog.GetsesType(fType:String):String;
begin
Result := '0';
If fType = 'ftString' then
   Result := '1'
   else
If fType = 'ftInteger' then
   Result := '3'
   else
If fType = 'ftWord' then
   Result := '4'
   else
If fType = 'ftFloat' then
   Result := '6'
   else
If fType = 'ftDate' then
   Result := '9'
   else
If fType = 'ftTime' then
   Result := '10';

end;
}
{
function TFieldDescDialog.GetsesAlignment(fType:String):String;
begin
Result := '';
If fType = 'taLeftJustify' then
   Result := '0'
   else
If fType = 'taCenter' then
   Result := '1'
   else
If fType = 'taRightJustify' then
   Result := '2';


end;
}
{
function TFieldDescDialog.VerifyField:Boolean;
var
S,S3:String;
begin
Result := False;
If sesDataSet1.RecordCount = 0 then Exit;
With  sesDataSet1 do
begin
      S3 := FieldByName('fName').AsString;
      If S3 = '' then
         begin
         Application.MessageBox(PChar('Empty field name...'),
         PChar(Caption), MB_ICONSTOP);
         DBGrid1.SetFocus;
         Exit;
         end;

      S := FieldByName('fType').AsString;
      S := GetsesType(S);
      If S = '0' then
         begin
         Application.MessageBox(PChar('Unknown field type of the '+S3),
         PChar(Caption), MB_ICONSTOP);
         DBGrid1.SetFocus;
         Exit;
         end;

      //
      //S:=S+FieldByName('fWidth').AsString + sesFieldSeparator;

      //S:=S+FieldByName('fSize').AsString + sesFieldSeparator;
      //S:=S+FieldByName('fRequired').AsString + sesFieldSeparator;
      //
      S := FieldByName('fAlignment').AsString;
      S := GetsesAlignment(S);
      If S = '' then
         begin
         Application.MessageBox(PChar('Unknown alignment type of the '+S3),
         PChar(Caption), MB_ICONSTOP);
         DBGrid1.SetFocus;
         Exit;
         end;
end;
end;
}
procedure TFieldDescDialog.fNameEnter(Sender: TObject);
begin
TsesEdit(Sender).sesFrame.BevelOuter := bvLowered;
end;

procedure TFieldDescDialog.fNameExit(Sender: TObject);
begin
TsesEdit(Sender).sesFrame.BevelOuter := bvRaised;
end;

function  TFieldDescDialog.AddNewField(aFieldName:String):Boolean;
var
S1:String;
J:Integer;
begin
    S1 := ANSIUpperCase(aFieldName);
    J := FieldNameList.IndexOf(S1);
    If J = -1 then
            begin
            FieldNameList.Add(S1);
            Result := True;
            end
            else
            Result := False;
end;

function  TFieldDescDialog.IsFieldNameUnique(aFieldName:String):Boolean;
var
S1,S2:String;
J,N:Integer;
begin
S2 := ANSIUpperCase(sesDataSet1.FieldByName('fName').AsString);
S1 := ANSIUpperCase(aFieldName);
N:=0;
for J:=0 to FieldNameList.Count-1 do
        begin
        If S1 = FieldNameList[J] then
           N := N + 1;
        end;
Case N of
0:
  begin
  Result := True;
  Exit;
  end;
1:
  begin
  If S1 = S2 then
      begin
      Result := True;
      Exit;
      end;
  Result := False;
  Exit;
  end;
        else
        begin
        Result := False;
        Exit;
        end;
end; // Case N

end;

function  TFieldDescDialog.IsFieldExisted(aFieldName:String):Boolean;
var
S1:String;
J:Integer;
begin
S1 := ANSIUpperCase(aFieldName);
J := FieldNameList.IndexOf(S1);
Result := (J > -1);
end;

procedure TFieldDescDialog.MemoToDataSet;
var
I:Integer;
S:String;
sesField:TsesFieldDef;
begin
FieldNameList.Clear;

If Memo.Lines.Count = 0 then
   begin
   sesDataSet1.Open;
   EnterStatus.Empty := True;
   EnterStatus.Count := 0;
   EnterStatus.Restructured := False;
   EnterStatus.dsState := 0;
   Status := EnterStatus;
   Exit;
   end;

sesDataSet1.Open;
for I:=0 to Memo.Lines.Count-1 do
    begin
    S:= Memo.Lines[I];
    sesField := sesDataSet.StringToFieldDef(S);
    If sesField.FName = '' then Continue;
    // checks up for unique name
    If not AddNewField(sesField.FName) then Continue;
    With sesField, sesDataSet1 do
         begin
         Append;
         FieldByName('fOrder').AsInteger := I+1;

         FieldByName('fName').AsString := FName;


         Case FDataType Of
         ftString:FieldByName('fType').AsString := '1';
         ftInteger:FieldByName('fType').AsString := '3';
         ftWord:FieldByName('fType').AsString := '4';
         ftFloat:FieldByName('fType').AsString := '6';
         ftDate:FieldByName('fType').AsString := '9';
         ftTime:FieldByName('fType').AsString := '10';
         end; // case
         FieldByName('fLabel').AsString := FLabel;
         FieldByName('fWidth').AsInteger := FWidth;
         FieldByName('fSize').AsInteger := FSize;
         If FRequired then
                FieldByName('fRequired').AsString := '1'
                else
                FieldByName('fRequired').AsString := '0';
         Case FAlignment Of
         taLeftJustify: FieldByName('fAlignment').AsString := '0';
         taCenter: FieldByName('fAlignment').AsString := '1';
         taRightJustify: FieldByName('fAlignment').AsString := '2';
         end; // case
         Post;
         end; // with

    end;// loop
   sesDataSet1.First;
   EnterStatus.Empty := False;
   EnterStatus.Count := sesDataSet1.RecordCount;
   EnterStatus.Restructured := False;
   EnterStatus.dsState := 0;
   Status := EnterStatus;

end;

procedure TFieldDescDialog.ActionList1Update(Action: TBasicAction;
  var Handled: Boolean);
begin
SaveButton.Enabled := Status.Restructured;
If Status.Restructured then
   stBar.Panels[1].Text := 'Modified'
   else
   stBar.Panels[1].Text := '';

    EditAction.Enabled := (Status.dsState = 0)and (sesDataSet1.RecordCount > 0);
    AppendAction.Enabled := (Status.dsState = 0);
    InsertAction.Enabled := (Status.dsState = 0);
    CancelEditAction.Enabled := not  (Status.dsState = 0);
    PostAction.Enabled := not (Status.dsState = 0);
    UpAction.Enabled := (Status.dsState = 0) and (sesDataSet1.RecordCount > 0) and
                        (sesDataSet1.RecNo > 1);
    DownAction.Enabled := (Status.dsState = 0) and (sesDataSet1.RecordCount > 0) and
                        (sesDataSet1.RecNo < sesDataSet1.RecordCount);
    DeleteAction.Enabled := (Status.dsState = 0) and (sesDataSet1.RecordCount > 0);

Handled := True;
end;

procedure TFieldDescDialog.ReadValuesFromDataSet;
var
I:Integer;
Field:TField;
begin
for I:=0 to sesLogger1.EditsCount-1 do
    begin
    Field := sesDataSet1.FindField(sesLogger1.Edits[I].Name);
    If Field <> nil then
       begin
       sesLogger1.Edits[I].Value := Field.Value;
       end;
    end;
sesLogger1.Complete;
sesLogger1.Complete;
end;

procedure TFieldDescDialog.WriteValuesToDataSet;
var
I:Integer;
Field:TField;
begin
for I:=0 to sesLogger1.EditsCount-1 do
    begin
    Field := sesDataSet1.FindField(sesLogger1.Edits[I].Name);
    If Field <> nil then
       begin
       Field.Value := sesLogger1.Edits[I].Value;
       end;

    end;
sesLogger1.Complete;    
end;

procedure TFieldDescDialog.CheckBox1Click(Sender: TObject);
begin
Self.ShowHint := CheckBox1.Checked;
end;

procedure TFieldDescDialog.AppendActionExecute(Sender: TObject);
begin
AddRecord(True);
end;

procedure TFieldDescDialog.PostActionExecute(Sender: TObject);
var
Complete:Boolean;
begin
Complete := sesLogger1.Complete;

If not Complete then
   begin
   sesLogger1.IsEditCompleted.SetFocus;
   Exit;
   end
   else
   begin
   WriteValuesToDataSet;
   try
   If sesDataSet1.State <> dsBrowse then
           sesDataSet1.Post;
   Status.Empty := False;
   Status.Count := sesDataSet.RecordCount;
   Status.Restructured := True;

Case Status.dsState Of
1: Status.RecNo := sesDataSet1.RecNo;
2,3:
   begin
   ReOrder;
   DBGrid1.SetFocus;
   end;
end; // case EditMode Of

   OutToMemo;
Case Status.dsState Of
1:
begin
sesDataSet1.RecNo := Status.RecNo;
Status.dsState:=0;
DBGrid1.SetFocus;

end;
2:
begin
sesDataSet1.RecNo := Status.RecNo;
If AutoAdd.Checked then
   begin
   AddRecord(False);
   end
   else
   begin
   Status.dsState:=0;
   DBGrid1.SetFocus;
   end;
end;
3:
begin
sesDataSet1.Last;
If AutoAdd.Checked then
   begin
   AddRecord(True);
   end
   else
   begin
   Status.dsState:=0;
   DBGrid1.SetFocus;
   end;
end;
end; // case EditMode Of


      except
      end;
   end;

end;
procedure TFieldDescDialog.EditActionExecute(Sender: TObject);
begin
//1. sets the status flag of edit mode
Status.dsState := 1;

//2. stores the current cursor dataset position
Status.RecNo := sesDataset1.RecNo;

//3. tests the current dataset state and resets it into dsBrowse, if any...
If((sesDataSet1.State = dsEdit) or(sesDataSet1.State = dsInsert)) then
  sesDataSet1.Cancel;

//4. appends a new record into the dataset
sesDataSet1.Edit;

//5. sets focus on the first editors by order...
fName.SetFocus;

end;

procedure TFieldDescDialog.InsertActionExecute(Sender: TObject);
begin
AddRecord(False);
end;

procedure TFieldDescDialog.AddRecord(Append:Boolean);
begin
//1. sets the status flag of edit mode
If Append then
Status.dsState := 3
else
Status.dsState := 2;

//2. stores the current cursor dataset position
If not Append then
Status.RecNo := sesDataset1.RecNo;

//3. tests the current dataset state and resets it into dsBrowse, if any...
If((sesDataSet1.State = dsEdit) or(sesDataSet1.State = dsInsert)) then
  sesDataSet1.Cancel;

//4. appends a new record into the dataset
If Append then
sesDataSet1.Append
else
sesDataSet1.Insert;

//5. initializes the editors(Edit box)
sesLogger1.InitEdits;

//6. sets focus on the first editors by order...
fName.SetFocus;

end;

procedure TFieldDescDialog.ReOrder;
var
I:Integer;
begin
With sesDataset1 do
   begin
   I:= 1;
   First;
   While not Eof do
      begin
      Edit;
      FieldByName('fOrder').AsInteger := I;
      Post;
      I:=I+1;
      Next;
      end;
   end;
end;


procedure TFieldDescDialog.CancelEditActionExecute(Sender: TObject);
begin
sesDataSet1.Cancel;
Status.dsState := 0;
ReadValuesFromDataSet;
DBGrid1.SetFocus;
end;

procedure TFieldDescDialog.DeleteActionExecute(Sender: TObject);

begin
sesDataSet1.Delete;
Status.dsState := 0;
Status.Restructured := True;
Status.RecNo := sesDataSet1.RecNo;
sesDataSet1.DisableControls;
ReOrder;
OutToMemo;
sesDataSet1.EnableControls;

ReadValuesFromDataSet;
If Status.RecNo <= sesDataSet1.RecordCount then
   sesDataSet1.RecNo :=Status.RecNo;
DBGrid1.SetFocus;

end;

procedure TFieldDescDialog.SaveButtonClick(Sender: TObject);
begin
If Application.MessageBox('The structure of the DataSet has been changed...'+#13+#10+
'The all data of the DataSet may be lost!'+#13+#10+'Make the structure changes permanent?',
PChar(Caption),MB_ICONWARNING+MB_YESNO)= IDYES then
    begin
    ModalResult := mrOK;
    end
    else
    Exit;
end;

procedure TFieldDescDialog.UpActionExecute(Sender: TObject);
var
F:TsesFieldDef;
begin
Status.Restructured := True;
Status.RecNo := sesDataSet1.RecNo;
Status.RecNo := Status.RecNo -1;

F.FName := sesDataSet1.FieldByName('fName').AsString;
F.FDataType := TFieldType(sesDataSet1.FieldByName('fType').AsInteger);
F.FLabel := sesDataSet1.FieldByName('fLabel').AsString;
F.FWidth := sesDataSet1.FieldByName('fWidth').AsInteger;
F.FSize := sesDataSet1.FieldByName('fSize').AsInteger;
F.FRequired := (sesDataSet1.FieldByName('fRequired').AsString = '1');
F.FAlignment := TAlignment(sesDataSet1.FieldByName('fAlignment').AsInteger);

sesDataSet1.Delete;
sesDataSet1.RecNo := Status.RecNo;
sesDataSet1.Insert;

sesDataSet1.FieldByName('fName').AsString := F.FName ;
sesDataSet1.FieldByName('fType').AsInteger := Integer(F.FDataType);
sesDataSet1.FieldByName('fLabel').AsString := F.FLabel;
sesDataSet1.FieldByName('fWidth').AsInteger := F.FWidth;
sesDataSet1.FieldByName('fSize').AsInteger := F.FSize;
If F.FRequired then
   sesDataSet1.FieldByName('fRequired').AsString := '1'
   else
   sesDataSet1.FieldByName('fRequired').AsString := '0';

sesDataSet1.FieldByName('fAlignment').AsInteger := Integer(F.FAlignment);

sesDataSet1.Post;
sesDataSet1.DisableControls;
Reorder;
OutToMemo;
sesDataSet1.EnableControls;
sesDataSet1.RecNo := Status.RecNo;
ReadValuesFromDataSet;
end;

procedure TFieldDescDialog.DownActionExecute(Sender: TObject);
var
F:TsesFieldDef;
begin
Status.Restructured := True;
Status.RecNo := sesDataSet1.RecNo;
Status.RecNo := Status.RecNo+1;

F.FName := sesDataSet1.FieldByName('fName').AsString;
F.FDataType := TFieldType(sesDataSet1.FieldByName('fType').AsInteger);
F.FLabel := sesDataSet1.FieldByName('fLabel').AsString;
F.FWidth := sesDataSet1.FieldByName('fWidth').AsInteger;
F.FSize := sesDataSet1.FieldByName('fSize').AsInteger;
F.FRequired := (sesDataSet1.FieldByName('fRequired').AsString = '1');
F.FAlignment := TAlignment(sesDataSet1.FieldByName('fAlignment').AsInteger);

sesDataSet1.Delete;
If sesDataSet1.RecNo < sesDataSet1.RecordCount then
   begin
   sesDataSet1.Next;
   sesDataSet1.Insert;
   end
   else
   sesDataSet1.Append;

sesDataSet1.FieldByName('fName').AsString := F.FName ;
sesDataSet1.FieldByName('fType').AsInteger := Integer(F.FDataType);
sesDataSet1.FieldByName('fLabel').AsString := F.FLabel;
sesDataSet1.FieldByName('fWidth').AsInteger := F.FWidth;
sesDataSet1.FieldByName('fSize').AsInteger := F.FSize;
If F.FRequired then
   sesDataSet1.FieldByName('fRequired').AsString := '1'
   else
   sesDataSet1.FieldByName('fRequired').AsString := '0';

sesDataSet1.FieldByName('fAlignment').AsInteger := Integer(F.FAlignment);

sesDataSet1.Post;
Status.RecNo := sesDataSet1.RecNo;
sesDataSet1.DisableControls;
Reorder;
OutToMemo;
sesDataSet1.EnableControls;
sesDataSet1.RecNo := Status.RecNo;
ReadValuesFromDataSet;

end;

procedure TFieldDescDialog.TypeChanged;
begin
If not (IsStringCardinal(fType.Text) = 0) then Exit;
Case fType.AsCardinal Of
1:
   begin
   fSize.ReadOnly := False;
   fSize.sesData.Ranged := True;
   If Trim(fSize.Text) = '0' then
      fSize.Text := '240';
   end;
else
begin
   fSize.ReadOnly := True;
   fSize.sesData.Ranged := False;
   fSize.Text := '0';

end;
end; // case
end;
procedure TFieldDescDialog.FormCreate(Sender: TObject);
begin
FieldNameList:=TStringList.Create;
end;

procedure TFieldDescDialog.FormDestroy(Sender: TObject);
begin
FieldNameList.Free;
end;

procedure TFieldDescDialog.fNameCheckText(Sender: TObject;
  var TextIsValid: Boolean);
begin
TextIsValid := True;
If (Trim(fName.Text) = '') then
   begin
   TextIsValid := False;
   Exit;
   end;

If sesDataSet1.State = dsInsert then
begin
TextIsValid := (not IsFieldExisted(fName.Text));
end
else
begin
TextIsValid := (IsFieldNameUnique(fName.Text));
end;

end;

procedure TFieldDescDialog.fNameUserErrorProcess(Sender: TObject;
  var fModalResult: TModalResult);
begin
If Application.MessageBox(PChar('The field name must have a unique valid value ...'+#13+#10+
'Retry the Field Name edit?'),
 PChar(Caption),MB_ICONSTOP+MB_YESNO) = IDYES then

   fModalResult := mrRetry
   else
   fModalResult := mrAbort;
   
end;

procedure TFieldDescDialog.fTypeChange(Sender: TObject);
begin
TypeChanged;
end;

end.

