unit QuickSortUnit;
{       .
  :
Ni#Ti@Nj#Tj...
@ -   ,
Ni#Ti -  :
   Ni -       ;
   Ti -   .
}
interface
uses
  Windows, Messages, SysUtils, Classes, Controls,sesSystem, Forms;

const
SortFieldSeparator=';';
SortFieldTypeSeparator='\';
CompareFloatNull=0.00000001;

type

TSortDataType = (sdString,sdInteger,sdFloat,sdDate,sdTime,sdStringInteger,sdUnknown);

TSortFieldType=record
 //      
 // number by order in record
 FieldNumberToSort:Integer;
 //     
 // type of the field
 FieldType:TSortDataType;
 {
 0 - String;
 1 - Integer;
 2 - Float;
 3 - Date;
 4 - Time;
 5 - String_Integer.
 }
 end;

 PStr = ^AStr;
 AStr = record
 S:String;
 end;

 PSortRecord = ^SortRecord;
 SortRecord = record
 Values:array of String;
 end;

TQuickSortStorage = class(TList)
private
FFieldCount:Integer;
FRecordCount:Integer;
FStorage:array of array of String;
FSortFieldTypes:array of TSortFieldType;
FCaseInSensitive:Boolean;
function GetFieldCount:Integer;
procedure SetFieldCount(Value:Integer);
function GetRecordCount:Integer;
procedure SetRecordCount(Value:Integer);

function GetSortBy:String;
procedure SetSortBy(Value:String);
function CompareSortItems(Item1, Item2: Pointer): Integer;
procedure QuickSortExecute(SortList: PPointerList; L, R: Integer);
protected
SortProc:TListSortCompare;
public

//  :
//      :
{Ni\Ti, ..., Nj\Tj
 -
 - Ni -       0...m-1, m -    ;
 - "\" -   \;
 - Ti -   :
        0 - String;
        1 - Integer;
        2 - Double(Float);
        3 - Date;
        4 - String+Integer(+ ;
 - "," -   }
property SortBy:String read GetSortBy write SetSortBy;

//     -  !
property FieldCount:Integer read GetFieldCount write SetFieldCount;

//   -  !
property RecordCount:Integer read GetRecordCount write SetRecordCount;

//       
property CaseInSensitive:Boolean read FCaseInSensitive write FCaseInSensitive;

//   :       .
constructor CreateSortStorage(const aRecordNumber,aFieldNumber:Integer);

//  
destructor Destroy;override;

//  
//   
function GetField(const FieldNumber:Integer; RecordNumber:Integer):String;

//   
procedure SetField(const FieldNumber:Integer; RecordNumber:Integer; Value:String);

//  
procedure SortExecute;

//*

end;



//      :    
function ParseStrInt(Source:String; var StrText:String; var IntText:Integer):Boolean;

//    ,    
function CompareInt(const Int1,Int2:String):Integer;

//    ,    
function CompareFloat(const Float1,Float2:String):Integer;

//   ,    
function CompareDate(const Date1,Date2:String):Integer;

//   ,    
function CompareTime(const Time1,Time2:String):Integer;

//    String+Integer,    
function CompareStrInt(const StrInt1,StrInt2:String):Integer;


var
// common reference sort storage pointer
SortStorage:TQuickSortStorage = nil;

implementation

//    String+Integer,    
function CompareStrInt(const StrInt1,StrInt2:String):Integer;
var
i1,i2:Integer;
s1:String;
begin
ParseStrInt(StrInt1, s1,i1);
ParseStrInt(StrInt2, s1,i2);

Result := AnsiCompareStr(s1,s1);
If Result = 0 then
begin
If i1 > i2 then
   Result := 1
   else
   If i1 < i2 then
      Result := -1
      else
      Result := 0;
end;
end;

//   ,    
function CompareDate(const Date1,Date2:String):Integer;
var
d1,d2:TDate;
begin
d1:=StrToDate(Date1);
d2:=StrToDate(Date2);

If d1 > d2 then
   Result := 1
   else
   If d1 < d2 then
      Result := -1
      else
      Result := 0;
end;

function CompareTime(const Time1,Time2:String):Integer;
var
d1,d2:TTime;
begin
d1:=StrToTime(Time1);
d2:=StrToTime(Time2);

If d1 > d2 then
   Result := 1
   else
   If d1 < d2 then
      Result := -1
      else
      Result := 0;
end;

//    ,    
function CompareFloat(const Float1,Float2:String):Integer;
var
f1,f2,df:Double;
begin
f1:=StrToFloat(Float1);
f2:=StrToFloat(Float2);
df := f1 - f2;
If df > CompareFloatNull then
   Result := 1
   else
   If df < -CompareFloatNull then
      Result := -1
      else
      Result := 0;
end;

//    ,    
function CompareInt(const Int1,Int2:String):Integer;
var
i1,i2:Integer;
begin
i1:=StrToInt(Int1);
i2:=StrToInt(Int2);
If i1 > i2 then
   Result := 1
   else
   If i1 < i2 then
      Result := -1
      else
      Result := 0;
end;

function TQuickSortStorage.CompareSortItems(Item1, Item2: Pointer): Integer;
var
i:Integer;
P1,P2:PSortRecord;
Value1,Value2:String;
CompareResult:SmallInt;
begin
Result := 0;
//1.   -  
If  FSortFieldTypes = nil then Exit; //

P1:=Item1;
P2:=Item2;

CompareResult:=0;
//     
for i:=0 to High(FSortFieldTypes) do
    begin
    If FCaseInSensitive then
       begin
       Value1 :=ANSIUpperCase(P1^.Values[FSortFieldTypes[i].FieldNumberToSort]);
       Value2 :=ANSIUpperCase(P2^.Values[FSortFieldTypes[i].FieldNumberToSort]);
       end
       else
       begin
       Value1 :=P1^.Values[FSortFieldTypes[i].FieldNumberToSort];
       Value2 :=P2^.Values[FSortFieldTypes[i].FieldNumberToSort];
       end;

    Case FSortFieldTypes[i].FieldType Of
    sdString: // String
      begin
      CompareResult := AnsiCompareStr(Value1,Value2);
      end;
    sdInteger: // 
      begin
      If Value1='' then Value1:='0';
      If Value2='' then Value2:='0';
      CompareResult := CompareInt(Value1,Value2);
      end;
    sdFloat: // 
      begin
      If Value1='' then Value1:='0';
      If Value2='' then Value2:='0';
      CompareResult := CompareFloat(Value1,Value2);
      end;
    sdDate: // 
      begin
      If Value1='' then Value1:=DateToStr(Date-100);
      If Value2='' then Value2:=DateToStr(Date-100);
      CompareResult := CompareDate(Value1,Value2);
      end;
    sdTime: // 
      begin
      If Value1='' then Value1:='00:00:00';//TimeToStr(Date-100);
      If Value2='' then Value2:='00:00:00'; //TimeToStr(Date-100);
      CompareResult := CompareTime(Value1,Value2);
      end;

    sdStringInteger: // String + Integer
      begin
      CompareResult := CompareStrInt(Value1,Value2);
      end;

         end; // Case Of FSortFieldTypes[i].FieldType
    If CompareResult <> 0 then
       begin
       Result := CompareResult;
       Exit;
       end;
    end; //      
Result := 0;
end;
procedure TQuickSortStorage.SetField(const FieldNumber:Integer; RecordNumber:Integer; Value:String);
begin
If (FieldNumber > (FFieldCount-1)) Or (RecordNumber > (FRecordCount)) then Exit;
FStorage[RecordNumber, FieldNumber] := Value;
end;

function TQuickSortStorage.GetField(const FieldNumber:Integer; RecordNumber:Integer):String;
var
P:PSortRecord;
begin
Result := '';
If (FieldNumber > (FFieldCount-1)) Or (RecordNumber > (FRecordCount)) then Exit;
//1.   -  
If  FSortFieldTypes = nil then
    begin
    //  ,     
    Result := FStorage[RecordNumber,FieldNumber];
    Exit;
    end;

P:=List[RecordNumber];
Result := P^.Values[FieldNumber];
end;

function TQuickSortStorage.GetFieldCount:Integer;
begin
Result := FFieldCount;
end;

procedure TQuickSortStorage.SetFieldCount(Value:Integer);
begin
FFieldCount := Value;
end;

function TQuickSortStorage.GetRecordCount:Integer;
begin
Result := FRecordCount;
end;

procedure TQuickSortStorage.SetRecordCount(Value:Integer);
var
I:Integer;
begin
FRecordCount := Value;
SetLength(FStorage, FRecordCount);
for I:=0 to High(FStorage) do
    SetLength(FStorage[I], FFieldCount);
end;

constructor TQuickSortStorage.CreateSortStorage(const aRecordNumber,aFieldNumber:Integer);
var
i:Integer;
begin
inherited Create;
//SortProc := CompareSortItems;
FCaseInSensitive := True;
FStorage := nil;
FSortFieldTypes := nil;
FieldCount := aFieldNumber;
RecordCount := aRecordNumber;
for i:= 0 to RecordCount-1 do
    begin
    Add(Addr(FStorage[i]));
    end;
end;

destructor TQuickSortStorage.Destroy;
begin
FStorage := nil;
inherited Destroy;
end;

function TQuickSortStorage.GetSortBy:String;
var
i:Integer;
begin
Result := '';
If FSortFieldTypes = nil then Exit;
for i:=0 to High(FSortFieldTypes) do
    begin
    If Result = '' then
       Result:=IntToStr(FSortFieldTypes[i].FieldNumberToSort) +
               SortFieldTypeSeparator +
               IntToStr(Integer(FSortFieldTypes[i].FieldType))
       else
       Result:=SortFieldSeparator + IntToStr(FSortFieldTypes[i].FieldNumberToSort) +
               SortFieldTypeSeparator +
               IntToStr(Integer(FSortFieldTypes[i].FieldType));


    end;
end;

procedure TQuickSortStorage.SetSortBy(Value:String);
var
tl:TFastList;
i,N:Integer;
begin
If Value = '' then
   begin
   FSortFieldTypes := nil;
   Exit;
   end;
tl:=TFastList.Create;
ParseString(Value,SortFieldSeparator,'', tl);
If tl.Count = 0 then
   begin
   tl.Free;
   FSortFieldTypes := nil;
   Exit;
   end;

SetLength(FSortFieldTypes, tl.Count);
for i:=0 to tl.Count-1 do
    begin
    N:=Pos(SortFieldTypeSeparator,tl[i]);
    If N=0 then Continue;
    FSortFieldTypes[i].FieldNumberToSort := StrToInt(Copy(tl[i],1,N-1));
    FSortFieldTypes[i].FieldType := TSortDataType(StrToInt(Copy(tl[i],N+1,10)));
    end;
tl.Free;
end;

procedure TQuickSortStorage.SortExecute;
begin
//1.   -  
If  FSortFieldTypes = nil then Exit; //

  if (List <> nil) and (Count > 0) then
    QuickSortExecute(List, 0, Count - 1);
end;

procedure TQuickSortStorage.QuickSortExecute(SortList: PPointerList; L, R: Integer);
var
  I, J: Integer;
  P, T: Pointer;
begin
try
  repeat
    I := L;
    J := R;
    P := SortList^[(L + R) shr 1];
    repeat
      while CompareSortItems(SortList^[I], P) < 0 do
        Inc(I);
      while CompareSortItems(SortList^[J], P) > 0 do
        Dec(J);
      if I <= J then
      begin
        T := SortList^[I];
        SortList^[I] := SortList^[J];
        SortList^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then
      QuickSortExecute(SortList, L, J);
    L := I;
  until I >= R;
except
Application.MessageBox('Invalid compare data...','Sort',MB_ICONSTOP);
Exit;
end;
  
end;


//      :    
function ParseStrInt(Source:String; var StrText:String; var IntText:Integer):Boolean;
var
ip,ip1,ipM:Integer;
ipS1:String;
ipC:Char;
begin
Result := False;
StrText := '';
IntText := 0;
ipS1 := Trim(Source);
ipM := Length(ipS1);
If ipM = 0 then Exit;
ip1:=-1;
for ip:=ipM downto 1 do
    begin
    ipC := ipS1[ip];
    If (ipC < '0') or (ipC > '9') then
        begin
        ip1:=ip;
        Break;
        end;
    end;
Result := True;
If ip1 = ipM then
   begin
   IntText := 0;
   StrText := ipS1;
   Exit;
   end;

If ip1 = -1 then
   begin
   IntText := StrToInt(ipS1);
   StrText := '';
   Exit;
   end;
StrText := Copy(ipS1,1, ip1);
IntText := StrToInt(Copy(ipS1,ip1+1, 1000));

end;



end.
