
{*******************************************************}
{ Delphi Text DataSet is the base component for creation of the TsesDataSet}
{*******************************************************}

unit sesDataSetUnit;

{
   This file contains a  "light" TDataSet implementation which works with
 text files or memory strings.  Text file or TStrings are interpreted as a multiple column
 and  multiple row table.
If no sesFields is presented, Text file or TStrings are interpreted as a single column
 and  multiple row table with one predefined field named "Line" with data size = 240 bytes
 and DataType = ftString.
Restriction!
1.Total length of dataset record must not be greater, than MaxStrLen.
Note. You may set the variable MaxStrLen to any cardinal value, you need...
2. sesDataSet avoids null exception on Field.Required=True, if value is Null,
   so, you need to process "Empty Event" event on base of data-aware controls...
3. sesFilter of this version does not support any edit operations with
   a "sesFiltered" dataset, so, use sesFilter;sesFiltered properties for
   restrictings of the records view of the dataset only...
}
{To create or modify DataSet structure use:
- design-time: sesFields property editor;
- run-time with sesDataSet:
 the first method:
  1. Close;
  2. sesFieldsClear;
    either
    3. sesAddFieldDef(NewFieldDesc:TsesFieldDef);
    or
    3. sesInsertFieldDef(Index:Integer; NewFieldDesc:TsesFieldDef);
    Note. Both functions will return "True", if NewFieldDesc is valid.
  4. Open;(if need...)
 the second method:
  1. Close;
  2. set FileName to file name with stored structure as the first string.
  3. Open.
  Note. You may use the restored from file structure in design-time as well...
}
interface

uses
  sesSystem,DB, Classes,Controls,QuickSortUnit,sesStrings;

var
  MaxStrLen:Integer = 2048; { This is the max string length value and total record size}

  // change this variable, if need...
  // this is divider in strings, which can consists of either field definition or data values.
  sesFieldSeparator:Char = ';';
  // divider between fields' definitions in the first structural line of text in file
  sesFieldDefSeparator:Char ='@';

type

// ses field description
TsesFieldDef = record
  FName:String;          // Field Name is required!
  FDataType:TFieldType;  // Field Data Type is required!
  {sesDataSet supports only the following data types("classic"):
  ftString, ftInteger, ftWord, ftFloat, ftDate, ftTime.
  }
  {optional}
  FLabel:String; // display label as Caption or Title for data-aware control

  FWidth:Integer; // display dynamic Desired with of the field by CHARACTERS:
                  // real control width must be calculated just at draw time...

  FSize:Integer; // required for ftString type only, always null for other types,
                 //if one is not assigned, "240" value will be used...

  FRequired:Boolean; // if field must not be empty: consider test it by data-aware control rather...

  FAlignment:TAlignment; // text alignment while control will render the field value

  Valid:Boolean; // this structure is verified well and ready to use by sesDataSet,
                 // using by import functions
  end;

// import of field structure regulating options:
// they define the sesBorrowFieldDef function while it process a imported field
TImportOption = (impDateTimeAsDate, impMemoAsString,impCurrencyAsFloat, impAllIntAsInteger);
{
 impDateTimeAsDate  ftDateTime data type will be interpreted as ftDate

 impMemoAsString    ftMemo & ftFmtMemo data type will be interpreted as ftString

 impCurrencyAsFloat ftCurrency data type will be interpreted as ftFloat

 impAllIntAsInteger ftSmallInt & ftLargeInt data type will be interpreted as ftInteger
}
// Notes. Without setting the corresponding options, the field of "non-ses" types will
//         be ignored(or be formed not Valid definition).

TImportOptions = set of TImportOption;

// defines internal actions on DataSet's Close event
TCloseMode = (cmNone, cmClearAll, cmData, cmStruData);
{
 cmNone         no data or structure will be saved, even FileName is assigned,
                you should use SaveToFile method to explicitly save your data and/or structure
                to external file;
                DataStrings and sesFields(Structure) will not be cleared!
                Default!
 cmClearAll     the same as cmNone, but all data(DataStrings) and the Structure(sesFields)
                WILL BE DELETED, even they have been created in design-time;
                you should us this option with dynamically created datasets.

 cmData         all changes of data storage will be saved to file on conditions:
                 1. data has been changed; 2. FileName property is assigned.
                 DataStrings and sesFields(Structure) will be cleared!
                 In this case you may lose the structure or data you entered in the
                 design-time!

 cmStruData     the same as cmData, but the Structure will be
                save to file as the first formatted string.
}

{As TsesDataset has been designed for memory data operations, wich structure usually
is defined at design-time,
Default settings CloseMode = cmNone;}

// defines internal actions on DataSet's Open event
TOpenMode = (omNone,omData,omStructureData);
{
 omNone            no data or structure will be read, even FileName property
                   is not empty, this mode is the most suitable to operate
                   data and structure defined in the design-time.
                   Default!

 omData            forces to read the data strings from given by FileName file,
                   if it is present, otherwise, keeps the DataStrings without any changes,
                   leaves sesFields designed in design-time permanent...


 omStructureData   forces to replace the existing Structure from the first
                   string of the file, however, the old structure will be kept,
                   if this formatted string is not present or not valid.
                   Attention. The design-time structure will be lost!

}
{Note. If OpenMode = omNone, TsesDataSet will do NOTHING during Open, so,
 you should work only with memory strings and design-time defined structure...
 "Do nothing" means, that data will not be loaded from file, even the one is assigned...
 Attention! The "omNone" value of OpenMode is default!}


// Seek start up settings
TSeekOption = (soFirst,soCurrent,soIgnoreCurrent);
{
 soFirst         from the begin of dataset(DEFAULT)
 soCurrent       from the current cursor position
 soIgnoreCurrent ignores the current record and begins the search from +1 or -1 position
                 on depend of foForward value;
                 the last mode is suitable for "Search Again" operations...
}

{ TRecInfo }

{   This structure is used to access additional information stored in
  each record buffer which follows the actual record data.

    Buffer: PChar;
   ||
   \/
    --------------------------------------------
    |  Record Data  | Bookmark | Bookmark Flag |
    --------------------------------------------
                    ^-- PRecInfo = Buffer + FRecInfoOfs

  Keep in mind that this is just an example of how the record buffer
  can be used to store additional information besides the actual record
  data.  There is no requirement that TDataSet implementations do it this
  way.

  For the purposes of this demo, the bookmark format used is just an integer
  value.  For an actual implementation the bookmark would most likely be
  a native bookmark type (as with BDE), or a fabricated bookmark for
  data providers which do not natively support bookmarks (this might be
  a variant array of key values for instance).

  The BookmarkFlag is used to determine if the record buffer contains a
  valid bookmark and has special values for when the dataset is positioned
  on the "cracks" at BOF and EOF. }

  PRecInfo = ^TRecInfo;
  TRecInfo = packed record
    Bookmark: Integer;
    BookmarkFlag: TBookmarkFlag;
  end;

  {TRecRef structure}
  PRecRef = ^TRecRef;
  TRecRef = packed record
    Bookmark: Integer;
    BookmarkFlag: TBookmarkFlag;
    RecNo:Integer;
      end;

{ TsesDataSet }
  TsesDataSet = class(TDataSet)
  private
    FData: TStrings;
//!!!    FDataStorage:TStrings;
    FBufferData:TStrings;
    FSaveBufferFlag:Boolean;
    FsesFilter:TStrings;
    FsesFiltered:Boolean;
    FRecBufSize: Integer;
    FRecInfoOfs: Integer;
    FCurRec: Integer;
    FFileName: string;
    FLastBookmark: Integer;
    FSaveChanges: Boolean;
    FInternalOpen:Boolean;



    FsesFields:TStrings; // strings with each field definition:
    {Name(0);DisplayLabel(1);DataType(2);DisplayWidth(3);Size(4)
    CustNo;Customer Nom;1;56;4
    }
    FFieldsList:TWorkList; // ses the underlined type Field list:
    {
    Name(0);
    DisplayLabel(1);
    DataType(2);       1 - Str, 3 - Int; 4 - Word, 6 - Float, 9 - Date, 10 - Time
    DisplayWidth(3);   Symbols!
    Size(4);           String only!
    Required(5);     0 - False, 1 - True
    Alignment(6)     0 - left, 2 - center, 2 - right, .
    }

    FFieldValues:TWorkList; // field values list by record:
    {ValueField_0;ValueField_1;ValueField_2;...ValueField_Count-1;}

    // option, regulated the Import process: borrow field structure from another dataset...
    FImportOptions:TImportOptions;

    // the two fields below define OnClose&OnOpen default sesDataSet actions, see above...
    FCloseMode:TCloseMode;
    FOpenMode:TOpenMode;


    FsesFieldsReady:Boolean; // good fields list has been created


    // defines full replacement of the existing Structure, if present, from
    // the first string from file
    FReplaceStructureFromFile:Boolean;

    FWorkList1,FWorkList2:TWorkList;
    procedure SetData(Value:TStrings);
    procedure SetsesFields(Value:TStrings);


  protected
    { Overriden abstract methods (required) }
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    //merges FData and FBufferData into Fdata target
    procedure MergeData;
    //process sesFilter operation
    procedure FilterData;


    function GetsesFieldOrder(aFieldName:String):Integer;
    function GetsesFieldType(aFieldName:String):TSortDataType;

    function GetFieldOrder(aField:TField):Integer;
    function GetFieldText(aField:TField; aString:String):String;
    function ReplaceStringByFieldValue(aField:TField; aSource:String;
             aFieldValue:String):String;
    function GetsesFieldText(aFieldNo:Integer; aRecNo:Integer):String;

    function InitString:String;


    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    procedure SetsesFilter(Value:TStrings);
    procedure SetsesFiltered(Value:Boolean);

    { Additional overrides (optional) }
    function GetRecordCount: Integer; override;
    function GetRecNo: Integer; override;
    procedure SetRecNo(Value: Integer); override;

    // ses functions
    procedure SetDisplayProperties;
    //procedure FillBuffer(aBuffer:PChar);
  public
    {Sort controls functions}
    // sorts the FData strings by OrderBy list of fields name, separated by semicolon,
    {returns False, if:
    - OrderBy is empty or invalid;
    - sesFields is empty;
    - FData is empty;
    - during sorting a error type conversion has been encountered/
    Parameters:
    Ascending - if False, sort by descending order, otherwise - by ascending;
    CaseInSensetive - if True, sort will ignore case of all string fields;
    StringAsStringInt - if True, sort will interprete all string fields as two parts:
                        String + Integer, so, process sorting such, as:
                        [A1,A2,A3,A20,A11,A100] will produce the new set, as:
                        [A1,A2,A3,A11,A20,A100], instead usual:
                        [A1,A11,A100,A2,A20,A3]...
    }
function Sort(const OrderBy:String; const Ascending:Boolean;
    CaseInSensitive:Boolean; StringAsStringInt:Boolean ):Boolean;


    // data storage modified flag
    property Modified:Boolean read FSaveChanges write FSaveChanges;


    //Searches the dataset for a specified record and makes that record the current record.
    function Locate(const KeyFields: string; const KeyValues: Variant;
                Options: TLocateOptions): Boolean;override;

// the extended sesDataSet locate function...
//Searches the dataset for a specified record and
// returns False, if record is not found.
function Seek(const KeyFields: string; const KeyValues: Variant;
  Options: TLocateOptions; SeekOption:TSeekOption; goForward:Boolean): Boolean;

// the extended sesDataSet locate function...
//Searches the dataset for a specified record and
// returns False, if record is not found,
// note, that the Search values are given by string list, separated by sesFieldSeparator
function Search(const KeyFields: string; const KeyValues: String;
  Options: TLocateOptions; SeekOption:TSeekOption; goForward:Boolean): Boolean;

    // The structure maintenance functions

    //imports the sesFields structure from a existing and open DataSet,
    // using the ImportOptions and selecting fields by FieldFilter,
    // which must present semicolon separated list of including field names.
    // note. If FieldFilter is empty string, it means all fields will be imported.
    function Import(const FieldFilter:String; const DataSet:TDataSet):Boolean;

    procedure sesFieldsClear; // clears all fields description, no structure will exists...
    // appends the new field definition into the end if the sesFields' list
    function sesAddFieldDef(const sesField:TsesFieldDef):Boolean;
    // inserts the new field definition just before the field with number Index of the sesFields' list
    function sesInsertFieldDef(const Index:Integer; const sesField:TsesFieldDef):Boolean;
    // returns the valid ses field definition on base the existing Field
    function sesBorrowFieldDef(const Field:TField):TsesFieldDef;

    // parses the FieldDefString with "string" representation of a field definition and
    // returns formatted sesField structure...
    function StringToFieldDef(const FieldDefString:String):TsesFieldDef;

    // converts the sesFieldDef into "string" representation of a field definition and
    // returns definition string...
    function FieldDefToString(const sesFieldDef:TsesFieldDef):String;
    // creates and returns the full desctiption of the Structure in formatted string:
    {FieldDef0@FieldDef1@...FieldDef(Count-1)
    where FieldDef0 - string representation of a field definition number 0
    }
    function StructureToString:String;
    //creates sesFields structure on base of formatted StruString
    function StringToStructure(const StruString:String):Boolean;


{Data functions}
//removes all data strings from data storage,
// but does nothing, if Dataset is Active
procedure ClearData;

{
 saves FData strings to the file with full name aFileName ,
  with the first string as the strings representation of the exeisting Structure, if
  WithStructure is True;
 returns True, if successful saving has been executed...
}
function SaveToFile(const aFileName:String; const WithStructure:Boolean):Boolean;

//removes all structure definitions strings from sesFields,
// but does nothing, if Dataset is Active
procedure ClearStructure;

//saves FData into FBufferData
procedure SaveData;

    // we need to override constructor due to made FData as persistent
    constructor Create(AOwner:TComponent);override; // ses functions
    destructor Destroy;override; // to free the FData only
    function GetFieldData(Field: TField; Buffer: Pointer; NativeFormat: Boolean): Boolean;override;
  published
    // publishes standard usual property, Active := True/False works as well...
    property Active;

    // data storage as TStringList: be careful with direct access...
    property DataStrings:TStrings read FData write SetData;

    //  A Name of the external file storage, it may be empty...
    property FileName: TsesFileName read FFileName write FFileName;

    // options, which regulate the Import process: borrow field structure from another dataset...
    property ImportOptions:TImportOptions read FImportOptions write FImportOptions;

    // defines default actions, which will be executed On Close event.
    property CloseMode:TCloseMode read FCloseMode write FCloseMode;

    // defines default actions, which will be executed On Open event.
    property OpenMode:TOpenMode read FOpenMode write FOpenMode;

    // real dataset fields Structure As TStringList representation,
    // be free to use as usual Fields property after dataset open...
    property sesFields:TStrings read FsesFields write SetsesFields;

    {sesFilter}
    {defines the formatted ses Filter expression:
    the first string: Filter fields list, separated by sesFieldSeparator;
    the second string: Filter fields' values list, separated by sesFieldSeparator;
    }
    property sesFilter:TStrings read FsesFilter write SetsesFilter;
    // sets whether On or Off the sesFilter mode
    property sesFiltered:Boolean read FsesFiltered write SetsesFiltered;

    // standard TDataSet's events properties
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property BeforeRefresh;
    property AfterRefresh;

  end;


implementation

uses Windows, SysUtils, Forms;

{ TsesDataSet }
//process sesFilter operation
 procedure TsesDataSet.FilterData;
 var
 rActive:Boolean;
 I,J:Integer;
 Recs:TStringList;
 begin
 If FsesFiltered then
    begin
    If not Active then Exit;
    If FsesFilter.Count < 2 then Exit;
    If (Trim(FsesFilter[0]) = '') or (Trim(FsesFilter[1]) = '') then Exit;
    If FSaveBufferFlag then
       SaveData;
    try
    DisableControls;
    Recs:=TStringList.Create;
    If Search(FsesFilter[0],FsesFilter[1], [loCaseInsensitive, loPartialKey],
          soFirst, True) then
          begin
          Recs.Add(IntToStr(RecNo));
          While Search(FsesFilter[0],FsesFilter[1], [loCaseInsensitive, loPartialKey],
                       soIgnoreCurrent, True) do
                Recs.Add(IntToStr(RecNo));
          end;
    Close;
    FData.Clear;
    for I:=0 to Recs.Count - 1 do
        begin
        J := StrToInt(Recs[I]) - 1;
        FData.Add(FBufferData[J]);
        end;
    Open;
    finally
    EnableControls;
    Recs.Free;
    end;
    end
    else
    begin
    rActive := Active;
    If Active then Close;
    MergeData;
    If rActive then
       begin
       Open;
       If RecordCount > 0 then
       First;
       end;
    end;

 end;
    procedure TsesDataSet.SetsesFilter(Value:TStrings);
    begin
    FsesFilter.Assign(Value);
    end;

    procedure TsesDataSet.SetsesFiltered(Value:Boolean);
    begin
    If FsesFiltered <> Value then
       begin
       FsesFiltered := Value;
       FilterData;
       end;
    end;


// we need to override constructor due to made FData as persistent
constructor TsesDataSet.Create(AOwner:TComponent);
begin
inherited Create(AOwner);
FsesFieldsReady:=False;
FData := TStringList.Create;
FBufferData := TStringList.Create;
FsesFields:=TStringList.Create;
FsesFilter := TStringList.Create;
FInternalOpen:= False;
FFieldsList:=TWorkList.Create; // ses type field list
FFieldValues:=TWorkList.Create; // field values list by record
    FWorkList1:=TWorkList.Create;
    FWorkList2:=TWorkList.Create;

// import field structure options:
FImportOptions := [impDateTimeAsDate, impMemoAsString,impCurrencyAsFloat, impAllIntAsInteger];

{ by default, On Close event the sesDataset will NOT :
  - save data string to file with or no the Structure;
  - clear the DataStrings memory data storage;
  - clear sesFields list,
  by other words, will do NOTHING...
}
FCloseMode := cmNone;

{ by default, On Open event the sesDataset will do, if FileName property is assigned
  and not empty , and the file really exists:
- clear the DataStrings, all data entered at designed time will be lost;
- load data from file into DataStrings storage;
- interprete data by existings sesFields structure, if the last is not existed at
  Open time, text strings will be processed as one single column, milti rows table.
}
FOpenMode := omNone;

Modified := False;
FSaveBufferFlag := True;
end;

destructor TsesDataSet.Destroy; // to free the FVariables only
begin
If FInternalOpen then
   begin
   FCloseMode := cmNone;
   InternalClose;
   end;

    FData.Free;
    FBufferData.Free;
    FsesFields.Free;
    FsesFilter.Free;
    FFieldsList.Free;
    FFieldValues.Free;
    FWorkList1.Free;
    FWorkList2.Free;

inherited Destroy;
end;

procedure TsesDataSet.SetsesFields(Value:TStrings);
begin
    FsesFields.Assign(Value);
end;

//removes all data strings from data storage,
// but does nothing, if Dataset is Active
procedure TsesDataSet.ClearData;
begin
If Active then Exit;
FData.Clear;
FBufferData.Clear;
  { Reset these internal flags }
  FLastBookmark := 0;
  FCurRec := -1;
end;

//removes all structure definitions strings from sesFields,
// but does nothing, if Dataset is Active
procedure TsesDataSet.ClearStructure;
begin
If Active then Exit;
FsesFields.Clear;
end;

procedure TsesDataSet.SetData(Value:TStrings);
begin
    FData.Assign(Value);
end;

{ This method is called by TDataSet.Open and also when FieldDefs need to
  be updated (usually by the DataSet designer).  Everything which is
  allocated or initialized in this method should also be freed or
  uninitialized in the InternalClose method. }

procedure TsesDataSet.InternalOpen;
var
  I: Integer;
  fDef:String;
  Handle: THandle;
  FindData: TWin32FindData;

begin


If FOpenMode <> omNone then
begin
  If Trim(FFileName) <> '' then
     begin
     Handle := FindFirstFile(PChar(FFileName), FindData);
     If Handle <> INVALID_HANDLE_VALUE then
        begin
        Windows.FindClose(Handle);
        FData.LoadFromFile(FFileName);
        If FOpenMode = omStructureData then
           begin
           FsesFields.Clear;
           If FData.Count > 0 then
              begin
              fDef := FData[0];
              // this is "safety" label to verify validity of the first formatted string...
              If Copy(fDef,1,3) = (sesFieldDefSeparator+sesFieldDefSeparator+sesFieldDefSeparator) then
                 begin
                 fDef := Copy(fDef,4,2048); //all remainder may be the structure descriptor
                 ParseString(fDef,sesFieldDefSeparator,'', FWorkList1);
                 If FWorkList1.Count > 0 then
                    begin

                    for I:=0 to FWorkList1.Count-1 do
                        FsesFields.Add(FWorkList1[I]);
                    FData.Delete(0);
                    end;
                 end;
              end;
           end;
        end;
     end;
end; // FOpenOptions

  { Fabricate integral bookmark values }
  If FData.Count > 0 then
  begin
  for I := 1 to FData.Count do
    FData.Objects[I-1] := Pointer(I);
  FLastBookmark := FData.Count;
  end
  else
  begin
  { Reset these internal flags }
  FLastBookmark := 0;
  FCurRec := -1;
  end;

  { Initialize our internal position.
    We use -1 to indicate the "crack" before the first record. }
  FCurRec := -1;

  { Initialize an offset value to find the TRecInfo in each buffer }
  FRecInfoOfs := MaxStrLen;

  { Calculate the size of the record buffers.
    Note: This is NOT the same as the RecordSize property which
    only gets the size of the data in the record buffer }
  FRecBufSize := FRecInfoOfs + SizeOf(TRecInfo);

  { Tell TDataSet how big our Bookmarks are (REQUIRED) }
  BookmarkSize := SizeOf(Integer);

  { Initialize the FieldDefs }
  FsesFieldsReady := False;
  InternalInitFieldDefs;
  // we will ignore always persistent fields
//  CreateFields;

  { Create TField components when no persistent fields have been created }
  if DefaultFields then
     begin
     CreateFields;
     If FsesFieldsReady then
        SetDisplayProperties;
     end;

  { Bind the TField components to the physical fields }
  BindFields(True);
    FInternalOpen := True;
    Modified := False;

end;

//saves FData into FBufferData
procedure TsesDataSet.SaveData;
begin
//this version simply rewrites Buffer Data into FData
FBufferData.BeginUpdate;
FBufferData.Text := FData.Text;
FBufferData.EndUpdate;
FSaveBufferFlag := False;
end;

//merges FData and FBufferData into Fdata target
procedure TsesDataSet.MergeData;
begin
//this version simply rewrites Buffer Data into FData
FData.BeginUpdate;
FData.Text := FBufferData.Text;
FData.EndUpdate;
end;
procedure TsesDataSet.SetDisplayProperties;
var
I:Integer;
F:TField;
sesField:TsesFieldDef;
S:String;
begin
If FsesFields.Count = 0 then
   begin
   F:= Self.FindField('Line');
   If F <> nil then
      begin
      F.DisplayWidth := 50;
      F.DisplayLabel := 'Line of text';
      F.Alignment := taLeftJustify;
      end;
   Exit;
   end;
For I:= 0  to FsesFields.Count - 1 do
   begin
   S := FsesFields[I];
   sesField := StringToFieldDef(S);
   If sesField.FName = '' then Continue;
   F:= Self.FindField(sesField.FName);
   If F <> nil then
      begin
      F.DisplayWidth :=sesField.FWidth;
      F.DisplayLabel := sesField.FLabel;
      F.Alignment := sesField.FAlignment;
      end;
   end;
end;

procedure TsesDataSet.InternalClose;
var
  aFileName:String;
begin
  aFileName := Trim(FFileName);

  If (Trim(FFileName) <> '') and
     (FSaveChanges)          and
        ((FCloseMode = cmData) or (FCloseMode = cmStruData))
                             then
                                 begin
                                 SaveToFile(FFileName,(FCloseMode = cmStruData));
                                 FData.Clear;
                                 FsesFields.Clear;
                                 end;
        If FCloseMode = cmClearAll then
           begin
           ClearData;
           ClearStructure;
           end;

  FInternalOpen := False;

  { Destroy the TField components if no persistent fields }
  if DefaultFields then DestroyFields;

  { Reset these internal flags }
  FLastBookmark := 0;
  FCurRec := -1;
end;

{ This property is used while opening the dataset.
  It indicates if data is available even though the
  current state is still dsInActive. }

function TsesDataSet.IsCursorOpen: Boolean;
begin
//  Result := FData.Count > 0;//Assigned(FData);
Result := FInternalOpen;
end;

function TsesDataSet.StringToFieldDef(const FieldDefString:String):TsesFieldDef;
var
J,N,A:Integer;
begin
    {
    Name(0);
    DataType(1);       1 - Str, 3 - Int; 4 - Word, 6 - Float, 9 - Date, 10 - Time
    the above fields are always required!
    --------------------------------
    DisplayLabel(2);
    DisplayWidth(3);   Symbols!
    Size(4);           String only!
    Required(5);     0 - False, 1 - True
    Alignment(6)     0 - left, 1 - center, 2 - right.
    Totals 7 fields
    }
Result.FName := '';
With Result do
begin
       ParseString(FieldDefString,sesFieldSeparator,'',FFieldsList);
       If FFieldsList.Count < 2 then Exit;
       FName := FFieldsList[0];
       J := StrToInt(FFieldsList[1]);
       FDataType := TFieldType(J);
       N := FFieldsList.Count;
       Case N of
       2,3:
          begin
          If N = 2 then
             FLabel:=''
             else
             FLabel := FFieldsList[2];
          Case  FDataType Of
          ftString:
             begin
             FSize := 240;
             FWidth := 25;
             FAlignment := taLeftJustify;
             end;
          ftDate,ftTime:
             begin
             FSize := 0;
             FWidth := 12;
             FAlignment := taLeftJustify;
             end;
          ftInteger,ftWord:
             begin
             FSize := 0;
             FWidth := 6;
             FAlignment := taCenter;
             end;
          ftFloat:
             begin
             FSize := 0;
             FWidth := 15;
             FAlignment := taRightJustify;
             end;

          end; // Case of DataType
       FRequired := False;
       end; // 2
       4:
          begin
          FLabel := FFieldsList[2];
          FWidth := StrToInt(FFieldsList[3]);
          Case  FDataType Of
          ftString:
             begin
             FSize := 240;
             FAlignment := taLeftJustify;
             end;
          ftDate,ftTime:
             begin
             FSize := 0;
             FAlignment := taLeftJustify;
             end;
          ftInteger,ftWord:
             begin
             FSize := 0;
             FAlignment := taCenter;
             end;
          ftFloat:
             begin
             FSize := 0;
             FAlignment := taRightJustify;
             end;

          end; // Case of DataType
       FRequired := False;
       end; // 4
       5:
          begin
          FLabel := FFieldsList[2];
          FWidth := StrToInt(FFieldsList[3]);
          FSize := StrToInt(FFieldsList[4]);
          Case  FDataType Of
          ftString:
             begin
             FAlignment := taLeftJustify;
             end;
          ftDate,ftTime:
             begin
             FSize := 0;
             FAlignment := taLeftJustify;
             end;
          ftInteger,ftWord:
             begin
             FSize := 0;
             FAlignment := taCenter;
             end;
          ftFloat:
             begin
             FSize := 0;
             FAlignment := taRightJustify;
             end;

          end; // Case of DataType
       FRequired := False;
       end; // 5
       6:
          begin
          FLabel := FFieldsList[2];
          FWidth := StrToInt(FFieldsList[3]);
          FSize := StrToInt(FFieldsList[4]);
          FRequired := Trim(FFieldsList[5]) = '1';
          Case  FDataType Of
          ftString:
             begin
             FAlignment := taLeftJustify;
             end;
          ftDate,ftTime:
             begin
             FSize := 0;
             FAlignment := taLeftJustify;
             end;
          ftInteger,ftWord:
             begin
             FSize := 0;
             FAlignment := taCenter;
             end;
          ftFloat:
             begin
             FSize := 0;
             FAlignment := taRightJustify;
             end;

          end; // Case of DataType
       end; // 6
       7:
          begin
          FLabel := FFieldsList[2];
          FWidth := StrToInt(FFieldsList[3]);
          FSize := StrToInt(FFieldsList[4]);
          FRequired := Trim(FFieldsList[5]) = '1';
          A := StrToInt(FFieldsList[6]);
          Case A Of
          0: FAlignment := taLeftJustify;
          1: FAlignment := taCenter;
          2: FAlignment := taRightJustify;
          end; // Case A of
          Case  FDataType Of
          ftString:
             begin
             end;
          ftDate,ftTime:
             begin
             FSize := 0;
             end;
          ftInteger,ftWord:
             begin
             FSize := 0;
             end;
          ftFloat:
             begin
             FSize := 0;
             end;

          end; // Case of DataType
       end; // 7

       end; // case Of N

end; // with Result
end;


{ For this simple example we just create one FieldDef, but a more complete
  TDataSet implementation would create multiple FieldDefs based on the
  actual data. }

procedure TsesDataSet.InternalInitFieldDefs;
var
I:Integer;
DefaultInit,RetryInit:Boolean;
sesField:TsesFieldDef;
begin
  DefaultInit := False;
  FsesFieldsReady := False;
  If not Assigned(FsesFields) then
     DefaultInit := True
     else
     If FsesFields.Count = 0 then
        DefaultInit := True
        else
        If Trim(FsesFields[0]) = '' then
           DefaultInit := True;

RetryInit := False;
If DefaultInit then
   begin
   FieldDefs.Clear;
          with FieldDefs.AddFieldDef do begin
               Name := 'Line';
               DataType := ftString;
               Required := False;
               Size := 240;
               end;
   Exit;
   end
   else
   begin
   with FieldDefs,sesField do begin
   Clear;

   for I:=0 to FsesFields.Count-1 do
       begin
       sesField := StringToFieldDef(FsesFields[I]);
       If FName = '' then
           begin
           RetryInit := True;
          Application.MessageBox(PChar('Invalid field definition: '+#13+#10+
          FsesFields[I] + #13+#10+
          'The valid format is:'+#13+#10+
          'Name;DataType[;DisplayLabel;DisplayWidth(characters);DataSize(String only);'+
          'Required(0/1);Alignment(0/1/2)]'+#13+#10+
          #13+#10+
          'Result:'+#13+#10+
          'Only one text field "Line" will be used...'+#13+#10+
          'Advise: consider to change the structure...'),
          PChar('sesDataSet: '+Self.Name),MB_ICONWARNING);
          Break;
          end;

      If FDataType = ftString then
      begin

          with AddFieldDef do begin
               Name := FName;
               DataType := ftString;
               Required := FRequired;
               Size := FSize;
               end;
      end
      else
      begin
      with AddFieldDef do begin
        Name := FName;
        DataType := FDataType;
        Required := FRequired;
      end;
      end;
//       FieldDefs.Add(FName, FDataType, FSize, False);

    end; // with FieldDef

       end; // loop

   end;
If RetryInit then
   begin
   FieldDefs.Clear;
   FieldDefs.Add('Line', ftString, 255, False);
   end
   else
   FsesFieldsReady := True;
end;

{ This is the exception handler which is called if an exception is raised
  while the component is being stream in or streamed out.  In most cases this
  should be implemented useing the application exception handler as follows. }

procedure TsesDataSet.InternalHandleException;
begin
  Application.HandleException(Self);
end;

{ Bookmarks }
{ ========= }

{ In this sample the bookmarks are stored in the Object property of the
  TStringList holding the data.  Positioning to a bookmark just requires
  finding the offset of the bookmark in the TStrings.Objects and using that
  value as the new current record pointer. }

procedure TsesDataSet.InternalGotoBookmark(Bookmark: Pointer);
var
  Index: Integer;
begin
  Index := FData.IndexOfObject(TObject(PInteger(Bookmark)^));
  if Index <> -1 then
    FCurRec := Index else
    DatabaseError('Bookmark not found');
end;

{ This function does the same thing as InternalGotoBookmark, but it takes
  a record buffer as a parameter instead }

procedure TsesDataSet.InternalSetToRecord(Buffer: PChar);
begin
  InternalGotoBookmark(@PRecInfo(Buffer + FRecInfoOfs).Bookmark);
end;

{ Bookmark flags are used to indicate if a particular record is the first
  or last record in the dataset.  This is necessary for "crack" handling.
  If the bookmark flag is bfBOF or bfEOF then the bookmark is not actually
  used; InternalFirst, or InternalLast are called instead by TDataSet. }

function TsesDataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag;
end;

procedure TsesDataSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag := Value;
end;

{ These methods provide a way to read and write bookmark data into the
  record buffer without actually repositioning the current record }

procedure TsesDataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PInteger(Data)^ := PRecInfo(Buffer + FRecInfoOfs).Bookmark;
end;

procedure TsesDataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PRecInfo(Buffer + FRecInfoOfs).Bookmark := PInteger(Data)^;
end;

{ Record / Field Access }
{ ===================== }

{ This method returns the size of just the data in the record buffer.
  Do not confuse this with RecBufSize which also includes any additonal
  structures stored in the record buffer (such as TRecInfo). }

function TsesDataSet.GetRecordSize: Word;
begin
  Result := MaxStrLen;
end;

{ TDataSet calls this method to allocate the record buffer.  Here we use
  FRecBufSize which is equal to the size of the data plus the size of the
  TRecInfo structure. }

function TsesDataSet.AllocRecordBuffer: PChar;
begin
  GetMem(Result, FRecBufSize);
end;

{ Again, TDataSet calls this method to free the record buffer.
  Note: Make sure the value of FRecBufSize does not change before all
  allocated buffers are freed. }

procedure TsesDataSet.FreeRecordBuffer(var Buffer: PChar);
begin
  FreeMem(Buffer, FRecBufSize);
end;

{
procedure TsesDataSet.FillBuffer(aBuffer:PChar);
var
I:Integer;
F:TField;
S,S1,S2,S3:String;
FCount:Integer;
Offset:Integer;
B:array[0..MaxStrLen] of Char;
begin
If FsesFieldsReady then
begin
If FsesFields.Count = 1 then
   StrLCopy(aBuffer, PChar(FData[FCurRec]), MaxStrLen)
   else
   begin
   S:= FData[FCurRec];
   If S = '' then
      begin
      StrLCopy(aBuffer, PChar(FData[FCurRec]), MaxStrLen);
      Exit;
      end;

   ParseString(S,sesFieldSeparator,'',FWorkList1);
   If FsesFields.Count <> FWorkList1.Count then
      begin
      StrLCopy(aBuffer, PChar(''), MaxStrLen);
      raise Exception.Create('Invalid record structure of '+Self.Name);
      Exit;
      end;
   Offset := 0;
   FillChar(B, RecordSize, 32);
   for I:=0 to FWorkList1.Count-1 do
      begin
      S1:=FFieldsList[0]; // field name
      F:= Self.FindField(S1);
      If F = nil then Continue;
      S2 := FWorkList1[I]; // Field Value text
      //StrLCopy(PBuf(aBuffer + Offset)^.Value,PChar(S2), F.Size);
      Move(S2[1],B[Offset],Length(S2));
      Offset := Offset + F.Size;
      end;
   B[Offset]:=#0;
   StrLCopy(aBuffer,B,RecordSize);
   //Move(B[0],aBuffer,RecordSize);
   end;
end
else
StrLCopy(aBuffer, PChar(FData[FCurRec]), MaxStrLen);
end;
}

{ This multi-purpose function does 3 jobs.  It retrieves data for either
  the current, the prior, or the next record.  It must return the status
  (TGetResult), and raise an exception if DoCheck is True. }

function TsesDataSet.GetRecord(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
begin
  if FData.Count < 1 then
    Result := grEOF else
  begin
    Result := grOK;
    case GetMode of
      gmNext:
        if FCurRec >= RecordCount - 1  then
          Result := grEOF else
          Inc(FCurRec);
      gmPrior:
        if FCurRec <= 0 then
          Result := grBOF else
          Dec(FCurRec);
      gmCurrent:
        if (FCurRec < 0) or (FCurRec >= RecordCount) then
          Result := grError;
    end;
    if Result = grOK then
    begin
         // we have to move each string of FData to Buffer
         StrLCopy(Buffer, PChar(FData[FCurRec]), MaxStrLen);
         //FillBuffer(Buffer);
         with PRecInfo(Buffer + FRecInfoOfs)^ do
           begin
           BookmarkFlag := bfCurrent;
           Bookmark := Integer(FData.Objects[FCurRec]);
           end;

    end else
      if (Result = grError) and DoCheck then DatabaseError('No Records');
  end;
end;

{ This routine is called to initialize a record buffer.  In this sample,
  we fill the buffer with zero values, but we might have code to initialize
  default values or do other things as well. }

procedure TsesDataSet.InternalInitRecord(Buffer: PChar);
begin
  FillChar(Buffer^, RecordSize, 0);
end;

function TsesDataSet.GetFieldOrder(aField:TField):Integer;
var
J,N:Integer;
F,Fname:String;
begin
Result := -1;
Fname := Trim(AnsiUpperCase(aField.FieldName));
for J:= 0 to FsesFields.Count-1 do
    begin
    N:=Pos(sesFieldSeparator, FsesFields[J]);
    If N = 0 then Exit;
    F:= AnsiUpperCase(Copy(FsesFields[J],1,N-1));
    F:=Trim(F);
    If F = FName then
       begin
       Result := J;
       Exit;
       end;
    end;
end;

function TsesDataSet.GetsesFieldOrder(aFieldName:String):Integer;
var
J,N:Integer;
F,Fname:String;
begin
Result := -1;
Fname := Trim(AnsiUpperCase(aFieldName));
for J:= 0 to FsesFields.Count-1 do
    begin
    N:=Pos(sesFieldSeparator, FsesFields[J]);
    If N = 0 then Exit;
    F:= AnsiUpperCase(Copy(FsesFields[J],1,N-1));
    F:=Trim(F);
    If F = FName then
       begin
       Result := J;
       Exit;
       end;
    end;
end;

function TsesDataSet.GetsesFieldType(aFieldName:String):TSortDataType;
var
J,N:Integer;
S,F,Fname:String;
begin
Result := sdUnknown;
Fname := Trim(AnsiUpperCase(aFieldName));
for J:= 0 to FsesFields.Count-1 do
    begin
    N:=Pos(sesFieldSeparator, FsesFields[J]);
    If N = 0 then Exit;
    F:= AnsiUpperCase(Copy(FsesFields[J],1,N-1));
    F:=Trim(F);
    If F = FName then
       begin
       S:=Copy(FsesFields[J],N+1,10);
       N:=Pos(sesFieldSeparator, S);
       If N > 0 then
          S:=Copy(S,1,N-1);
       S:=Trim(S);
       If not IsStringCardinal(S) = 0 then Exit;
       N:=StrToInt(S);
       Case N of
       1: Result := sdString;
       3,4: Result := sdInteger;
       6: Result := sdFloat;
       9: Result := sdDate;
       10: Result := sdTime;
       end; // case N

       Exit;
       end;
    end;
end;

function TsesDataSet.GetFieldText(aField:TField; aString:String):String;
var
FOrder:Integer;
begin
Result := '';
FOrder := GetFieldOrder(aField);
If (FOrder < 0) then Exit;
ParseString(aString,sesFieldSeparator,'', FWorkList2);
If FOrder > (FWorkList2.Count-1) then Exit;
Result := FWorkList2[FOrder];
end;

function TsesDataSet.GetsesFieldText(aFieldNo:Integer; aRecNo:Integer):String;
var
FOrder:Integer;
begin
Result := '';
FOrder := aFieldNo;
If (FOrder < 0) or (FOrder > FsesFields.Count-1) then Exit;
If (aRecNo < 0) or (aRecNo > FData.Count-1) then Exit;
ParseString(FData[aRecNo],sesFieldSeparator,'', FWorkList2);
If FOrder > (FWorkList2.Count-1) then Exit;
Result := FWorkList2[FOrder];
end;

function TsesDataSet.ReplaceStringByFieldValue(aField:TField; aSource:String;
aFieldValue:String):String;
var
J,FOrder:Integer;
S:String;
begin
Result := '';

FOrder := GetFieldOrder(aField);
If (FOrder < 0) then Exit;

// process the case of empty(insert or append) string
If Trim(aSource) = '' then
   begin
   aSource := '';
   for J:=0 to FsesFields.Count-1 do
       begin
       If J = 0 then
          aSource := ' '
          else
          aSource := aSource + sesFieldSeparator+' ';
       end;
   end;

ParseString(aSource,sesFieldSeparator,'', FWorkList2);
If FOrder > (FWorkList2.Count-1) then Exit;
Result := '';
for J:= 0 to FWorkList2.Count-1 do
  begin
  If J = FOrder then
     S:= aFieldValue
     else
     S:= FWorkList2[J];
  If J = 0 then
     Result := S
     else
     Result := Result + sesFieldSeparator + S;
  end;
end;

{ Here we copy the data from the record buffer into a field's buffer.
  This function, and SetFieldData, are more complex when supporting
  calculated fields, filters, and other more advanced features.
  See TBDEDataSet for a more complete example. }

function TsesDataSet.GetFieldData(Field: TField; Buffer: Pointer; NativeFormat: Boolean): Boolean;
var
S,FValue:String;
P:PChar;
IntValue:LongInt;
FloatValue:Double;
WordValue:Word;
DateValue:TDate;
TimeValue:TTime;
pDate:TParsedDate;
pTime:TParsedTime;
begin
// where from do calls GetFieldData issue?
// I think, nobody knows, even Borland...
// By theory, calls are generatated by your code or data-aware controls, may be...
{So, consider the following advise:
don't touch this text without serious reason!}
If (not FInternalOpen)  then
   begin
   P:=#0;
   StrLCopy(Buffer, P, 1);
   Result := False;
   Exit;
   end;
   If FsesFieldsReady then
      begin
     // multiple column's description is presented as well...

{Notes. The task, indeed, is simple...
1. Data of the current record is stored in ActiveBuffer(unformatted string).
   ActiveBuffer really point to the current string of FData, but it is not always...
2. Buffer may(MUST!) point to Field's buffer, but it may be nil sometimes...
3. We have to a) extract field value from ActiveBuffer; b) move it to field's buffer.
   That's all!
}

      // this is encountered with DBGrid ... ?
      // What way can Buffer be nil?!
      // work around:
      If Buffer = nil then
          begin
          //Really, this is encountered only with Required field...
          If (State in [dsEdit,dsInsert]) and (Field.Required) then
             Result :=True
             else
             Result := False;
          Exit;
          end;

      P:=ActiveBuffer; // full text of the current string is in ActiveBuffer
      // we have to send to Buffer the part of the entire string, which
      // is corresponding to the value of Field
      If (FData.Count = 0) and not ((Self.State = dsEdit ) or (Self.State = dsInsert )) then
         S:=' '
         else
         S:=String(P); // source string

      FValue := GetFieldText(Field, S);
      If FValue = '' then
                       begin
                       Buffer := nil;
                       Result := False;
                       Exit;
                       end;

      Case Field.DataType Of
      ftString:
                begin
                P:=PChar(FValue);
                StrLCopy(Buffer, P, Length(FValue));
                Result := PChar(Buffer)^ <> #0;
                end;
      ftInteger:
                begin
                If not (IsStringInteger(FValue) = 0) then
                       begin
                       Buffer := nil;
                       Result := False;
                       Exit;
                       end
                       else
                        begin
                        IntValue := StrToInt(FValue);
                        Move(IntValue,Buffer^, SizeOf(IntValue));
                        Result := True;
                        end;
                end;
      ftWord:
                begin
                If not (IsStringCardinal(FValue) = 0) then
                       begin
                       Buffer := nil;
                       Result := False;
                       Exit;
                       end
                       else
                       begin
                       WordValue := StrToInt(FValue);
                       Move(WordValue,Buffer^, SizeOf(WordValue));
                       Result := True;
                       end;
                end;
      ftFloat:
                begin
                If not (IsStringFloat(FValue) = 0) then
                       begin
                       Buffer := nil;
                       Result := False;
                       Exit;
                       end
                       else
                       begin
                       FloatValue := StrToFloat(FValue);
                       Move(FloatValue,Buffer^, SizeOf(FloatValue));
                       Result := True;
                      end;
                end;
      ftDate:
                begin
                If not (IsStringDate(FValue,pDate) = 0) then
                       begin
                       Buffer := nil;
                       Result := False;
                       Exit;
                       end
                       else
                       begin
                       DateValue := StrToDate(FValue);
                       Move(DateValue,Buffer^, SizeOf(DateValue));
                       Result := True;
                       end;
                end;
      ftTime:
                begin
                If not (IsStringTime(FValue,pTime) = 0) then
                       begin
                       Buffer := nil;
                       Result := False;
                       Exit;
                       end
                       else
                       begin
                       TimeValue := StrToTime(FValue);
                       Move(TimeValue,Buffer^, SizeOf(TimeValue));
                       Result := True;
                       end;
                end;

      else
      begin
                P:=PChar(FValue);
                StrLCopy(Buffer, P, Length(FValue));
                Result := PChar(Buffer)^ <> #0;

      end;
      end;// Case DataType

      end
      else
      begin
      // the single default column is presented
      StrLCopy(Buffer, ActiveBuffer, Field.Size);
      Result := PChar(Buffer)^ <> #0;
      end;
//
end;

function TsesDataSet.InitString:String;
begin
  If not FsesFieldsReady then Result := '';


end;
procedure TsesDataSet.SetFieldData(Field: TField; Buffer: Pointer);
var
SourceString,S,S1:String;
P,P1:PChar;
IntValue:LongInt;
FloatValue:Double;
WordValue:Word;
DateValue:TDateTime;
h,m,se:Integer;
begin
If (not FInternalOpen)then
   begin
   Exit;
   end;
  //StrLCopy(ActiveBuffer, Buffer, Field.Size);
  P:=ActiveBuffer; // source string for replacement
  S:=String(P);

  P1:=Buffer;      // source field Text to replace into source string

  If FsesFieldsReady then
      begin
      // multiple column's description is presented as well...
      // we have to send to ActiveBuffer the entire string, part of which will
      // be replaced by the value of Field

      Case Field.DataType Of
      ftString:
                begin
                If Buffer = nil then
                   S1:=''
                   else
                   S1:=String(P1);
                end;
      ftInteger:
                begin
                If Buffer = nil then
                   S1:=''
                   else
                   begin
                   Move(Buffer^,IntValue,SizeOf(IntValue));
                   S1:=IntToStr(IntValue);
                   end;
                end;
      ftWord:
                begin
                If Buffer = nil then
                   S1:=''
                   else
                   begin
                   Move(Buffer^,WordValue,SizeOf(WordValue));
                   S1:=IntToStr(WordValue);
                   end;
                end;
      ftFloat:
                begin
                If Buffer = nil then
                   S1:=''
                   else
                   begin
                   Move(Buffer^,FloatValue,SizeOf(FloatValue));
                   S1:=FloatToStr(FloatValue);
                   end;
                end;
      ftDate:
                begin
                If Buffer = nil then
                   IntValue := 0
                   else
                   // ooooo! What does Borland do with date!
                   Move(Buffer^,IntValue,SizeOf(IntValue));
                DateValue := IntValue;
                DateValue := DateValue - 693594.0; // ??? Why? But it works!
                S1:=DateToStr(DateValue);
                end;
      ftTime:
                begin
                If Buffer = nil then
                   IntValue := 0
                   else
                   Move(Buffer^,IntValue,SizeOf(IntValue));
                se:= IntValue div 1000; // secs
                h:= se div 3600;
                m := (se - h*3600) div 60;
                se := (se - h*3600 - m*60);
                S1:=IntToStr(h)+TimeSeparator+
                    IntToStr(m)+TimeSeparator+
                    IntToStr(se);
                end;

      else
      begin
      S1:='';
      end;
      end;// Case DataType

      SourceString := ReplaceStringByFieldValue(Field,S,S1);
      end
      else
      begin
      // the single default column is presented
      S1:=String(P1);
      SourceString := S1;
      end;
  P1:=PChar(SourceString);
  StrLCopy(ActiveBuffer, P1, Length(SourceString)); //Field.Size);
  DataEvent(deFieldChange, Longint(Field));
end;

{ Record Navigation / Editing }
{ =========================== }

{ This method is called by TDataSet.First.  Crack behavior is required.
  That is we must position to a special place *before* the first record.
  Otherwise, we will actually end up on the second record after Resync
  is called. }

procedure TsesDataSet.InternalFirst;
begin
  FCurRec := -1;
end;

{ Again, we position to the crack *after* the last record here. }

procedure TsesDataSet.InternalLast;
begin
  FCurRec := FData.Count;
end;

{ This method is called by TDataSet.Post.  Most implmentations would write
  the changes directly to the associated datasource, but here we simply set
  a flag to write the changes when we close the dateset. }

procedure TsesDataSet.InternalPost;
begin

  FSaveChanges := True;
  { For inserts, just update the data in the string list }
  if State = dsEdit then
  begin
  FData[FCurRec] := ActiveBuffer;
  end
  else
  begin
    { If inserting (or appending), increment the bookmark counter and
      store the data }
  If FData.Count = 0 then
    begin
    Inc(FLastBookmark);
    FData.AddObject(ActiveBuffer, Pointer(FLastBookmark));
    end
    else
    begin
    Inc(FLastBookmark);
    FData.InsertObject(FCurRec, ActiveBuffer, Pointer(FLastBookmark));
    end;
  end;
end;

{ This method is similar to InternalPost above, but the operation is always
  an insert or append and takes a pointer to a record buffer as well. }

procedure TsesDataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
  FSaveChanges := True;
  Inc(FLastBookmark);
  if Append then InternalLast;
  FData.InsertObject(FCurRec, PChar(Buffer), Pointer(FLastBookmark));
end;

{ This method is called by TDataSet.Delete to delete the current record }

procedure TsesDataSet.InternalDelete;
begin
  FSaveChanges := True;
  FData.Delete(FCurRec);
  if FCurRec >= FData.Count then
    Dec(FCurRec);
end;

{ Optional Methods }
{ ================ }

{ The following methods are optional.  When provided they will allow the
  DBGrid and other data aware controls to track the current cursor postion
  relative to the number of records in the dataset.  Because we are dealing
  with a small, static data store (a stringlist), these are very easy to
  implement.  However, for many data sources (SQL servers), the concept of
  record numbers and record counts do not really apply. }

function TsesDataSet.GetRecordCount: Longint;
begin
  Result := FData.Count;
end;

function TsesDataSet.GetRecNo: Longint;
begin
  UpdateCursorPos;
  if (FCurRec = -1) and (RecordCount > 0) then
    Result := 1 else
    Result := FCurRec + 1;
end;

procedure TsesDataSet.SetRecNo(Value: Integer);
begin
  if (Value >= 0) and (Value < FData.Count) then
  begin
    FCurRec := Value - 1;
    Resync([]);
  end;
end;

// the classic DataSet Locate...
//Searches the dataset for a specified record and makes that record the current record.
function TsesDataSet.Locate(const KeyFields: string; const KeyValues: Variant;
  Options: TLocateOptions): Boolean;
begin
Result := Seek(KeyFields,KeyValues,Options, soFirst,True);
end;
// the extended sesDataSet locate function...
//Searches the dataset for a specified record and
// returns False, if record is not found,
// note, that the Search values are given by string list, separated by sesFieldSeparator
function TsesDataSet.Search(const KeyFields: string; const KeyValues: String;
  Options: TLocateOptions; SeekOption:TSeekOption; goForward:Boolean): Boolean;
var
I:Integer;
List:TWorkList;
V:Variant;
S:String;
begin
try
List:=TWorkList.Create;
ParseString(KeyValues,';','',List);
If List.Count = 1 then
   begin
   S:= List[0];
   V:=S;
   end
   else
   begin
   V := VarArrayCreate([0, List.Count-1], varVariant);
   for I:=0 to List.Count-1 do
       V[I] := List[I];

   end;

      Result := Seek(KeyFields, V, Options,
                   SeekOption, goForward);

finally
List.Free;
end;

end;

// the extended sesDataSet locate function...
//Searches the dataset for a specified record and
// returns False, if record is not found.
function TsesDataSet.Seek(const KeyFields: string; const KeyValues: Variant;
  Options: TLocateOptions; SeekOption:TSeekOption; goForward:Boolean): Boolean;
var
  I, FieldCount, PartialLength,J1,J2: Integer;
  Buffer: PChar;
  CaseInsensitive: Boolean;
  MultiKey:Boolean;
  Found:Boolean;
  F:TField;
  V:Variant;
  S:String;
  D1,D2:TDateTime;
  OldPos:Integer;
function EndOfLoop:Boolean;
begin
If goForward then
   Result := Self.Eof
   else
   Result := Self.Bof;
end;
procedure GotoNextRecord;
begin
If goForward then
   Next
   else
   Prior;

end;
begin
  Result := False;
  MultiKey := False;

  If Trim(KeyFields) = '' then
              begin
              Application.MessageBox(PChar(sesDataSet_LocateFieldListIsEmpty),PChar(Self.Name),MB_ICONSTOP);
              Exit;
              end;

  try
  CheckBrowseMode;
  //CursorPosChanged;
  OldPos := Self.RecNo;
  ParseString(KeyFields,';','',FWorkList1);
    CaseInsensitive := loCaseInsensitive in Options;
        FieldCount := FWorkList1.Count;

       if VarIsArray(KeyValues) then
           MultiKey := True
           else
           begin
           MultiKey:=False;
           V := KeyValues;
           end;

// from where?
If SeekOption = soFirst then
   First
   else
   If (SeekOption = soIgnoreCurrent) and (goForward) then
      Next
      else
      If (SeekOption = soIgnoreCurrent) and (not goForward) then
         Prior;
      // so, if SeekOption = soCurrent, the start search position is current...

// where?
While not EndOfLoop do
begin
  Found:=True;
  for I := 0 to FieldCount - 1 do
      begin
      F:=FindField(FWorkList1[I]);
      If F = nil then
              begin
              Application.MessageBox(PChar(sesDataSet_LocateUnknownField+':'+FWorkList1[I]),PChar(Self.Name),MB_ICONSTOP);
              Exit;
              end;
      If F.DataType = ftString then
         begin
        PartialLength := 0;
        If MultiKey then
           S:= VarToStr(KeyValues[I])
           else
           S:= VarToStr(KeyValues);
           J1:= Length(S);
           J2:= Length(F.AsString);
        if (loPartialKey in Options) then
            begin
            If CaseInsensitive then
               Found := ANSIUpperCase(Copy(F.AsString,1, J1)) =
                        ANSIUpperCase(Copy(S,1,J1))
                else
                Found := Copy(F.AsString,1,J1) = Copy(S,1,J1);
            end
            else
            begin
            If CaseInsensitive then
               Found := ANSIUpperCase(F.AsString) =ANSIUpperCase(S)
                else
                Found := (F.AsString = S);
            end;

         If not Found then
            begin
            Found:=False;
            Break;
            end;

         end
         else
         begin
  Case F.DataType Of
   ftDate:
         begin
         If MultiKey then
            D1:= VarToDateTime(KeyValues[I])
            else
            D1:=VarToDateTime(KeyValues);
            D2 := F.AsDateTime;
         If D1 <> D2 then
            begin
            Found:=False;
            Break;
            end;
         end;
   ftTime:
         begin
         If MultiKey then
            D1:= VarToDateTime(KeyValues[I])
            else
            D1:=VarToDateTime(KeyValues);
            D2 := F.AsDateTime;
         If D1 <> D2 then
            begin
            Found:=False;
            Break;
            end;
         end;

   ftInteger,ftWord,ftFloat:
         begin
         If MultiKey then
            V:= KeyValues[I]
            else
            V:=KeyValues;
         If F.Value <> V then
            begin
            Found:=False;
            Break;
            end;
         end;
   end; // Case Of DataType
         end;
      end; // end loop
      If Found then
         begin
         Result := True;
         Exit;
         end;

GotoNextRecord;

end; // Eof or BoF

Self.RecNo := OldPos;


  except
  Application.MessageBox(PChar(sesDataSet_InvalidLocateConditions),PChar(Self.Name),MB_ICONSTOP);
  Exit;
  end;
end;

    // The structure maintenance functions
    procedure TsesDataSet.sesFieldsClear; // clears all fields description, no structure will exists...
    begin
    sesFields.Clear;

    end;

    // appends the new field definition into the end if the sesFields' list
    function TsesDataSet.sesAddFieldDef(const sesField:TsesFieldDef):Boolean;
    var
    S:String;
    begin
    Result := False;
    S:=FieldDefToString(sesField);
    If S = '' then Exit;
    FsesFields.Append(S);
    Result := True;
    end;

    // inserts the new field definition just before the field with number Index of the sesFields' list
    function TsesDataSet.sesInsertFieldDef(const Index:Integer; const sesField:TsesFieldDef):Boolean;
    var
    S:String;
    begin
    Result := False;
    If (Index < 0) or (Index > FsesFields.Count-1) then Exit;
    Result := False;
    S:=FieldDefToString(sesField);
    If S = '' then Exit;
    FsesFields.Insert(Index,S);
    Result := True;
    end;

    // returns the valid ses field definition on base the existing Field
    // this function's behaviour is determined by ImportOptions
    function TsesDataSet.sesBorrowFieldDef(const Field:TField):TsesFieldDef;
    begin
    Result.Valid := False;
    Result.FName := Field.FieldName;
    Case  Field.DataType Of
    ftString,ftInteger,ftWord,ftFloat,
    ftDate,ftTime: Result.FDataType := Field.DataType;
    ftDateTime:
                     begin
                     If impDateTimeAsDate in FImportOptions then
                        Result.FDataType := ftDate
                        else
                        Exit;
                     end;
    ftMemo,ftFmtMemo:
                     begin
                     If impMemoAsString in FImportOptions then
                        Result.FDataType := ftString
                        else
                        Exit;
                     end;
    ftSmallInt,ftLargeInt,ftAutoInc:
                     begin
                     If impAllIntAsInteger in FImportOptions then
                        Result.FDataType := ftInteger
                        else
                        Exit;
                     end;

    ftCurrency:
                     begin
                     If impCurrencyAsFloat in FImportOptions then
                        Result.FDataType := ftFloat
                        else
                        Exit;
                     end;


    else
       begin
       Result.FDataType := Field.DataType;
       Exit;
       end;
    end; // Case Field.DataType


    Result.FLabel := Field.DisplayLabel;
    Result.FWidth := Field.DisplayWidth;
    Result.FSize := Field.Size;
    Result.FRequired := Field.Required;
    Result.FAlignment := Field.Alignment;
    Result.Valid := True;
    end;

    // converts the sesFieldDef into "string" representation of a field definition and
    // returns definition string...
    function TsesDataSet.FieldDefToString(const sesFieldDef:TsesFieldDef):String;
    begin
    Result := '';
    With sesFieldDef do
      begin
      If Trim(FName) = '' then Exit;
      If not (FDataType in [ftString,ftInteger,ftWord,ftFloat,ftDate,ftTime]) then Exit;
      Result := Result+FName + sesFieldSeparator;
      Result := Result+IntToStr(Integer(FDataType)) + sesFieldSeparator;
      Result:= Result+FLabel + sesFieldSeparator;
      If FWidth = 0 then
         Result:= Result+'10' + sesFieldSeparator
         else
         Result:= Result+IntToStr(FWidth) + sesFieldSeparator;
      Case FDataType Of
      ftString:
               begin
               If FSize <= 0 then
                  Result:= Result+'240' + sesFieldSeparator
                  else
                  Result:= Result+IntToStr(FSize) + sesFieldSeparator;
               end;
       else
               begin
               Result:= Result+'0' + sesFieldSeparator;
               end;
      end; // case FDataType
      If FRequired then
         Result:= Result+'1' + sesFieldSeparator
         else
         Result:= Result+'0' + sesFieldSeparator;
      Result:= Result+IntToStr(Integer(FAlignment));
      end;

    end;

    // creates and returns the full desctiption of the Structure in formatted string:
    {FieldDef0@FieldDef1@...FieldDef(Count-1)
    where FieldDef0 - string representation of a field definition number 0
    }
function TsesDataSet.StructureToString:String;
var
I:Integer;
begin
Result := '';
If FsesFields.Count = 0 then Exit;
Result := sesFieldDefSeparator+sesFieldDefSeparator+sesFieldDefSeparator;
For I:=0 to  FsesFields.Count-1 do
    begin
    If I = FsesFields.Count-1 then
       Result := Result + FsesFields[I]
       else
       Result := Result + FsesFields[I] + sesFieldDefSeparator;
    end;
end;
//creates sesFields structure on base of formatted StruString
function TsesDataSet.StringToStructure(const StruString:String):Boolean;
var
I:Integer;
S:String;
begin

FsesFields.Clear;
Result := False;
If Copy(StruString,1,3) <> sesFieldDefSeparator+sesFieldDefSeparator+sesFieldDefSeparator then Exit;

S:=Copy(StruString,4,2048);
ParseString(S,sesFieldDefSeparator,'', FWorkList1);
If FWorkList1.Count = 0 then Exit;

For I:=0 to  FWorkList1.Count-1 do
    FsesFields.Add(FWorkList1[I]);
Result := True;
end;


{
 saves FData strings to the file with full name aFileName ,
  with the first string as the strings representation of the exeisting Structure, if
  WithStructure is True;
 returns True, if successful saving has been executed...
}
function TsesDataSet.SaveToFile(const aFileName:String; const WithStructure:Boolean):Boolean;
var
S:String;
begin
Result := False;
If aFileName = '' then Exit;
If (WithStructure) and (FsesFields.Count > 0) then
    begin
    S := StructureToString;
    If S = '' then Exit;
    FData.Insert(0,S);
    end;
try
FData.SaveToFile(aFileName);
If (WithStructure) and (FsesFields.Count > 0) then
    begin
    FData.Delete(0);
    end;
Result := True;
except
Result := False;
If (WithStructure) and (FsesFields.Count > 0) then
    begin
    FData.Delete(0);
    end;
end;

end;

// sorts the FData strings by OrderBy list of fields name
    {returns False, if:
    - OrderBy is empty or invalid;
    - sesFields is empty;
    - FData is empty;
    - during sorting a error type conversion has been encountered/
    }
function TsesDataSet.Sort(const OrderBy:String; const Ascending:Boolean;
    CaseInSensitive:Boolean; StringAsStringInt:Boolean ):Boolean;
var
I,J,N:Integer;
RCloseMode:TCloseMode;
ROpenMode:TOpenMode;
FAct:Boolean;
SortPars:String;
fOrder:Integer;
S,fTypeStr:String;
fType:TSortDataType;
begin
Result := False;
If FData.Count = 0 then Exit;
If Trim(OrderBy) = '' then Exit;
If FsesFields.Count = 0 then Exit;
ParseString(OrderBy,';','', FWorkList1);
If FWorkList1.Count = 0 then Exit;
// check up the fields list and fills sort parameters
SortPars:='';
for I:=0 to FWorkList1.Count - 1 do
    begin
    fOrder:= GetsesFieldOrder(FWorkList1[I]);
    If (fOrder = -1) then Exit;
    fType := GetsesFieldType(FWorkList1[I]);
    If fType = sdUnknown then Exit;
    If (fType = sdString) and  StringAsStringInt then
       fType := sdStringInteger;
    fTypeStr := IntToStr(Integer(fType));

    If I = FWorkList1.Count - 1 then
       SortPars := SortPars + IntToStr(fOrder)+SortFieldTypeSeparator+fTypeStr
       else
       SortPars := SortPars + IntToStr(fOrder)+SortFieldTypeSeparator+fTypeStr+SortFieldSeparator;
    end;

FAct := Active;
RCloseMode:= FCloseMode;
FCloseMode := cmNone;
ROpenMode:= FOpenMode;
FOpenMode := omNone;
Self.Close;


SortStorage := TQuickSortStorage.CreateSortStorage(FData.Count, FsesFields.Count);
//1. fill storage
for I:= 0 to fData.Count-1 do
    begin
    for J:=0 to FsesFields.Count-1 do
        SortStorage.SetField(J,I, GetsesFieldText(J,I));
    end;

//2. set sort params
SortStorage.CaseInSensitive := CaseInSensitive;
SortStorage.SortBy := SortPars;

try
//3. execute sorting
SortStorage.SortExecute;
except
FCloseMode := RCloseMode;
//FOpenOptions:=FOpenOpt;
FOpenMode := ROpenMode;
If FAct then
Self.Open;
Exit;
end;
//4. retrieve new ordered strings
If Ascending then
begin
for I:= 0 to fData.Count-1 do
    begin
    S:='';
    for J:=0 to FsesFields.Count-1 do
        begin
        If J = FsesFields.Count-1 then
           S := S +SortStorage.GetField(J,I)
           else
           S := S +SortStorage.GetField(J,I) + sesFieldSeparator;

        end;
    FData[I] := S;
    end;
end
else
begin
N:=0;
for I:= fData.Count-1 downto 0 do
    begin
    S:='';
    for J:=0 to FsesFields.Count-1 do
        begin
        If J = FsesFields.Count-1 then
           S := S +SortStorage.GetField(J,I)
           else
           S := S +SortStorage.GetField(J,I) + sesFieldSeparator;

        end;
    FData[N] := S;
    N:=N+1;
    end;
end;
//5. free sort storage
SortStorage.Free;

//6. restore dataset status
Result := True;
FCloseMode := RCloseMode;

If FAct then
Self.Open;
FOpenMode := ROpenMode;
end;

//imports the sesFields structure from a existing and open DataSet,
// using the ImportOptions and selecting fields by FieldFilter,
// which must present semicolon separated list of including field names.
// note. If FieldFilter is empty string, it means all fields will be imported.
function TsesDataSet.Import(const FieldFilter:String; const DataSet:TDataSet):Boolean;
var
I,J:Integer;
fDef:TsesFieldDef;
All:Boolean;
F:TField;
begin
Result := False;
If DataSet.Active = False then Exit;
ClearStructure;
FWorkList1.Clear;
FWorkList2.Clear;
If FieldFilter <> '' then
   ParseString(FieldFilter,';','', FWorkList2);
If FWorkList2.Count > 0 then
   All := False
   else
   All := True;
for I:=0 to DataSet.FieldCount-1 do
    begin
    F := DataSet.Fields[I];
    If not All then
       begin
       J := FWorkList2.IndexOf(F.FieldName);
       If J = -1 then Continue;
       end;
    fDef := sesBorrowFieldDef(F);
    If not fDef.Valid  then Continue;
    sesAddFieldDef(fDef);
    end;

end;
end.
