unit sesEditUnit;

interface

uses
  sesStrings, sesSystem,
  Windows, Messages, SysUtils, Classes, Graphics, stdctrls,Controls,
  DB,FileCtrl,Inifiles,Math,Buttons,extctrls, Forms, Dialogs, Menus;


const
sesFalse = '0';
sesTrue='1';


type

{TsesLogger}
TsesLogger = class;
TLoggerFileStorage = TIniFile;

// TsesLogger defaults type: if FileName does not contain a dir part
TDirDefaultType = (dfWinDir,dfExecute,dfUser);

// the all properties of this type will be outlined by Inspector with the picture "+"
TGroupProperties = class(TPersistent);

// Dir type for dialog property Editor
TDirectory = String;

{TsesCustomEdit class}
TsesCustomEdit = class;


{sesEditor types}
TsesCustomEditor = class;


// The TCheckTextEvent type points to a method that handles the additional
// process of the TsesEdit Text value
TCheckTextEvent = procedure (Sender: TObject; var TextIsValid:Boolean) of object;

//The UserErrorProcess type points to a method that shows the user the
// error description message by dialog or form linked with CheckText event
TUserErrorProcess = procedure (Sender: TObject; var fModalResult:TModalResult) of object;


TLoggedType = (ltNone,ltUnique,ltDated);

TClassicEditControlOption = (goCanExitOnError, goCheckChar, goCheck);
TClassicEditControlOptions = set of TClassicEditControlOption;




TOnEditButtonClick = procedure(Sender: TObject; ScreenPopUpPoint,FormPopUpPoint:TPoint;
var aSource:String; var Changed:Boolean) of object;

TOnErrorKey = procedure(Sender: TObject;
var Key:Char; var Replace:Boolean) of object;




{TsesCustomEdit types}

// Edit State
TsesState = (ssInit,ssError,ssEdit,ssComplete);

// PopUp menu options define an apperance of the PopUp menu
TPopupOption = (poCalculator,poHistory,poPickList,poUserForm, poDatePicker, poTimePicker);
TPopupOptions = set of TPopupOption;

{Init processing: begin a new edit session}
TInitReset = (irInitial, irCurrent, irLastValid);

{Error processing}
// show Error/Hint Dialog
TErrorDialog = (edNone,edShow);
// Exit allowing: no or if Edit is having the State = sesInit(just after init)
TErrorExit = (eeNone,eeIfInit);
// generate sound on error: no or system Beep or special sound wave file from Windows dir
TErrorSound = (esNone,esBeep,esPlayWave);

// Edit state view:
TStateView = (svNone, svShowFrame);

{ kind of interpetation the PickList strings:
 plStandard -  only One field in each string;
 plCodeCaption - two field: tested Code and desciption(Caption),
                 which must be divided by PickListSeparator.
 plGrid - multi-fields in each string, divided by PickListSeparator,
          the first string(index=0) must consist of headers.
}
TsesPickListKind =(plStandard, plCodeCaption, plGrid, plDataSet, plNone);

{PickButton default action}
TPickButtonAction = (pbNone,pbCalculator,pbHistory,pbPickList,pbUserForm,pbDatePicker,pbTimePicker);

// autosize options during the changing of Layout
//TAutosizeOption = (aoNone, aoDecEditor, aoIncTotalWidth);

// TVertAlign type defines 3 kinds of vertical subject alignment:
TVertAlignment =(dtTop, dtCenter, dtBottom);

// Caption layout relative to Edit&Button types:
TsesLayout = (clTop,clLeft);

// Default Button glyphs
TDefaultGlyph = (dgNone, dgUserForm,dgPickList, dgHistory, dgCalculator,dgDatePicker,dgTimePicker);

  // declare additional class to public Container properties we needed...
  TsesCustomPanel = class(TCustomPanel)
  public
   property Align;
   property BevelInner;
   property BevelOuter;
   property BevelWidth;
   property BorderStyle;
   property BorderWidth;
   property Color;
   property Caption;
   property Font;
   property ParentFont;

  end;

  // the properties of the exterior frame of the TsesCustomEdit(container)
  TsesFrame = class(TGroupProperties)
   private
   FInteriorIndent:Word; // Edit&Button offset from BevelInner


   procedure SetBevelInner(const Value:TPanelBevel);
   procedure SetBevelOuter(const Value:TPanelBevel);
   procedure SetBevelWidth(const Value:TBevelWidth);
   procedure SetBorderStyle(const Value:TBorderStyle);
   procedure SetBorderWidth(const Value:TBorderWidth);
   procedure SetInteriorIndent(const Value:Word);


   function GetBevelInner:TPanelBevel;
   function GetBevelOuter:TPanelBevel;
   function GetBevelWidth:TBevelWidth;
   function GetBorderStyle:TBorderStyle;
   function GetBorderWidth:TBorderWidth;
   function GetInteriorIndent:Word;

   public
   Frame:TsesCustomPanel;
   published

   property BevelInner:TPanelBevel read GetBevelInner write SetBevelInner;
   property BevelOuter:TPanelBevel read GetBevelOuter write SetBevelOuter;
   property BevelWidth:TBevelWidth read GetBevelWidth write SetBevelWidth;
   property BorderStyle:TBorderStyle read GetBorderStyle write SetBorderStyle;
   property BorderWidth:TBorderWidth read GetBorderWidth write SetBorderWidth;
   property InteriorIndent:Word read GetInteriorIndent write SetInteriorIndent;
   end;

  // the properties of the DisplayLabel
  TsesCaption = class(TGroupProperties)
   private
   FAlignment:TAlignment;         // Caption text horizontal alignment
   FLayout:TsesLayout;
   FVisible:Boolean;              // Caption visibility
   FVertAlignment:TVertAlignment; // Caption text vertical alignment
   FWordWrap:Boolean;

   procedure SetAlignment(const Value:TAlignment);
   procedure SetLayout(const Value:TsesLayout);
   procedure SetVisible(const Value:Boolean);
   procedure SetVertAlignment(const Value:TVertAlignment);
   procedure SetWordWrap(const Value:Boolean);

   function GetAlignment:TAlignment;
   function GetVisible:Boolean;
   function GetVertAlignment:TVertAlignment;
   function GetWordWrap:Boolean;

   public
   Frame:TsesCustomPanel;
   published
   property Alignment:TAlignment read GetAlignment write SetAlignment;
   property Layout:TsesLayout read FLayout write SetLayout; // Caption layout relative to Edit&Button
   property Visible:Boolean read GetVisible write SetVisible;
   property VertAlignment:TVertAlignment read GetVertAlignment write SetVertAlignment;
   property WordWrap:Boolean read GetWordWrap write SetWordWrap;
   end;

   // the properties of the Edit's Button component
  TsesButton = class(TGroupProperties)
   private
   EditorGap:Integer; // the gap between Edit&Button
   FButtonWidth:Integer;
   FButtonVisible:Boolean;
   FDefaultGlyph:TDefaultGlyph;
   procedure SetCaption(const Value:String);
   procedure SetDefaultGlyph(const Value:TDefaultGlyph);
   procedure SetEditGap(const Value:Integer);
   procedure SetEnabled(const Value:Boolean);

   procedure SetGlyph(const Value:TBitMap);
   procedure SetNumGlyphs(const Value:TNumGlyphs);
   procedure SetLayout(const Value:TButtonLayout);
   procedure SetVisible(const Value:Boolean);
   procedure SetFlat(const Value:Boolean);
   procedure SetWidth(const Value:Integer);

   function GetCaption:String;
   function GetEnabled:Boolean;
   function GetGlyph:TBitMap;
   function GetNumGlyphs:TNumGlyphs;
   function GetLayout:TButtonLayout;
   function GetFlat:Boolean;
   function GetVisible:Boolean;
   function GetWidth:Integer;

   public

   Edit:TsesCustomEdit;

   published
   property Caption:String read GetCaption write SetCaption;
   property DefaultGlyph:TDefaultGlyph read FDefaultGlyph write SetDefaultGlyph;
   property EditGap:Integer read EditorGap write SetEditGap;
   property Enabled:Boolean read GetEnabled write SetEnabled;
   property Flat:Boolean read GetFlat write SetFlat;
   property Glyph:TBitMap read GetGlyph write SetGlyph;
   property NumGlyphs:TNumGlyphs read GetNumGlyphs write SetNumGlyphs;
   property Layout:TButtonLayout read GetLayout write SetLayout;
   property Width:Integer read GetWidth write SetWidth;
   property Visible:Boolean read GetVisible write SetVisible;

   end;


   // the properies of the Editor component
  TsesEditGroup = class(TGroupProperties)
   private
   procedure SetColor(const Value:TColor);
   procedure SetFont(const Value:TFont);
   procedure SetParentFont(const Value:Boolean);
   procedure SetWidth(const Value:Integer);


   function GetColor:TColor;
   function GetFont:TFont;
   function GetParentFont:Boolean;
   function GetWidth:Integer;

   public
   Edit:TsesCustomEditor;
   property Width:Integer read GetWidth write SetWidth;
   published
   property Color:TColor read GetColor write SetColor;
   property Font:TFont read GetFont write SetFont;
   property ParentFont:Boolean read GetParentFont write SetParentFont;
   end;

   // the group of Data properies of the Editor component
  TsesDataGroup = class(TGroupProperties)
   private

   procedure SetRanged(const Value:Boolean);
   procedure SetRangeMinOrFalse(const Value:String);
   procedure SetRangeMaxOrTrue(const Value:String);

   function GetRangeMinOrFalse:String;
   function GetRangeMaxOrTrue:String;
   function GetRanged:Boolean;

   public
   Editor:TsesCustomEditor;
   published
    property Ranged:Boolean read GetRanged write SetRanged;
    property RangeMinOrFalse:String read GetRangeMinOrFalse write SetRangeMinOrFalse;
    property RangeMaxOrTrue:String read GetRangeMaxOrTrue write SetRangeMaxOrTrue;
   {
   property Font:TFont read GetFont write SetFont;
   property ParentFont:Boolean read GetParentFont write SetParentFont;
   property Text:String read FText write FText;
   property Width:Integer read GetWidth write SetWidth;
   }
   end;



  TsesCustomEdit = class(TCustomPanel)
  private
    { Private declarations }
    // internal variables
    InsideRect:TRect;
    // resizing internal flag
    FForceAdjust:Boolean;
    FResizing:Boolean;

    // properies' fields
    FButton:TsesButton;
    FCancel:Boolean;
    FDecimalKeyAsSeparator:Boolean;
    FEnabled:Boolean;

    FErrorDialog:TErrorDialog;
    FErrorExit:TErrorExit;
    FErrorSound:TErrorSound;

    FGroupIndex:Integer;

    FInitReset:TInitReset;

    FLogger:TsesLogger;
    FLoggerLimit:Word;
    FLoggedType:TLoggedType;

    FOnCheckText:TCheckTextEvent;
    FOnUserErrorProcess:TUserErrorProcess;

    FOnChange: TNotifyEvent;
    FOnComplete:TNotifyEvent;
    FOnPickButtonClick:TNotifyEvent;

    FPopupOptions:TPopupOptions;
    FPickDataSet:TDataSet;
    FPickList:TStrings;
    FPickListKind:TsesPickListKind;
    FPickListSeparator:String;
    FReadOnly:Boolean;
    FRequired:Boolean;
    FsesCaption:TsesCaption;
    FsesData:TsesDataGroup;
    FsesEditor:TsesEditGroup;
    FsesState:TsesState;
    FsesFrame:TsesFrame;
    FStateView:TStateView;
    FTabStopEditor:Boolean;
    FTabNext:TWinControl;
    FValue:Variant;

  function GetDataType:TClassicDataType;
  function  GetEnterAsTab:Boolean;
  function GetModified:Boolean;

  function GetsesEnabled:Boolean;
  function GetsesTabStop:Boolean;
  function GetText:TCaption;

  procedure SetDataType(const Value:TClassicDataType);
  procedure SetEnterAsTab(const Value:Boolean);
  procedure SetLogger(const Value:TsesLogger);
  procedure SetModified(const Value:Boolean);
  procedure SetPopupOptions(const Value:TPopupOptions);

  procedure SetPickList(const Value:TStrings);
  procedure SetPickDataSet(const Value:TDataSet);
  procedure SetReadOnly(const Value:Boolean);
  procedure SetStateView(const Value:TStateView);
  procedure SetText(const Value:TCaption);
  procedure SetsesEnabled(const Value:Boolean);
  procedure SetsesState(const Value:TsesState);
  procedure SetsesTabStop(const Value:Boolean);
  procedure SetVarValue(const aValue:Variant);

  protected
    { Protected declarations }
  procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
  function DoCheckText:Boolean;
  function DoUserErrorProcess:TModalResult;

  function GetAsVariant:Variant;
  function GetCaptionRect:TRect;
  function PopUpMenuEnabled:Boolean;
  procedure PassFocus;

  procedure AdjustClientSize(const aLayout:TsesLayout);
  procedure AdjustActualSize;
  procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
  procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
  procedure DoExit;override;
  procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  procedure Loaded;override;
  procedure Notification(AComponent: TComponent;
    Operation: TOperation);override;
  procedure Paint; override;
  procedure PickButtonClick(Sender:TObject);
  procedure PopUpMenuClick(Sender:TObject);
  procedure WMSize(var Message:TWMSize); message WM_SIZE;
//  procedure TABSTOPCHANGED(var Message:TMessage); message CM_TABSTOPCHANGED;
  public
    { Public declarations }
    Editor:TsesCustomEditor; // editor component
    PickButton:TSpeedButton; // Pick List button
    PickButtonAction:TPickButtonAction;  // ???
    PopUpMenuButton:TSpeedButton; // PopUp Menu Button

    // old text value before logging, set to last valid value
    LastValidValue:String;
    // initial value after loaded, set during design time only...
    InitialValue:String;

    property Cancel:Boolean read FCancel write FCancel;
    property Caption;
    property DataType:TClassicDataType read GetDataType write SetDataType;
    property DecimalKeyAsSeparator:Boolean read FDecimalKeyAsSeparator write FDecimalKeyAsSeparator;

    property EnterAsTab:Boolean read GetEnterAsTab write SetEnterAsTab;

    property ErrorDialog:TErrorDialog read FErrorDialog write FErrorDialog;
    property ErrorExit:TErrorExit read FErrorExit write FErrorExit;
    property ErrorSound:TErrorSound read FErrorSound write FErrorSound;

    property GroupIndex:Integer read FGroupIndex write FGroupIndex;

    property InitReset:TInitReset read FInitReset write FInitReset;

    property Logger:TsesLogger read FLogger write SetLogger;
    property LoggerLimit:Word read FLoggerLimit write FLoggerLimit;
    property LoggedType:TLoggedType read FLoggedType write FLoggedType;

    property Modified:Boolean read GetModified write SetModified;

    property OnCheckText:TCheckTextEvent read FOnCheckText write FOnCheckText;
    property OnUserErrorProcess:TUserErrorProcess read FOnUserErrorProcess write FOnUserErrorProcess;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnComplete: TNotifyEvent read FOnComplete write FOnComplete;
    property OnPickButtonClick:TNotifyEvent read FOnPickButtonClick write FOnPickButtonClick;

    property PopupOptions:TPopupOptions read FPopupOptions write SetPopupOptions;
    property PickDataSet:TDataSet read FPickDataSet write SetPickDataSet;
    property PickList:TStrings read FPickList write SetPickList;
    property PickListKind:TsesPickListKind read FPickListKind write FPickListKind;
    property PickListSeparator:String read FPickListSeparator write FPickListSeparator;

    property ReadOnly:Boolean read FReadOnly write SetReadOnly;
    property Required:Boolean read FRequired write FRequired;

    property sesState:TsesState read FsesState write SetsesState;
    {grouped properties}
    property sesButton:TsesButton read FButton write FButton;
    property sesCaption:TsesCaption read FsesCaption write FsesCaption;
    property sesData:TsesDataGroup read FsesData write FsesData;
    property sesEditor:TsesEditGroup read FsesEditor write FsesEditor;
    property sesFrame:TsesFrame read FsesFrame write FsesFrame;

    property StateView:TStateView read FStateView write SetStateView;
    property TabStopEditor:Boolean read GetsesTabStop write SetsesTabStop;
    property TabNext:TWinControl read FTabNext write FTabNext;

    property Text:TCaption read GetText write SetText;
    property Value:Variant read GetAsVariant write SetVarValue;


    procedure SetFocus; override;

    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;

    // reposition functions
    procedure AdjustPositions;
    function GetControlsRect:TRect;
    function GetTotalBorderWidth:Integer;

    // logger functions
    procedure PostLog;
    // PickList functions
    function GetPickListCaption(const Full:Boolean):String;

    // edit session functions
    function CheckText:TEditErrorType;
    //takes the focus from the focused TSesEdit regardless its data validity and
    // passes it to the aWinControl
    procedure PassFocusTo(aWinControl:TWinControl);
    procedure SetInitText;
    function GetInitText:String;
    function GetState:TsesState;

    // error processing functions
    procedure sesSayMessage(ErrorKind:TEditErrorType);

    // Edit access functions
    function AsBoolean:Boolean;
    function AsCardinal:Cardinal;
    function AsInteger:Integer;
    function AsRound(const Precision:Integer):Double;
    function AsFloat:Double;

    function AsDate:TDate;
    function AsDateNullTime:TDateTime;
    function AsTime:TTime;
    function AsTimeNullDate:TDateTime;


  published
    { Published declarations }
    property Enabled:Boolean read GetsesEnabled write SetsesEnabled;
  end;


{TsesEdit}
{The main edit control properties}
  TsesEdit = class(TsesCustomEdit)
  published
  property Align;
  property Anchors;
          property Cancel;
  property Caption;
  property Color;
  property Constraints;
          property DataType;
          property DecimalKeyAsSeparator;
          property EnterAsTab;
          property ErrorDialog;
          property ErrorExit;
          property ErrorSound;
          property GroupIndex;
          property InitReset;
          property Logger;
          property LoggerLimit;
          property LoggedType;
          property PopupOptions;
          property PickDataSet;
          property PickList;
          property PickListKind;
          property PickListSeparator;

  property UseDockManager default True;
  property DockSite;
  property DragCursor;
  property DragKind;
  property DragMode;

  property Font;
          property OnCheckText;
          property OnUserErrorProcess;
          property OnChange;
          property OnComplete;
          property OnPickButtonClick;
          property OnEnter;
          property OnExit;
  property ParentBiDiMode;
  property ParentFont;
  property ParentColor;
  property ParentShowHint;
  property PopUpMenu;

  property ReadOnly;
    property Required;
    property sesButton;
    property sesCaption;
    property sesData;
    property sesEditor;
    property sesFrame;
    property TabStopEditor;
    property TabNext;
    property StateView;

  property ShowHint;
  property TabOrder;
          property Text;
  property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnGetSiteInfo;
        property OnKeyDown;
        property OnKeyUp;
        property OnKeyPress;

    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;


  end;

{TsesLogger}
TsesLogger = class(TComponent)
  private
  FActive:Boolean;
  FDirDefault:TDirectory;
  FDirDefaultType:TDirDefaultType;
  FEditsList:TStringList;
  FFileNameStorage:String;
  FFileName:TFileName;
  FFileStorage:TLoggerFileStorage;
  FInitTruncation:Boolean;
  FStatus:Word;

  // trigger events

  // fires just after linked dataset has been set at a new record
  FOnAfterScroll:TNotifyEvent;

//  FOnPost:TNotifyEvent;
//  FOnAppend:TNotifyEvent;

  function GetEdit(Index:Integer):TsesCustomEdit;
  function GetEditsCount:Integer;

  procedure SetActive(const Value:Boolean);
  procedure SetDirDefault(const Value:TDirectory);
  procedure SetDirDefaultType(const Value:TDirDefaultType);
  procedure SetFileName(const Value:TFileName);

  protected
  procedure Loaded;override;
//  procedure Paint;override;
  procedure SetRuntimeDirDefault;
  function Activate(Start:Boolean):Word;
  procedure DeActivate;

  public


  property Edits[Index:Integer]:TsesCustomEdit read GetEdit;
  property EditsCount:Integer read GetEditsCount;
  property FileStorage:TLoggerFileStorage read FFileStorage;
  property FileNameRuntime:String read FFileNameStorage;


  { Logger status:
    0 - Active;
    1 - inActive;
    2 - Error on creation of Storage
  }
  property Status:Word read FStatus write FStatus;

  constructor Create(AOwner:TComponent);override;
  destructor Destroy;override;

   // log functions
  function GetHistoryList(const Sender:TsesCustomEdit;
   const List:TStrings):Boolean;
  procedure Log(const Sender:TsesCustomEdit);
  procedure LogEdits;
  procedure RegisterEdit(const Sender:TsesCustomEdit);
  // truncate a history list of Sender to it's threshold, if assigned
  procedure TruncateLog(const Sender:TsesCustomEdit; UpdateFile:Boolean);
  // truncate a history list of all sesEdits to it's threshold, if assigned
  procedure TruncateLogs;

  // service functions
  function  ReadControlRect(const Sender:TObject):Boolean;
  procedure SaveControlRect(const Sender:TObject);

  // Edits group operations
  function Complete:Boolean;overload;
  function Complete(GroupIndex:Integer):Boolean;overload;
  function EditByName(const EditName:String):TsesEdit;
// resets all edits to initial state and  Text to Init Text
  procedure InitEdits;overload;
  procedure InitEdits(GroupIndex:Integer);overload;

  function IsEditCompleted:TsesEdit;overload;
  function IsEditCompleted(GroupIndex:Integer):TsesEdit;overload;

  procedure SetEditsState(const Value:TsesState);overload;
  procedure SetEditsState(const Value:TsesState; GroupIndex:Integer);overload;

  procedure PassFocusTo(AWinControl:TWinControl);

  function TrySetFocus:TsesCustomEdit;overload;
  function TrySetFocus(GroupIndex:Integer):TsesCustomEdit;overload;

  procedure RefreshEdits;

  published
  property Active:Boolean read FActive write SetActive;
  property DirDefault:TDirectory read FDirDefault write SetDirDefault;
  property DirDefaultType:TDirDefaultType read FDirDefaultType write SetDirDefaultType;
  property FileName:TFileName read FFileName write SetFileName;
  property InitTruncation:Boolean read FInitTruncation write FInitTruncation;
  property OnAfterScroll:TNotifyEvent read   FOnAfterScroll write   FOnAfterScroll;




//  property Visible;
  end;

// custom class Editor component
  TsesCustomEditor = class(TCustomEdit)
  private
    { Private declarations }
  FCheckTextResult:TEditErrorType;
  FCanLoseFocus:Boolean;
  FSeparatorPressed:Boolean;
  FRanged:Boolean;
  FRangeMinOrFalse:String;
  FRangeMaxOrTrue:String;
//  FButtonKind:TEditButtonKind;
  FEnterAsTab:Boolean;
  FErrorSoundType:TErrorSound;
  FFocusPassedOn:Boolean;

//  FOnButtonClick:TNotifyEvent;


  FButtonHotKey:TShortCut;

  FClassicDataType:TClassicDataType;

  function IsValidChar(Key: Char): Boolean;
  procedure SetClassicDataType(Value:TClassicDataType);
  procedure DoChangeText(Sender:TObject);

  protected
    { Protected declarations }
  procedure UserExitAction;
  procedure ActivateCharCheckPopUpForm(Key:Char);

  property ErrorSoundType:TErrorSound read FErrorSoundType write FErrorSoundType;

  property EnterAsTab:Boolean read FEnterAsTab write FEnterAsTab;

//  property ButtonKind:TEditButtonKind read FButtonKind write SetEditButtonKind;
  property ButtonHotKey:TShortCut read FButtonHotKey write FButtonHotKey;

//  procedure Loaded; override;
  function IsStringBoolean:Boolean;
  procedure CreateParams(var Params: TCreateParams); override;
  procedure CreateWnd; override;
  procedure WMSize(var Message: TWMSize); message WM_SIZE;
  procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  procedure KeyUp(var Key: Word; Shift: TShiftState);override;
  procedure KeyPress(var Key: Char); override;
  procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
  procedure CMExit(var Message: TCMExit);   message CM_EXIT;
  procedure CEOperation(var Message: TWMKey); message CM_DIALOGKEY;

  procedure ActivateErrorWindow;
  procedure DeActivateErrorWindow;
  public
    { Public declarations }
    property CanLoseFocus:Boolean read FCanLoseFocus write FCanLoseFocus;
    property CheckTextResult:TEditErrorType read FCheckTextResult write FCheckTextResult;
    property Color;
    property EditDataType:TClassicDataType read FClassicDataType write SetClassicDataType;
    property Font;
    property ParentFont;
    property Ranged:Boolean read FRanged write FRanged;
    property RangeMinOrFalse:String read FRangeMinOrFalse write FRangeMinOrFalse;
    property RangeMaxOrTrue:String read FRangeMaxOrTrue write FRangeMaxOrTrue;

//  property Button: TSpeedButton read FButton;
  constructor Create(AOwner:TComponent);override;
  // override this method for custom text check up
  function CheckText:TEditErrorType;virtual;

  published
    { Published declarations }
  end;



procedure AdjustPosition(Sender:TObject);
procedure AdjustActualSizes(Sender:TObject);

var
LastEdit:TsesCustomEdit;

implementation

uses sesAdviseUnit;

{$R sesButtons.RES}

{sesButton data}
var
  sesButtonResNames: array[0..6] of PChar = (
    '', 'USERFORM', 'PICKLIST', 'HISTORY','CALC','DATEPICKER','TIMEPICKER');

  sesButtonGlyphs: array[0..6] of TBitmap;


// common procedure for adjusting of components on the client area of the panel
procedure AdjustPosition(Sender:TObject);
begin
TsesCustomEdit(Sender).AdjustPositions;
//TsesCustomEdit(Sender).sesAdjustHeight;
end;

// common procedure for adjusting of components on the client area of the panel
procedure AdjustActualSizes(Sender:TObject);
begin
TsesCustomEdit(Sender).AdjustActualSize;

end;


{TsesLogger}
  constructor TsesLogger.Create(AOwner:TComponent);
  begin
  inherited Create(AOwner);
{
  FIcon:=TBitMap.Create;
  FIcon.LoadFromResourceName(HInstance, 'LOGGEROF');
  FIconLoaded :=False;

  Height:=12;
  Width:=18;
Constraints.MaxHeight:=12;
Constraints.MinHeight:=12;
Constraints.MaxWidth:=18;
Constraints.MinWidth:=18;
  ShowHint := True;
  Hint:=sesLogger_1;

 }
  FActive := False;
  FFileName := '';
  FStatus := 1;
  FFileStorage:=nil;
  FDirDefault := '';
  FDirDefaultType:=dfExecute;
  FEditsList:=TStringList.Create;
  FInitTruncation:=True;
  end;
  
  destructor TsesLogger.Destroy;
  begin
  If FStatus = 1 then
     FileStorage.Free;
  FEditsList.Free;
//  FIcon.Free;
  inherited Destroy;
  end;

// passes the focus from the focused TsesEdit to aWinControl
procedure TsesLogger.PassFocusTo(AWinControl:TWinControl);
  var
  I,N:Integer;
  begin
  If aWinControl = nil then Exit;
  If not aWinControl.CanFocus then Exit;
  N:=FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I := 0 to N do
      begin
      If TsesCustomEdit(FEditsList.Objects[I]).Editor.Focused then
         begin
         TsesCustomEdit(FEditsList.Objects[I]).PassFocusTo(aWinControl);
         If not aWinControl.Focused then
            aWinControl.SetFocus;
         Exit;
         end;
      end;
  If not aWinControl.Focused then
     aWinControl.SetFocus;

  end;


  // does nothing, excluding firing the OnAfterScroll Event
  procedure TsesLogger.RefreshEdits;
  begin
  If Assigned(FOnAfterScroll) then
     begin
     FOnAfterScroll(Self);
     IsEditCompleted;
     end;
  end;

  // if assigned LoggerLimit sesEdit history list will be truncated
  procedure TsesLogger.TruncateLog(const Sender:TsesCustomEdit; UpdateFile:Boolean);
var
  S,S1:TStringList;
  I,J,N,M,Recs:Integer;
  Section,Key,Value:String;

begin
  If not FActive then Exit;
  If FFileStorage = nil then Exit;
  If Sender.LoggerLimit = 0 then Exit;
  Section := Sender.Name;
  If not FFileStorage.SectionExists(Section) then Exit;

  S := TStringList.Create;
  S1 := TStringList.Create;
  try
    FFileStorage.ReadSection(Section, S);
    If S.Count > Sender.LoggerLimit then
       begin
       // try to truncate
       M:=S.Count-1;
       N:=Sender.LoggerLimit - 1;
       Recs := S.Count - Sender.LoggerLimit;

       I:=Recs;
       For J := 0 to N do
           begin
           Key := S[I];
           Value := FFileStorage.ReadString(Section,Key,'');
           S1.Add(Key+'='+Value);
           I:=I+1;
           If I > M then Break;
           end;
       end;
  If S1.Count = 0 then Exit;
  FFileStorage.EraseSection(Section);
  M:=S1.Count-1;
  For I := 0 to M do
      begin
      N := Pos('=',S1[I]);
      Key := Copy(S1[I],1,N-1);
      Value := Copy(S1[I],N+1,500);
      FFileStorage.WriteString(Section,Key,Value);
      end;
  If UpdateFile then
     FFileStorage.UpdateFile;
  finally
    S.Free;
    S1.Free;
  end;
end;

  procedure TsesLogger.TruncateLogs;
  var
  I,N:Integer;
  begin
  If not FActive then Exit;
  If FFileStorage = nil then Exit;
  try
  N:=FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
    TruncateLog(TsesCustomEdit(FEditsList.Objects[I]),False);

  FFileStorage.UpdateFile;
  except
  end;
  end;

  procedure TsesLogger.RegisterEdit(const Sender:TsesCustomEdit);
  var
  N:Integer;
  S:String;
  begin
  S:=Sender.Name;
  N:=FEditsList.IndexOf(S);
  If N < 0 then
     FEditsList.AddObject(S,Sender);
  end;

  procedure TsesLogger.LogEdits;
  var
  I,N:Integer;
  begin
  If not FActive then Exit;
  N:=FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I := 0 to N do
      TsesCustomEdit(FEditsList.Objects[I]).PostLog;
  end;


  procedure TsesLogger.Log(const Sender:TsesCustomEdit);
  var
  Section,Key:String;
  begin
  If Sender.LoggedType = ltNone then Exit;
  If not FActive then Exit;
  try
With FFileStorage do
begin

  Section := Sender.Name;
  If Sender.LoggedType = ltDated then
     begin
     Key := DateTimeToStr(Now);
     WriteString(Section,Key, Sender.Text);
     end
     else
     begin
     Key := Sender.Text;
     WriteString(Section,Key, '0');
     end;
  UpdateFile;
end; // with FFileStorage
  except
  Exit;
  end;
  end;

  function TsesLogger.GetHistoryList(const Sender:TsesCustomEdit;
   const List:TStrings):Boolean;
  var
  Section,Key,Value:String;
  I,N:Integer;

  begin
  Result := False;
  List.Clear;
  If not FActive then Exit;
  If FStatus <> 0 then Exit;
  If Sender.LoggedType = ltNone then Exit;

  try
  Section := Sender.Name;
  With FFileStorage do
       begin
       If not SectionExists(Section) then Exit;
       If Sender.LoggedType = ltDated then
          begin
          ReadSection(Section,List);
          N := List.Count-1;
          For I:=0 to N do
              begin
              Key:=List[I];
              Value := ReadString(Section,Key,'');
              List[I] := Key + LogFieldSeparator + Value;
              end;
          end // Dated
          else
          begin
          ReadSection(Section,List);
          end;
       end;
  Result := False;     
  except
  Exit;
  end;
  end;

  procedure TsesLogger.InitEdits;
  var
  I,N:Integer;
  begin
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
     begin
     TsesCustomEdit(FEditsList.Objects[I]).SetInitText;
     end;
  finally
  end;
  end;

  procedure TsesLogger.InitEdits(GroupIndex:Integer);
  var
  I,N:Integer;
  begin
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
     begin
     If TsesCustomEdit(FEditsList.Objects[I]).GroupIndex <> GroupIndex then Continue;
     TsesCustomEdit(FEditsList.Objects[I]).SetInitText;
     end;
  finally
  end;
  end;

  procedure TsesLogger.SetEditsState(const Value:TsesState);
  var
  I,N:Integer;
  begin
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
     begin
     If TsesCustomEdit(FEditsList.Objects[I]).ReadOnly = False then
     TsesCustomEdit(FEditsList.Objects[I]).sesState := Value;
     end;
  finally
  end;
  end;

  procedure TsesLogger.SetEditsState(const Value:TsesState; GroupIndex:Integer);
  var
  I,N:Integer;
  begin
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
     begin
     If TsesCustomEdit(FEditsList.Objects[I]).GroupIndex <> GroupIndex then Continue;
     If TsesCustomEdit(FEditsList.Objects[I]).ReadOnly = False then
     TsesCustomEdit(FEditsList.Objects[I]).sesState := Value;
     end;
  finally
  end;
  end;

  function TsesLogger.Complete:Boolean;
  var
  I,N:Integer;
  Res:Boolean;
  begin
  Result := False;
  Res:=True;
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
      If TsesCustomEdit(FEditsList.Objects[I]).GetState <> ssComplete then
         begin
         Res := False;
         end;
  finally
  end;
  Result := Res;
  end;

  function TsesLogger.Complete(GroupIndex:Integer):Boolean;
  var
  I,N:Integer;
  Res:Boolean;
  begin
  Result := False;
  Res:=True;
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
      begin
      If TsesCustomEdit(FEditsList.Objects[I]).GroupIndex <> GroupIndex then Continue;
      If TsesCustomEdit(FEditsList.Objects[I]).GetState <> ssComplete then
         begin
         Res := False;
         end;
      end;
  finally
  end;
  Result := Res;
  end;


  function TsesLogger.EditByName(const EditName:String):TsesEdit;
  var
  I,N:Integer;
  S:String;
  begin
  Result := nil;
  S := ANSIUpperCase(EditName);
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
      If ANSIUpperCase(TsesCustomEdit(FEditsList.Objects[I]).Name) = S then
         begin
         Result := TsesEdit(FEditsList.Objects[I]);
         Exit;
         end;
  finally
  end;
  end;

  function TsesLogger.IsEditCompleted:TsesEdit;
  var
  I,N:Integer;
  begin
  Result := nil;
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
      begin
      If TsesCustomEdit(FEditsList.Objects[I]).GetState <> ssComplete then
         begin
         Result := TsesEdit(FEditsList.Objects[I]);
         Exit;
         end;
      end;
  finally
  end;
  end;

  function TsesLogger.IsEditCompleted(GroupIndex:Integer):TsesEdit;
  var
  I,N:Integer;
  begin
  Result := nil;
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
     begin
     If TsesCustomEdit(FEditsList.Objects[I]).GroupIndex <> GroupIndex then Continue;
      If TsesCustomEdit(FEditsList.Objects[I]).GetState <> ssComplete then
         begin
         Result := TsesEdit(FEditsList.Objects[I]);
         Exit;
         end;
     end;
  finally
  end;
  end;

  function TsesLogger.TrySetFocus:TsesCustomEdit;
  var
  I,N:Integer;
  Edit:TsesCustomEdit;
  begin
  Result := nil;
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
  begin
      If TsesCustomEdit(FEditsList.Objects[I]).sesState <> ssComplete then
         begin
         Edit := TsesCustomEdit(FEditsList.Objects[I]);
         If Edit.CanFocus then
            begin
            Edit.SetFocus;
            Result := Edit;
            Exit;
            end;
         end;
  end;
  finally
  end;
  end;

  function TsesLogger.TrySetFocus(GroupIndex:Integer):TsesCustomEdit;
  var
  I,N:Integer;
  Edit:TsesCustomEdit;
  begin
  Result := nil;
  try
  N:= FEditsList.Count;
  If N = 0 then Exit;
  N:=N-1;
  for I:=0 to N do
  begin
     If TsesCustomEdit(FEditsList.Objects[I]).GroupIndex <> GroupIndex then Continue;
       If TsesCustomEdit(FEditsList.Objects[I]).sesState <> ssComplete then
         begin
         Edit := TsesCustomEdit(FEditsList.Objects[I]);
         If Edit.CanFocus then
            begin
            Edit.SetFocus;
            Result := Edit;
            Exit;
            end;
         end;
  end;
  finally
  end;
  end;

  procedure TsesLogger.SaveControlRect(const Sender:TObject);
  var
  Section:String;
  Control:TControl;
  begin
  If not FActive then Exit;
  try
  Control:=TControl(Sender);
  Section := Control.Name;
  With FFileStorage,Control do
       begin
       WriteInteger(Section,'Left', Left);
       WriteInteger(Section,'Top', Top);
       WriteInteger(Section,'Width', Width);
       WriteInteger(Section,'Height', Height);
       UpdateFile;
       end;
  except
  Exit;
  end;
  end;

  function  TsesLogger.ReadControlRect(const Sender:TObject):Boolean;
  var
  Section:String;
  Control:TControl;
  begin
  Result := False;
  If not FActive then Exit;
  try
  Control:=TControl(Sender);
  Section := Control.Name;

  With FFileStorage,Control do
       begin
       If not SectionExists(Section) then Exit;
       Left := ReadInteger(Section,'Left', Left);
       Top := ReadInteger(Section,'Top', Top);
       Width := ReadInteger(Section,'Width', Width);
       Height := ReadInteger(Section,'Height', Height);
       end;
  Result := True;
  except
  Exit;
  end;
  end;

  procedure TsesLogger.Loaded;
  begin
  inherited;
  if FActive then Activate(True);
  If FInitTruncation then TruncateLogs;
  end;
  
  function TsesLogger.Activate(Start:Boolean):Word;
  var
  Dir:String;
  S,S1:String;
  I:Integer;
  begin
  Result := 0;
If csDesigning in ComponentState then Exit;
  FActive := False;
  If not Start then
     begin
     DeActivate;
     Exit;
     end;


  If FFileStorage <> nil   then
     FFileStorage.Free;
  S:= Trim(FFileName);
  If S = '' then
       begin
       FStatus := 2; // empty file name
//       Hint :=sesLogger_2+':'+#13+#10+S;
//       Invalidate;
       Exit;
       end;
  Dir := ExtractFileDir(FFileName);
  If Dir = '' then
      begin
      SetRuntimeDirDefault;
      S1:=Trim(FDirDefault);
      If S1 = '' then
         begin
         FStatus:=3; // empty dir name in user defined file name
//         Hint :=sesLogger_3+':'+#13+#10+S1;
//         Invalidate;
         Exit;
         end;
     I:=Length(S1);
     If S1[I] <> '\' then
         begin
         FStatus:=7; // the last symbol is no slash
//         Hint :=sesLogger_7+':'+#13+#10+S1;
//         Invalidate;
         Exit;
         end;

      Case FDirDefaultType Of
      dfUser:
             begin
             If not DirectoryExists(FDirDefault) then
                begin
                FStatus:=3; // empty dir name in user defined file name
//                Hint :=sesLogger_3+':'+#13+#10+FDirDefault;;
//                Invalidate;
                Exit;
                end;
             FFileNameStorage := FDirDefault + S;
             end;
      dfWinDir,dfExecute:
             begin
             If not DirectoryExists(FDirDefault) then
                begin
                FStatus:=4; // Win or Execute dir does not exist
//                Hint :=sesLogger_4+':'+#13+#10+FDirDefault;
//                Invalidate;
                Exit;
                end;
             FFileNameStorage := FDirDefault + S;
             end;

      end; // case
      end
      else
      begin
      If not DirectoryExists(Dir) then
                begin
                FStatus:=5; // file name dir does not exist
//                Hint :=sesLogger_5+':'+#13+#10+Dir;
//                Invalidate;
                Exit;
                end;
      FFileNameStorage := S;
      end;
  // at last, begin to create Storage
  try
  FFileStorage:= TLoggerFileStorage.Create(FFileNameStorage);
  FStatus:=0; // ready!
  FActive := True;
//  Hint:=sesLogger_0+':'+#13+#10+FFileNameStorage;
  except
  FStatus:=6; // there is not enough memory
//  Hint :=sesLogger_6;
//  Invalidate;
  Exit;
  end;
  end;


  procedure TsesLogger.DeActivate;
  begin
  If FFileStorage <> nil   then
     FFileStorage.Free;
  FStatus:=1;
//  Hint:=sesLogger_1;
//  Invalidate;
  end;

  {
  procedure TsesLogger.Paint;
  begin
  Case FStatus Of
  0:
    begin
     If not FIconLoaded then
        begin
        FIconLoaded:=True;
        FIcon.LoadFromResourceName(HInstance, 'LOGGEROPEN');
        end;
    end;
  1,2,3,4,5,6,7:
     begin
     If FIconLoaded then
        begin
        FIconLoaded:=False;
        FIcon.LoadFromResourceName(HInstance, 'LOGGEROF');
        end;
     end;
  end; // Case Of FStatus
  Canvas.Draw(0,0, FIcon);
  end;
  }

  procedure TsesLogger.SetFileName(const Value:TFileName);
  begin
  If Value <> FFileName then
     begin
     FFileName := Trim(Value);
     end;
  end;

  procedure TsesLogger.SetDirDefaultType(const Value:TDirDefaultType);
  begin
  If FDirDefaultType <> Value then
     begin
     FDirDefaultType := Value;
     end;
  end;

  procedure TsesLogger.SetRuntimeDirDefault;
  var
  P:PChar;
  aP:array[0..255] of Char;
  L:Integer;
  begin
     Case FDirDefaultType of
     dfUser: FDirDefault:= Trim(FDirDefault);
     dfWinDir:
              begin
              P:=aP;
              GetWindowsDirectory(P,SizeOf(aP));
              FDirDefault := String(P);
              L:=Length(FDirDefault);
              If FDirDefault[L] <> '\' then
              FDirDefault:= FDirDefault+'\';
              end;
     dfExecute:
              begin
              FDirDefault:= ExtractFileDir(Application.ExeName);
              L:=Length(FDirDefault);
              If FDirDefault[L] <> '\' then
              FDirDefault:= FDirDefault+'\';
              end;

     end; // case Of Type
    end;

procedure TsesLogger.SetActive(const Value:Boolean);
begin
 if FActive <> Value then
    begin
    FActive := Value;
    end;
end;

procedure TsesLogger.SetDirDefault(const Value:TDirectory);
begin
If FDirDefault <> Value then
   begin
   FDirDefault := Value;
   FDirDefaultType := dfUser;
   end;
end;

function TsesLogger.GetEdit(Index:Integer):TsesCustomEdit;
begin
Result := nil;
If (Index < 0) Or (Index > FEditsList.Count-1) then Exit;
Result := TsesCustomEdit(FEditsList.Objects[Index]);
end;

function TsesLogger.GetEditsCount:Integer;
begin
Result := FEditsList.Count;
end;

{sesData data}
   procedure TsesDataGroup.SetRanged(const Value:Boolean);
   begin
   If Editor.Ranged <> Value then
      Editor.Ranged := Value;
   end;
   procedure TsesDataGroup.SetRangeMinOrFalse(const Value:String);
   begin
   If Editor.RangeMinOrFalse <> Value then
      Editor.RangeMinOrFalse := Value;
   end;

   procedure TsesDataGroup.SetRangeMaxOrTrue(const Value:String);
   begin
   If Editor.RangeMaxOrTrue <> Value then
      Editor.RangeMaxOrTrue := Value;
   end;

   function TsesDataGroup.GetRangeMinOrFalse:String;
   begin
   Result := Editor.RangeMinOrFalse;
   end;
   function TsesDataGroup.GetRangeMaxOrTrue:String;
   begin
   Result := Editor.RangeMaxOrTrue;
   end;

   function TsesDataGroup.GetRanged:Boolean;
   begin
   Result := Editor.Ranged;
   end;



function GetsesButtonGlyph(Kind: TDefaultGlyph): TBitmap;
var
I:Cardinal;
begin
I:=Cardinal(Kind);
If I > High(sesButtonResNames) then
  begin
  Result := nil;
  Exit;
  end;

If Kind = dgNone then
  begin
  Result := nil;
  Exit;
  end;
  if sesButtonGlyphs[I] = nil then
  begin
    sesButtonGlyphs[I] := TBitmap.Create;
    sesButtonGlyphs[I].LoadFromResourceName(HInstance, sesButtonResNames[I]);
  end;
  Result := sesButtonGlyphs[I];
end;




{TsesEditGroup}
   procedure TsesEditGroup.SetColor(const Value:TColor);
   begin
   If not Assigned(Edit) then Exit;
   If Edit.Color <> Value then
         begin
         Edit.Color := Value;
         Edit.Parent.Repaint;
         end;
   end;

   procedure TsesEditGroup.SetFont(const Value:TFont);
   begin
   If not Assigned(Edit) then Exit;
   If Edit.Font <> Value then
         begin
         Edit.Font := Value;
         AdjustPosition(Edit.Parent);

         end;
   end;

   procedure TsesEditGroup.SetParentFont(const Value:Boolean);
   begin
   If not Assigned(Edit) then Exit;
   If Edit.ParentFont <> Value then
         begin
         Edit.ParentFont := Value;
         Edit.Parent.Repaint;
         end;
   end;


   procedure TsesEditGroup.SetWidth(const Value:Integer);
   begin
   If not Assigned(Edit) then Exit;
   If Edit.Width <> Value then
         begin
         Edit.Width := Value;
         AdjustPosition(Edit.Parent);
         Edit.Parent.Repaint;
         end;
   end;

   function TsesEditGroup.GetColor:TColor;
   begin
   Result := clBlack;
   If not Assigned(Edit) then Exit;
   Result := Edit.Color;
   end;

   function TsesEditGroup.GetFont:TFont;
   begin
   Result := nil;
   If not Assigned(Edit) then Exit;
   If csLoading in Edit.ComponentState then Exit;
   Result := Edit.Font;
   end;

   function TsesEditGroup.GetParentFont:Boolean;
   begin
   Result := False;
   If not Assigned(Edit) then Exit;
   Result := Edit.ParentFont;;
   end;


   function TsesEditGroup.GetWidth:Integer;
   begin
   Result := 0;
   If not Assigned(Edit) then Exit;
   Result := Edit.Width;
   end;


{TsesCaption}
   procedure TsesCaption.SetAlignment(const Value:TAlignment);
   begin
   If not Assigned(Frame) then Exit;
        If Frame.Alignment <> Value then
           begin
           Frame.Alignment := Value;
           AdjustPosition(Frame);
           Frame.Invalidate;
           end;
   end;
   procedure TsesCaption.SetLayout(const Value:TsesLayout);
   var
   ParentEdit:TsesCustomEdit;
   begin
   If not Assigned(Frame) then Exit;
   If FLayout <> Value then
           begin
           ParentEdit:=TsesCustomEdit(Frame);
           ParentEdit.AdjustClientSize(Value);
           AdjustPosition(Frame);
           Frame.Invalidate;
           Exit;
           end;


     end;



   procedure TsesCaption.SetVisible(const Value:Boolean);
   begin
   If not Assigned(Frame) then Exit;
        If FVisible <> Value then
           begin
           FVisible := Value;
           AdjustActualSizes(Frame);
           AdjustPosition(Frame);
           Frame.Invalidate;
           end;
   end;

   procedure TsesCaption.SetVertAlignment(const Value:TVertAlignment);
   begin
   If not Assigned(Frame) then Exit;
        If FVertAlignment <> Value then
           begin
           FVertAlignment := Value;
           Frame.Invalidate;
           end;
   end;
   procedure TsesCaption.SetWordWrap(const Value:Boolean);
   begin
   If not Assigned(Frame) then Exit;
        If FWordWrap <> Value then
           begin
           FWordWrap := Value;
           AdjustPosition(Frame);
           Frame.Invalidate;
           end;
   end;

   function TsesCaption.GetAlignment:TAlignment;
   begin
   Result := taLeftJustify;
   If not Assigned(Frame) then Exit;
   Result := Frame.Alignment;
   end;

   function TsesCaption.GetVisible:Boolean;
   begin
   Result := FVisible;
   end;

   function TsesCaption.GetVertAlignment:TVertAlignment;
   begin
   Result := FVertAlignment;
   end;

   function TsesCaption.GetWordWrap:Boolean;
   begin
   Result := FWordWrap;
   end;

{TsesFrame}
   procedure TsesFrame.SetInteriorIndent(const Value:Word);
   begin
   If not Assigned(Frame) then Exit;
        If FInteriorIndent <> Value then
           begin
           FInteriorIndent := Value;
           AdjustPosition(Frame);
           Frame.Invalidate;
           end;
   end;

   procedure TsesFrame.SetBevelInner(const Value:TPanelBevel);
   begin
   If not Assigned(Frame) then Exit;
        If Frame.BevelInner <> Value then
           begin
           Frame.BevelInner := Value;
           AdjustPosition(Frame);
           Frame.Invalidate;
           end;
   end;

   procedure TsesFrame.SetBevelOuter(const Value:TPanelBevel);
   begin
   If not Assigned(Frame) then Exit;
        If Frame.BevelOuter <> Value then
           begin
           Frame.BevelOuter := Value;
           AdjustPosition(Frame);
           Frame.Invalidate;
           end;
   end;

   procedure TsesFrame.SetBevelWidth(const Value:TBevelWidth);
   begin
   If not Assigned(Frame) then Exit;
        If Frame.BevelWidth <> Value then
           begin
           Frame.BevelWidth := Value;
           AdjustPosition(Frame);
           Frame.Invalidate;
           end;
   end;

   procedure TsesFrame.SetBorderStyle(const Value:TBorderStyle);
   begin
   If not Assigned(Frame) then Exit;
        If Frame.BorderStyle <> Value then
           begin
           Frame.BorderStyle := Value;
           AdjustActualSizes(Frame);
           AdjustPosition(Frame);
           Frame.Invalidate;
           end;
   end;

   procedure TsesFrame.SetBorderWidth(const Value:TBorderWidth);
   begin
   If not Assigned(Frame) then Exit;
        If Frame.BorderWidth <> Value then
           begin
           Frame.BorderWidth := Value;
           AdjustPosition(Frame);
           Frame.Invalidate;
           end;
   end;

   function TsesFrame.GetInteriorIndent:Word;
   begin
   Result := FInteriorIndent;
   end;


   function TsesFrame.GetBevelInner:TPanelBevel;
   begin
   Result := bvNone;
   If not Assigned(Frame) then Exit;
   Result := Frame.BevelInner;
   end;

   function TsesFrame.GetBevelOuter:TPanelBevel;
   begin
   Result := bvNone;
   If not Assigned(Frame) then Exit;
   Result := Frame.BevelOuter;
   end;

   function TsesFrame.GetBevelWidth:TBevelWidth;
   begin
   Result := 1;
   If not Assigned(Frame) then Exit;
   Result := Frame.BevelWidth;
   end;

   function TsesFrame.GetBorderStyle:TBorderStyle;
   begin
   Result := bsNone;
   If not Assigned(Frame) then Exit;
   Result := Frame.BorderStyle;
   end;

   function TsesFrame.GetBorderWidth:TBorderWidth;
   begin
   Result := 0;
   If not Assigned(Frame) then Exit;
   Result := Frame.BorderWidth;
   end;

{TsesButton}
procedure TsesButton.SetCaption(const Value:String);
begin
If not Assigned(Edit) then Exit;
If Edit.PickButton.Caption <> Value then
   begin
   Edit.PickButton.Caption := Value;
   Edit.PickButton.Invalidate;
   end;

end;
procedure TsesButton.SetEnabled(const Value:Boolean);
begin
If not Assigned(Edit) then Exit;
If Edit.PickButton.Enabled <> Value then
   begin
   Edit.PickButton.Enabled := Value;
   Edit.PickButton.Invalidate;
   end;

end;

procedure TsesButton.SetDefaultGlyph(const Value:TDefaultGlyph);
begin
If not Assigned(Edit) then Exit;
If FDefaultGlyph <> Value then
   begin
   FDefaultGlyph := Value;
 //  Case FDefaultGlyph of
//   dgUser :;
//   else
   Edit.PickButton.Glyph:= GetsesButtonGlyph(Value);
  // end; // case FDefaultGlyph of
   Edit.PickButton.Invalidate;
   end;

end;

procedure TsesButton.SetEditGap(const Value:Integer);
begin
If not Assigned(Edit) then Exit;
If EditorGap <> Value then
   begin
   EditorGap := Value;
   AdjustPosition(Edit);
   Edit.Invalidate;
   end;

end;


procedure TsesButton.SetGlyph(const Value:TBitMap);
begin
If not Assigned(Edit) then Exit;
   Edit.PickButton.Glyph := Value;
   FDefaultGlyph := dgNone;
   Edit.Invalidate;

end;

procedure TsesButton.SetWidth(const Value:Integer);
begin
If not Assigned(Edit) then Exit;
If not GetVisible then Exit;

If Edit.PickButton.Width <> Value then
   begin
   Edit.PickButton.Width := Value;
   AdjustPosition(Edit);
   Edit.Invalidate;
   end;

end;

procedure TsesButton.SetFlat(const Value:Boolean);
begin
If not Assigned(Edit) then Exit;
If Edit.PickButton.Flat <> Value then
   begin
   Edit.PickButton.Flat := Value;
   Edit.PopUpMenuButton.Flat:=Value;
   Edit.Invalidate;
   end;

end;

procedure TsesButton.SetLayout(const Value:TButtonLayout);
begin
If not Assigned(Edit) then Exit;
If Edit.PickButton.Layout <> Value then
   begin
   Edit.PickButton.Layout := Value;
   Edit.PickButton.Invalidate;
   end;

end;


procedure TsesButton.SetVisible(const Value:Boolean);
begin
If not Assigned(Edit) then Exit;
If Value <> FButtonVisible then
   begin
   FButtonVisible := Value;
   Edit.PopUpMenuButton.Visible := Edit.PopUpMenuEnabled;
   If FButtonVisible then
      begin
      If Edit.PopUpMenuButton.Visible then
         Edit.PopUpMenuButton.Width := 12
         else
         Edit.PopUpMenuButton.Width := 0;
      If FButtonWidth > 0 then
         begin
         Edit.PickButton.Width := FButtonWidth;
         AdjustPosition(Edit);
         Edit.Invalidate;
         Exit;
         end
         else
         begin
         If Edit.PickButton.Width = 0 then
            Edit.PickButton.Width := 20;
         AdjustPosition(Edit);
         Edit.Invalidate;
         Exit;
         end;
      end
      else
      begin
      If Edit.PickButton.Width = 0 then
          FButtonWidth := 20
          else
          FButtonWidth := Edit.PickButton.Width;
      Edit.PickButton.Width := 0;
      Edit.PopUpMenuButton.Width := 0;
      end;
   AdjustPosition(Edit);
   Edit.Invalidate;
   end;

end;

procedure TsesButton.SetNumGlyphs(const Value:TNumGlyphs);
begin
If not Assigned(Edit) then Exit;
If Edit.PickButton.NumGlyphs <> Value then
   begin
   Edit.PickButton.NumGlyphs := Value;
   Edit.PickButton.Invalidate;
   end;

end;

function TsesButton.GetCaption:String;
begin
Result:='';
If not Assigned(Edit) then Exit;
Result := Edit.PickButton.Caption;
end;

function TsesButton.GetEnabled:Boolean;
begin
Result:=False;
If not Assigned(Edit) then Exit;
Result := Edit.PickButton.Enabled;
end;

function TsesButton.GetFlat:Boolean;
begin
Result := False;
If not Assigned(Edit) then Exit;
Result := Edit.PickButton.Flat;
end;

function TsesButton.GetVisible:Boolean;
begin
Result := False;
If not Assigned(Edit) then Exit;
Result := FButtonVisible;
end;

function TsesButton.GetGlyph:TBitMap;
begin
Result := nil;
If not Assigned(Edit) then Exit;
Result := Edit.PickButton.Glyph;
end;


function TsesButton.GetNumGlyphs:TNumGlyphs;
begin
Result := 1;
If not Assigned(Edit) then Exit;
Result := Edit.PickButton.NumGlyphs;
end;

function TsesButton.GetLayout:TButtonLayout;
begin
Result := blGlyphLeft;
If not Assigned(Edit) then Exit;
Result := Edit.PickButton.Layout;
end;

function TsesButton.GetWidth:Integer;
begin
Result := 0;
If not Assigned(Edit) then Exit;
If not FButtonVisible then
   Result := FButtonWidth
   else
   Result := Edit.PickButton.Width;
end;




{TsesCustomEdit}

constructor TsesCustomEdit.Create(AOwner:TComponent);
begin
inherited Create(AOwner);

FDecimalKeyAsSeparator := True;

FCancel := False;

FErrorDialog := edShow;
FErrorExit := eeIfInit;
FErrorSound := esNone;


FPickList:=TStringList.Create;


LastValidValue := abracadabra;
FEnabled:=True;
// internal only flags setting...
FForceAdjust:=False;
FResizing := False;

//Anchors  :=[];

Width := 145;
Height := 44;
BorderWidth := 1;
Alignment := taLeftJustify;

FLoggerLimit := 100;

FsesFrame := TsesFrame.Create;
FsesFrame.Frame := TsesCustomPanel(Self);
//

PickButton:=TSpeedButton.Create(Self);
With PickButton do
  begin
  Parent := Self;
  Name := 'PickButton';
  Height := 20;
  Width := 20;
  Margin := -1;
  Spacing := 1;
  OnClick := PickButtonClick;
  end;
FButton := TsesButton.Create;
FButton.Edit := TsesCustomEdit(Self);
FButton.EditorGap := 1;
FButton.FButtonVisible := True;
FButton.FButtonWidth := 0;

//PickButton.PopupMenu := sesAdviseForm.sesPopupMenu;
PopUpMenuButton:=TSpeedButton.Create(Self);
With PopUpMenuButton do
  begin
  Parent := Self;
  Name := 'PopUpMenuButton';
  Height := 12;
  Width := 0;
  Margin := -1;
  Spacing := 0;
  Glyph.LoadFromResourceName(HInstance, 'POPUPMENU');
  Visible := False;
  OnClick := PopUpMenuClick;
  end;

//Editor:=TsesCustomEditor.Create(Self);
Editor:=TsesCustomEditor.Create(Self);
With Editor do
  begin
  Parent := Self;
  Name := 'Editor';
  EditDataType := cdtString;
  end;

FsesEditor :=TsesEditGroup.Create;
FsesEditor.Edit := Editor;//TsesCustomEditor(Editor);

FsesFrame.FInteriorIndent := 1;

FsesCaption:=TsesCaption.Create;
FsesCaption.Frame := TsesCustomPanel(Self);

FsesCaption.FVisible := True;
FsesCaption.FAlignment := taLeftJustify;
FsesCaption.FWordWrap := False;
FsesCaption.FVertAlignment := dtCenter;
FsesCaption.FLayout := clTop;

FsesData := TsesDataGroup.Create;
FsesData.Editor := Editor;

Editor.TabStop := True;
TabStop := False;
FTabStopEditor := True;
FStateView := svShowFrame;
FsesState := ssInit;
FPickListKind := plNone;
FRequired := False;
FGroupIndex := 0;
end;

function TsesCustomEdit.GetPickListCaption(const Full:Boolean):String;
begin
Result := sesAdviseForm.GetPickListCaption(TsesCustomEdit(Self),Full);
end;


procedure TsesCustomEdit.WMSize(var Message:TWMSize);
begin
inherited;
AdjustPositions;
end;

{
procedure TsesCustomEdit.TABSTOPCHANGED(var Message:TMessage);
begin
//If csLoading in ComponentState then Exit;
//If not Assigned(Editor) then Exit;
//If Editor.TabStop <> Self.TabStop then
   Editor.TabStop := Self.TabStop;
end;
}
procedure TsesCustomEdit.PassFocus;
begin
If Assigned(FTabNext) then
   begin
   If FTabNext.CanFocus then
      FTabNext.SetFocus;
   end;

end;

function TsesCustomEdit.DoCheckText:Boolean;
var
Valid:Boolean;
begin
Result := True;
If Assigned(FOnCheckText) then
   begin
   Valid := True;
   FOnCheckText(Self,Valid);
   Result := Valid;
   end;

end;

function TsesCustomEdit.DoUserErrorProcess:TModalResult;
var
fModalResult:TModalResult;
begin
Result := mrRetry;
If Assigned(FOnUserErrorProcess) then
   begin
   fModalResult := mrRetry;
   FOnUserErrorProcess(Self,fModalResult);
   Result := fModalResult;
   end;

end;

procedure TsesCustomEdit.DoExit;
begin
If Assigned(OnExit) then OnExit(Self);
PassFocus;
end;

procedure TsesCustomEdit.Paint;
const
  Alignments: array[TAlignment] of Longint = (DT_LEFT, DT_RIGHT, DT_CENTER);
  VertAlignments: array[TVertAlignment] of Longint = (DT_TOP, DT_VCENTER, DT_BOTTOM);

var
  Rect: TRect;
  TopColor, BottomColor: TColor;
  FrameTopColor, FrameBottomColor:TColor;
  Flags: Longint;
  LineWidth:Integer;
  procedure AdjustColors(Bevel: TPanelBevel);
  begin
    TopColor := clBtnHighlight;
    if Bevel = bvLowered then TopColor := clBtnShadow;
    BottomColor := clBtnShadow;
    if Bevel = bvLowered then BottomColor := clBtnHighlight;
  end;


  procedure DrawState;
    begin
    If FStateView = svNone then Exit;
    If not Assigned(Editor) then Exit;
    // draw the Edit state box
    LineWidth := 1;
    Rect.Left := Editor.Left-LineWidth;//Self.Left;
    Rect.Top := Editor.Top-LineWidth;//Self.Top;
    Rect.Right := Rect.Left + Editor.Width + LineWidth*2;
    Rect.Bottom := Rect.Top + Editor.Height + LineWidth*2;
    If FReadOnly then
    begin
            FrameTopColor    := sesState_ReadOnly_TopColor;
            FrameBottomColor := sesState_ReadOnly_BottomColor;
    end
    else
    begin
    Case sesState Of
    ssInit:
            begin
            FrameTopColor    := sesState_Init_TopColor; // HelperBK9;//clBtnFace;
            FrameBottomColor := sesState_Init_BottomColor ;//HelperBK10;//clBtnFace;
            end;
    ssEdit:
            begin
            FrameTopColor    := sesState_Edit_TopColor;
            FrameBottomColor := sesState_Edit_BottomColor;
            end;

    ssError:
            begin
            FrameTopColor    := sesState_Error_TopColor;
            FrameBottomColor := sesState_Error_BottomColor;
            end;

    ssComplete:
            begin
            FrameTopColor    := sesState_Complete_TopColor;
            FrameBottomColor := sesState_Complete_BottomColor;
            end;

    end; // Case sesState
    end;
    Frame3D(Canvas, Rect, FrameTopColor, FrameBottomColor, LineWidth);

    {
    Canvas.Pen.Color:=clBlack;
    Canvas.Rectangle(Rect);
    //FillRect(Rect);
    //Ellipse(Rect);
    }
    end;
begin
  Rect := GetClientRect;
  if BevelOuter <> bvNone then
  begin
    AdjustColors(BevelOuter);
    Frame3D(Canvas, Rect, TopColor, BottomColor, BevelWidth);
  end;
  Frame3D(Canvas, Rect, Color, Color, BorderWidth);
  if BevelInner <> bvNone then
  begin
    AdjustColors(BevelInner);
    Frame3D(Canvas, Rect, TopColor, BottomColor, BevelWidth);
  end;
  with Canvas do
  begin
    Brush.Color := Color;
    FillRect(Rect);
    If not sesCaption.Visible then
       begin
       DrawState;
       Exit;
       end;

    Brush.Style := bsClear;
    Font := Self.Font;
    If sesCaption.WordWrap then
       begin
       // multiline text will be allowed, but no vert alignments will be produced
       // there is only hor text alignment...
       Flags := DT_WORDBREAK or Alignments[sesCaption.Alignment];
       end
       else
       begin
       Flags := DT_SINGLELINE or VertAlignments[sesCaption.VertAlignment] or Alignments[sesCaption.Alignment];
       end;

    Rect := GetControlsRect;
    Case sesCaption.FLayout Of
    clTop:
            begin
//            Rect.Right := Editor.Left;
            Dec(Rect.Bottom, Editor.Height);
            Rect.Left := Rect.Left + 1;
            end;
    clLeft:
            begin
            Rect.Right := Editor.Left-1;
//            Dec(Rect.Right, sesButton.Width);
            //Dec(Rect.Right, Editor.Width);
            Rect.Left := Rect.Left + 1;
            end;

     end; // Case Of Layout
    //Flags := DT_EXPANDTABS or DT_VCENTER or Alignments[Alignment];
    Flags := DrawTextBiDiModeFlags(Flags);
    DrawText(Handle, PChar(Caption), -1, Rect, Flags);

  end;
  DrawState;
end;

procedure TsesCustomEdit.sesSayMessage(ErrorKind:TEditErrorType);
const
sysBeep = $FFFFFFFF;
  var
  P:PChar;
  aP:array[0..255] of Char;
  S:String;
begin
Case  FErrorSound Of
esNone: Exit;
esBeep: Windows.Beep(2000,1500);//Windows.MessageBeep(sysBeep); // system computer speaker beep
esPlayWave:
   begin
   // we wait to find sound files in Windows dir
   P:=aP;
   GetWindowsDirectory(P,SizeOf(aP));
   S := String(P);
   S:=S+'\';
   Case ErrorKind of
   ceChar: sesAdviseForm.PlaySound(S+'Error_Char.wav');
   ceNumber: sesAdviseForm.PlaySound(S+'Error_Number.wav');
   ceBoolean: sesAdviseForm.PlaySound(S+'Error_Boolean.wav');
   ceRange: sesAdviseForm.PlaySound(S+'Error_Range.wav');
   cePickList: sesAdviseForm.PlaySound(S+'Error_PickList.wav');
   ceDate: sesAdviseForm.PlaySound(S+'Error_Date.wav');
   ceTime: sesAdviseForm.PlaySound(S+'Error_Time.wav');
   end; // Case Of ErrorKind
   end; // PlayWave
end; // Case FErrorSound Of

end;

function TsesCustomEdit.GetControlsRect:TRect;
var
Indent:Word;
WinBorderWidth:Integer;
begin
   // estimate the current InsideRect:

   WinBorderWidth := 4;
   Result.Left := BorderWidth;
   Result.Top := BorderWidth;
   Result.Right := Width - BorderWidth;
   Result.Bottom := Height - BorderWidth;
  if BevelOuter <> bvNone then
     begin
     Inc(Result.Left,BevelWidth);
     Inc(Result.Top,BevelWidth);
     Dec(Result.Right,BevelWidth);
     Dec(Result.Bottom,BevelWidth);
     end;

  if BevelInner <> bvNone then
     begin
     Inc(Result.Left,BevelWidth);
     Inc(Result.Top,BevelWidth);
     Dec(Result.Right,BevelWidth);
     Dec(Result.Bottom,BevelWidth);
     end;

  if BorderStyle <> bsNone then
     begin
     Inc(Result.Left,WinBorderWidth);
     Inc(Result.Top,WinBorderWidth);
     Dec(Result.Right,WinBorderWidth);
     Dec(Result.Bottom,WinBorderWidth);
     end;
Indent := sesFrame.InteriorIndent;
Inc(Result.Left,Indent);
Inc(Result.Top,Indent);
Dec(Result.Right,Indent);
Dec(Result.Bottom,Indent);

end;

function TsesCustomEdit.GetTotalBorderWidth:Integer;
begin
   // estimate the current InsideRect:
   Result:= BorderWidth + sesFrame.FInteriorIndent;
  if BevelOuter <> bvNone then
     Result := Result + BevelWidth;

  if BevelInner <> bvNone then
     Result := Result + BevelWidth;

  if BorderStyle <> bsNone then
     Result := Result + 4;


end;

procedure TsesCustomEdit.AdjustClientSize(const aLayout:TsesLayout);
var
Rect:TRect;
begin
If csFixups in ComponentState then Exit;
//If csReading in ComponentState then Exit;

If csLoading in ComponentState then
   begin
   sesCaption.FLayout := aLayout;
   Exit;
   end;
   sesCaption.FLayout := aLayout;
   If (aLayout = clLeft) and (not sesCaption.WordWrap) then
      begin
      Rect.Left := Self.Left;
      Rect.Top := Self.Top;
      Rect.Right := Self.Left + 1;
      Rect.Bottom := Self.Top + 1;
      inherited UpdateBoundsRect(Rect);
      end;
   AdjustActualSize;
   Invalidate;
end;

function TsesCustomEdit.GetState:TsesState;
var
Res:TEditErrorType;
begin
If ReadOnly then
   Result := ssComplete
   else
   begin
   Res := Editor.CheckText;
   Case Res Of
   ceOK: Result := ssComplete;
   else
   If FsesState = ssInit then
      Result := ssInit
      else
      Result := ssError;
   end; // Case
   end;
Repaint;   
end;
procedure TsesCustomEdit.AdjustPositions;
var
dR:Integer;
CaptionRect:TRect;
W,W1,CaptionWidth:Integer;
ActualWidth:Integer;
begin
If (not FForceAdjust) and  ((FResizing) or (csFixups in ComponentState))  then Exit;

//If csLoading in ComponentState then Exit;
// If csCreating	in  ControlState then Exit;
// If csReadingState in  ControlState then Exit;


InsideRect := GetControlsRect;
CaptionRect := GetCaptionRect;
CaptionWidth := CaptionRect.Right - CaptionRect.Left;

  // we think, that Edit control has resized it's Height already due to Autosize property...
  PickButton.Height := Editor.Height;
  PopUpMenuButton.Height:=Editor.Height;

    Case sesCaption.FLayout Of
    clTop:
            begin
            If sesButton.Visible then
               W:= Editor.Width + PopUpMenuButton.Width+PickButton.Width + sesButton.EditorGap
               else
               W := Editor.Width;
            ActualWidth := Max(CaptionWidth,W);
            ActualWidth := ActualWidth + GetTotalBorderWidth * 2;

            If (Self.Width < ActualWidth) then
               AdjustActualSize;

            // left sides:
            Editor.Left := InsideRect.Left;
            // Edit's width
            dR := InsideRect.Right - InsideRect.Left;
            If  sesButton.Visible then
                begin
                dR := dR - sesButton.Width;
                dR := dR -  sesButton.EditorGap;
                If PopUpMenuButton.Visible then
                dR := dR - PopUpMenuButton.Width;
                end;
//            If FsesAutosize then
            Editor.Width := dR;

            If  sesButton.Visible then
                begin
                PickButton.Left:=Editor.Left + Editor.Width + sesButton.EditorGap;
                If PopUpMenuButton.Visible then
                   PopUpMenuButton.Left :=PickButton.Left + PickButton.Width - 1
                   else
                   PopUpMenuButton.Left:=InsideRect.Right;
                end
                else
                begin
                PopUpMenuButton.Left := InsideRect.Right;
                PickButton.Left:= InsideRect.Right;// + 1;
                end;
            // top sides:
            If  FForceAdjust then
                begin
                Editor.Top := InsideRect.Bottom - Editor.Height - 2;
                PickButton.Height := PickButton.Height + 2;
                end
                else
                Editor.Top := InsideRect.Bottom - Editor.Height; // - 1;
            PickButton.Top := Editor.Top;
            PopUpMenuButton.Top := Editor.Top;
            end;
  clLeft:
            begin
               If sesButton.Visible then
                  begin
                  If PopUpMenuButton.Visible then
                     W:= PickButton.Width + PopUpMenuButton.Width + sesButton.EditorGap
                     else
                     W := PickButton.Width + sesButton.EditorGap;
                 end
                 else
                 W := 0;
               W:=  CaptionWidth + W  + GetTotalBorderWidth * 2;
               W1 := Self.Width - W;
               If W1 < 25 then
                  W1 := 25;
            Editor.Width := W1;


            If  sesButton.Visible then
                begin
                If PopUpMenuButton.Visible then
                   begin
                   PickButton.Left :=InsideRect.Right -
                   PopUpMenuButton.Width - PickButton.Width - sesButton.EditorGap;
                   PopUpMenuButton.Left :=PickButton.Left + PickButton.Width - 1;
                   end
                   else
                   begin
                   PickButton.Left :=InsideRect.Right -
                   PickButton.Width - sesButton.EditorGap;
                   PopUpMenuButton.Left:=InsideRect.Right;
                   end;
                Editor.Left := PickButton.Left - Editor.Width - sesButton.EditorGap;
                end
                else
                begin
                PopUpMenuButton.Left := InsideRect.Right;
                PickButton.Left:= InsideRect.Right;// + 1;
                Editor.Left := InsideRect.Right - Editor.Width-1;
                end;

              // top sides:
            Editor.Top := InsideRect.Bottom - Editor.Height;// - 1;
            PickButton.Top := Editor.Top;
            PopUpMenuButton.Top := Editor.Top;

            end;
    end; // Case Of Layout
//  PickButton.Left:= InsideRect.Right - PickButton.Width -1;// - sesButton.EditGap;

end;

procedure TsesCustomEdit.CMEnter(var Message: TCMGotFocus);
begin
inherited;
//Editor.SetFocus;
end;

function TsesCustomEdit.GetCaptionRect:TRect;
const
  Alignments: array[TAlignment] of Longint = (DT_LEFT, DT_RIGHT, DT_CENTER);
  VertAlignments: array[TVertAlignment] of Longint = (DT_TOP, DT_VCENTER, DT_BOTTOM);

var
  DC: HDC;
  Rect:TRect;
  Flags: Longint;
begin
  If not sesCaption.Visible then
     begin
     Result.Left :=0;
     Result.Top :=0;
     Result.Right :=0;
     Result.Bottom :=0;
     Exit;
     end;
  // set the Caption rendering parameters:
  If sesCaption.FWordWrap then
       begin
       // multiline text will be allowed, but no vert alignments will be produced
       // there is only hor text alignment...
       Flags := DT_WORDBREAK or Alignments[sesCaption.FAlignment];
       end
       else
       begin
       Flags := DT_SINGLELINE or VertAlignments[sesCaption.FVertAlignment] or
                Alignments[sesCaption.FAlignment];
       end;

  Flags := Flags or DT_CALCRECT;
  Flags := DrawTextBiDiModeFlags(Flags);
  // Caption always lies on (Left,Top) corner of the Client area of Parent(TCustomPanel)
  Rect := GetControlsRect; // Left and Top has been determined...
  // Now we need to estimate Right and Height...
  Case sesCaption.FLayout Of
  clTop:
        begin
        Dec(Rect.Bottom, Editor.Height);
        end;
  clLeft:
        begin
        If PickButton.Visible then
           begin
           Dec(Rect.Right, Editor.Width + PickButton.Width+sesButton.EditorGap);
           If PopUpMenuButton.Visible then
              Dec(Rect.Right,PopUpMenuButton.Width);
           end
           else
           Dec(Rect.Right, Editor.Width);
        end;
  end; // Case Of Layout
  // The first, take the Panel font as the Caption font and try to define it's metrics...
  DC := GetDC(0);
  Canvas.Handle := DC;
  Canvas.Font := Self.Font;
  DrawText(Canvas.Handle, PChar(Caption), Length(Caption), Rect, Flags);
  Canvas.Handle := 0;
  ReleaseDC(0, DC);

  Inc(Rect.Right, 3);
  Inc(Rect.Bottom, 1);

  Result := Rect;

end;

procedure TsesCustomEdit.SetPickList(const Value:TStrings);
begin
    PickList.Assign(Value);
end;

procedure TsesCustomEdit.CMTextChanged(var Message: TMessage);
begin
AdjustPositions;
Invalidate;
end;

procedure TsesCustomEdit.AdjustActualSize;
var
Caption_Height, Caption_Width:Integer;
  Rect,R:TRect;
  H,W,ActualWidth,ActualHeight:Integer;
begin
//Exit;
//If not FsesAutosize then Exit;
//If (csLoading in ComponentState)  then Exit;
If (csReading in ComponentState) then Exit;

               R:= GetCaptionRect;
               Caption_Height := R.Bottom - R.Top;
               Caption_Width := R.Right - R.Left;

               Rect.Left := Left;
               Rect.Top:=Top;

Case sesCaption.FLayout Of
clTop:
              begin
               Rect.Right := Left + Self.Width;
               If sesCaption.Visible then
               ActualHeight := Max(Self.Height, Caption_Height+Editor.Height + GetTotalBorderWidth * 2)
               else
               ActualHeight := Editor.Height + GetTotalBorderWidth * 2;
               Rect.Bottom := Top + ActualHeight;

               inherited UpdateBoundsRect(Rect);
              end;

clLeft:
   begin
   // height
   If sesCaption.Visible then
      H := Max(Editor.Height, Caption_Height)
      else
      H := Editor.Height;
   H := H + GetTotalBorderWidth * 2; // Total height
   ActualHeight := Max(Self.Height, H);
   Rect.Bottom := Self.Top + ActualHeight;

   // width
   W := Editor.Width + GetTotalBorderWidth * 2;
   If sesCaption.Visible then
      W := W + Caption_Width ;

   If PickButton.Visible then
      W := W + PickButton.Width+ sesButton.EditorGap;
   If PopUpMenuButton.Visible then
      W := W + PopUpMenuButton.Width;
   ActualWidth := Max(Self.Width, W);
   Rect.Right :=  Self.Left + ActualWidth;
               inherited UpdateBoundsRect(Rect);
   end;
end; // case sesCaption.Layout Of
Invalidate;
FResizing := False;
end;

procedure TsesCustomEdit.CMFontChanged(var Message: TMessage);
begin
  inherited;

AdjustActualSize;
AdjustPositions;

end;

procedure TsesCustomEdit.SetInitText;
begin
Case FInitReset Of
irInitial: Text := InitialValue;
irCurrent:;
irLastValid:
            begin
            If LastValidValue <> abracadabra then
               Text := LastValidValue;
            end;
end;// case FInitReset Of
FsesState := ssInit;
Repaint;
end;

function TsesCustomEdit.GetInitText:String;
begin
Result := InitialValue;
end;

procedure TsesCustomEdit.Loaded;
begin
inherited;
InitialValue := Text;
FsesState:=ssInit;
If FLogger <> nil then
   FLogger.RegisterEdit(Self);

If FPopupOptions = [] then
   PopUpMenuButton.Visible := PopUpMenuEnabled;

AdjustActualSize;
AdjustPositions;

end;

function TsesCustomEdit.PopUpMenuEnabled:Boolean;
var
I:Integer;
begin
I:= -1;
If poCalculator in FPopupOptions then I:=I+1;
If poPickList in FPopupOptions then I:=I+1;
If poUserForm in FPopupOptions then I:=I+1;
If poHistory in FPopupOptions then I:=I+1;
If poDatePicker in FPopupOptions then I:=I+1;
If poTimePicker in FPopupOptions then I:=I+1;
Result := (I > 0);
end;



procedure TsesCustomEdit.SetDataType(const Value:TClassicDataType);
begin
If not Assigned(Editor) then Exit;
If Editor.EditDataType <> Value then
   begin
   Editor.EditDataType := Value;
   end;

end;

function TsesCustomEdit.GetDataType:TClassicDataType;
begin
Result := cdtString;
If not Assigned(Editor) then Exit;
Result := Editor.EditDataType;
end;

function TsesCustomEdit.GetModified:Boolean;
begin
Result := False;
If not Assigned(Editor) then Exit;
Result := Editor.Modified;
end;


procedure TsesCustomEdit.SetStateView(const Value:TStateView);
begin
If FStateView <> Value then
   begin
   FStateView := Value;
   Invalidate;
   end;
end;

procedure TsesCustomEdit.SetsesState(const Value:TsesState);
begin
If FsesState <> Value then
   begin
   FsesState := Value;
   Invalidate;
   If not ((csLoading in ComponentState) or (csDesigning in ComponentState)) then
      begin
      If Assigned(FOnComplete) and (FsesState = ssComplete) then
         FOnComplete(Self);
      end;
   end;
end;

procedure TsesCustomEdit.SetsesEnabled(const Value:Boolean);
begin
If FEnabled <> Value then
   begin
   FEnabled := Value;
   Editor.Enabled := Value;
   PickButton.Enabled := Value;
   end;

end;

procedure TsesCustomEdit.SetModified(const Value:Boolean);
begin
If Editor.Modified <> Value then
   Editor.Modified := Value;
end;

procedure TsesCustomEdit.SetEnterAsTab(const Value:Boolean);
begin
If Editor.EnterAsTab <> Value then
   begin
   Editor.EnterAsTab := Value;
   end;

end;

function TsesCustomEdit.GetsesEnabled:Boolean;
begin
Result := FEnabled;
end;

function TsesCustomEdit.GetEnterAsTab:Boolean;
begin
Result := Editor.EnterAsTab;
end;

function TsesCustomEdit.GetsesTabStop:Boolean;
begin
//
Result := FTabStopEditor;
end;

procedure TsesCustomEdit.SetsesTabStop(const Value:Boolean);
begin

If FTabStopEditor <> Value then
   begin
   FTabStopEditor := Value;
   TabStop := False;
   Editor.TabStop := Value;
   end;
end;

function TsesCustomEdit.GetText:TCaption;
begin
Result := Editor.Text;
end;

procedure TsesCustomEdit.SetText(const Value:TCaption);
begin
If Editor.Text <> Value then
   begin
   If not ((csDesigning in ComponentState) or (csLoading in ComponentState)) then
     begin
     if Editor.HandleAllocated then
        begin
        SendMessage(Editor.Handle, EM_SETSEL, 0, -1);
        SendMessage(Editor.Handle, EM_REPLACESEL, 1, Longint(PChar(Value)));
        end
        else
        begin
        Editor.Text := Value;
        end;

     sesState:=ssEdit;
     Repaint;
     end
     else
     begin
     Editor.Modified := True;
     Editor.Text := Value;
     end;
   end;

end;
procedure TsesCustomEdit.SetVarValue(const aValue:Variant);
begin
If VarIsNull(aValue) then
   VarClear(FValue)
   else
   FValue := aValue;
   Editor.Text := VarToStr(aValue);
   Invalidate;

end;


procedure TsesCustomEdit.PostLog;
var
Old:String;
begin
Old:=LastValidValue;
If LastValidValue <> Text then
   LastValidValue := Text;

If FLoggedType <> ltNone then
   begin
   If FLogger <> nil then
      begin
      If Old <> Text then
         FLogger.Log(Self);
      end;
   end;

end;

procedure TsesCustomEdit.SetLogger(const Value:TsesLogger);
begin
If Value = nil then
   begin
   FLogger := nil;
   FLoggedType := ltNone;
   end
   else
   FLogger := Value;
end;


procedure TsesCustomEdit.SetPickDataSet(const Value:TDataSet);
begin
If Value = nil then
   begin
   FPickDataSet := nil;
   If FPickListKind = plDataSet then
      FPickListKind := plNone;
   end
   else
   FPickDataSet := Value;
end;

procedure TsesCustomEdit.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
  begin
    if (AComponent is TsesLogger) then
    begin
    FLogger := nil;
    FLoggedType := ltNone;
    end;

    if (AComponent is TDataSet) then
    begin
    FPickDataSet := nil;
    If FPickListKind = plDataSet then
       FPickListKind := plNone;
    end;

    if (AComponent is TWinControl) then
    begin
    If Assigned(FTabNext) then
       begin
       If AComponent.Name = FTabNext.Name then
          FTabNext := nil;

       end;
    end;

  end;
end;

procedure TsesCustomEdit.SetReadOnly(const Value:Boolean);
begin
If FReadOnly <> Value then
   begin
   FReadOnly := Value;
   Editor.ReadOnly := Value;
   Repaint;
   end;
end;
procedure TsesCustomEdit.SetPopupOptions(const Value:TPopupOptions);
begin
If FPopupOptions <> Value then
   begin
   FPopupOptions := Value;
   PopUpMenuButton.Visible := PopUpMenuEnabled;
   If PopUpMenuButton.Visible then
      PopUpMenuButton.Width := 12
      else
      PopUpMenuButton.Width := 0;
   // stack down: UserForm, PickList,History,Calc
   If poUserForm in FPopupOptions then
       begin
       sesButton.DefaultGlyph := dgUserForm;
       PickButtonAction := pbUserForm;
       AdjustPositions;
       Exit;
       end;

   If poPickList in FPopupOptions then
       begin
       sesButton.DefaultGlyph := dgPickList;
       PickButtonAction := pbPickList;
       AdjustPositions;
       Exit;
       end;

   If poDatePicker in FPopupOptions then
       begin
       sesButton.DefaultGlyph := dgDatePicker;
       PickButtonAction := pbDatePicker;
       AdjustPositions;
       Exit;
       end;

   If poTimePicker in FPopupOptions then
       begin
       sesButton.DefaultGlyph := dgTimePicker;
       PickButtonAction := pbTimePicker;
       AdjustPositions;
       Exit;
       end;
   If poHistory in FPopupOptions then
       begin
       sesButton.DefaultGlyph := dgHistory;
       PickButtonAction := pbHistory;
       AdjustPositions;
       Exit;
       end;
   If poCalculator in FPopupOptions then
       begin
       sesButton.DefaultGlyph := dgCalculator;
       PickButtonAction := pbCalculator;
       AdjustPositions;
       Exit;
       end;
       sesButton.DefaultGlyph := dgNone;
       PickButtonAction := pbNone;
   AdjustPositions;
   end;
end;

procedure TsesCustomEdit.PopUpMenuClick(Sender:TObject);
var
P,P1:TPoint;
begin
P.x:=PickButton.Left;
P.y:=PickButton.Top+PickButton.Height;
P1:=Self.ClientToScreen(P);
sesAdviseForm.PopUpEdit := Self;

If poCalculator in FPopupOptions then
        sesAdviseForm.Calc.Visible := True
        else
        sesAdviseForm.Calc.Visible := False;
If poPickList in FPopupOptions then
        sesAdviseForm.PickList.Visible := True
        else
        sesAdviseForm.PickList.Visible := False;

If poUserForm in FPopupOptions then
        sesAdviseForm.UserForm.Visible := True
        else
        sesAdviseForm.UserForm.Visible := False;

If poHistory in FPopupOptions then
        sesAdviseForm.History.Visible := True
        else
        sesAdviseForm.History.Visible := False;

If poDatePicker in FPopupOptions then
        sesAdviseForm.DatePicker.Visible := True
        else
        sesAdviseForm.DatePicker.Visible := False;
If poTimePicker in FPopupOptions then
        sesAdviseForm.TimePicker.Visible := True
        else
        sesAdviseForm.TimePicker.Visible := False;

sesAdviseForm.sesPopUpMenu.Popup(P1.x,P1.y);

end;

procedure TsesCustomEdit.PickButtonClick(Sender:TObject);
var
Res:TModalResult;
F:TCustomForm;
begin
Res:=0;
Case PickButtonAction Of
pbUserForm:
           begin
           If Assigned(FOnPickButtonClick) then
              begin
              FOnPickButtonClick(Self);
              Exit;
              end;

           end;
pbHistory:
           begin
           sesAdviseForm.ShowHistory(Self);
           Res := sesAdviseForm.ShowModal;
           end;

pbCalculator:
           begin
           sesAdviseForm.ShowCalc(Self);
           Res := sesAdviseForm.ShowModal;
           end;
pbPickList:
           begin
           sesAdviseForm.ShowPickList(Self);
           Res := sesAdviseForm.ShowModal;
           end;
pbDatePicker:
           begin
           sesAdviseForm.ShowDatePicker(Self);
           Res := sesAdviseForm.ShowModal;
           end;
pbTimePicker:
           begin
           sesAdviseForm.ShowTimePicker(Self);
           Res := sesAdviseForm.ShowModal;
           end;
end; // Case Of PickButtonAction

Case Res Of
mrCancel,mrRetry:;
mrAbort:
        begin
        F:=GetParentForm(Self);
        LoseFocusAllowed := True;
//        F.SelectNext(Self,True,True);
        F.Perform(WM_NEXTDLGCTL, 0,0);
//        Self.Editor.Handle,99999999);
        end;
end; // Case Of Res

end;

procedure TsesCustomEdit.SetFocus;
begin
If Editor.CanFocus then
   Editor.SetFocus;
end;
procedure TsesCustomEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  Case DataType Of
  cdtBoolean:
            begin
            Case Key Of
             VK_Down:
                     begin
                     If Trim(Text) = sesData.RangeMaxOrTrue then
                        Text := sesData.RangeMinOrFalse
                        else
                        Text := sesData.RangeMaxOrTrue;
                     end;

             VK_Up:
                begin
                Key:=0;
                PopUpMenuClick(Self);
                end;
              end; // Case Of Key

            end;
   else
   begin
   Case Key Of
     VK_Down:
                begin
                Key := 0;
                PickButtonClick(Self);
                end;
     VK_Up:
                begin
                Key:=0;
                PopUpMenuClick(Self);
                end;

   end; // Case Of Key

   end; // all other types
   end; // Case Of Datatype
end;

//takes the focus from the focused TSesEdit regardless its data validity and
// passes it to the aWinControl
procedure TsesCustomEdit.PassFocusTo(aWinControl:TWinControl);
begin
If aWinControl = nil then Exit;
If not aWinControl.CanFocus then Exit;

LoseFocusAllowed := True;
aWinControl.SetFocus;
LoseFocusAllowed := False;

end;


function TsesCustomEdit.AsCardinal:Cardinal;
begin
Result := StrToInt(Text);
end;

function TsesCustomEdit.AsInteger:Integer;
begin
Result := StrToInt(Text);
end;

function TsesCustomEdit.AsRound(const Precision:Integer):Double;
begin
Result := RoundExtended(StrToFloat(Text),Precision);
end;

function TsesCustomEdit.AsFloat:Double;
begin
Result := StrToFloat(Text);
end;

function TsesCustomEdit.AsDate:TDate;
begin
Result := StrToDate(Text);
end;

function TsesCustomEdit.AsDateNullTime:TDateTime;
begin
Result := StrToDate(Text);
end;

function TsesCustomEdit.AsTime:TTime;
begin
Result := StrToTime(Text);
end;

function TsesCustomEdit.AsTimeNullDate:TDateTime;
begin
Result := StrToDateTime(Text);
end;

function TsesCustomEdit.GetAsVariant:Variant;
begin
  If (Text <> '') then Result := Text else Result := Null;
end;

function TsesCustomEdit.CheckText:TEditErrorType;
begin
Result := Self.Editor.CheckText;
end;

function TsesCustomEdit.AsBoolean:Boolean;
begin
If sesData.RangeMaxOrTrue = Trim(Text) then
   Result := True
   else
   Result := False;
end;

destructor TsesCustomEdit.Destroy;
begin
sesEditor.Free;
sesFrame.Free;
sesButton.Free;
FPickList.Free;
inherited Destroy;
end;
{***End of implementation of TsesCustomEdit}

{TsesCustomEditor implementation}

constructor TsesCustomEditor.Create(AOwner:TComponent);
begin
inherited Create(AOwner);


FSeparatorPressed := False;

Autosize := True;
Width := 121;

FFocusPassedOn:= False;
//   
FClassicDataType := cdtString;
FEnterAsTab := True;
FButtonHotKey:=TextToShortCut('F12');

FCanLoseFocus := False;

TabStop:=False;

Self.OnChange := DoChangeText;

ControlStyle := ControlStyle - [csSetCaption];

end;

procedure TsesCustomEditor.SetClassicDataType(Value:TClassicDataType);
begin
If Value <> FClassicDataType then
   begin
   FClassicDataType := Value;
   {
   Case FClassicDataType Of
   ftString:;
   ftBoolean: If not IsStringBoolean then
              begin
              If (FRangeMinOrFalse = '') and (FRangeMaxOrTrue = '') then
                 begin
                 FRangeMinOrFalse := sesFalse;
                 FRangeMaxOrTrue := sesTrue;
                 end;
              Text := FRangeMinOrFalse;
              end;
   ftCardinal: If not (IsStringCardinal(Text) = 0) then Text := '0';
   ftInteger: If not (IsStringInteger(Text) = 0) then Text := '0';
   ftFloat: If not (IsStringFloat(Text) = 0) then Text := '0';
   end; // Case Of FClassicDataType
   }
   end;
end;


{
procedure TsesCustomEditor.SetEditRect;
var
  Loc: TRect;
begin
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc));
  If FButtonVisible then
  begin
  Loc.Bottom := ClientHeight + 1;
  Loc.Right := ClientWidth - FButton.Width - 4;
  Loc.Top := 0;
  Loc.Left := 0;
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
  FButton.Top := 0;
  FButton.Height := ClientHeight + 1;
  FButton.Left := ClientWidth - FButton.Width;
  end
  else
  begin
  Loc.Bottom := ClientHeight + 1;
  Loc.Right := ClientWidth;
  Loc.Top := 0;
  Loc.Left := 0;
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
  end;

end;
}
{
procedure TsesCustomEditor.SetButtonVisible(Value:Boolean);
begin
If Value <> FButtonVisible then
   begin
   FButtonVisible:=Value;
   If FButtonVisible then
      begin
  FButton := TSpeedButton.Create (Self);
  FButton.Name:='Button';
  FButton.Width := 20;
  FButton.Visible := False;
  FButton.Cursor:=crArrow;
  FButton.Parent := Self;
//  FButton.Glyph := nil;
//  FButton.Glyph.Handle := LoadBitmap(HInstance, 'EMPTY');
//  FButtonKind:=ebkCustom;
//  FButton.NumGlyphs := 1;
  FButton.Spacing :=0;
  FButton.Margin:=-1;
  FButton.OnClick := OnInternalButtonClick;
      end
      else
      begin
      FButton.Free;
      end;
   SetEditRect;
   Invalidate;
   end;
end;
}
procedure TsesCustomEditor.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
{  Params.Style := Params.Style and not WS_BORDER;  }
  Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN;
end;

procedure TsesCustomEditor.CreateWnd;
begin
  inherited CreateWnd;
end;


procedure TsesCustomEditor.WMSize(var Message: TWMSize);
begin
  inherited;
end;

{
procedure TsesCustomEditor.SetEditButtonKind(Value:TEditButtonKind);
begin
If Value <> FButtonKind then
   begin
   FButtonKind := Value;
// (ebkNone,ebkCustom,ebkToggle,ebkDropDown,ebkForm,ebkOpen);
   Case FButtonKind Of
   ebkNone: FButton.Glyph.Handle := LoadBitmap(HInstance, 'EMPTY');
   ebkCustom:;
   ebkToggle: FButton.Glyph.Handle := LoadBitmap(HInstance, 'TOGGLE');
   ebkDropDown: FButton.Glyph.Handle := LoadBitmap(HInstance, 'DOWN');
   ebkForm: FButton.Glyph.Handle := LoadBitmap(HInstance, 'FORM');
   ebkOpen: FButton.Glyph.Handle := LoadBitmap(HInstance, 'OPEN');

        end; // Case Of FButtonKind
   Invalidate;
   end;
end;
}

{
procedure TsesCustomEditor.SetEditEnabled(Value:Boolean);
begin
Enabled := Value;
     If FButton <> nil then
        begin
        FButton.Enabled := Value;
        end;
Invalidate;
end;
}
{
function TsesCustomEditor.GetButtonFlat:Boolean;
begin
Result := FButton.Flat;
end;
}
{
procedure TsesCustomEditor.SetButtonFlat(Value:Boolean);
begin
     If FButton <> nil then
        begin
        FButton.Flat := Value;
        end;
Invalidate;
end;


function TsesCustomEditor.GetButtonCaption:String;
begin
     If FButton <> nil then
        begin
        Result := FButton.Caption;
        end;
end;

procedure TsesCustomEditor.SetButtonCaption(Value:String);
begin
     If FButton <> nil then
        begin
        FButton.Caption := Value;
        Invalidate;
        end;
end;

function TsesCustomEditor.GetButtonWidth:Integer;
begin
     If FButton <> nil then
        begin
        Result := FButton.Width;
        end;
end;

procedure TsesCustomEditor.SetButtonWidth(Value:Integer);
begin

FButton.Width := Value;
SetEditRect;
Invalidate;
end;

function TsesCustomEditor.GetButtonLayout:TButtonLayout;
begin
     If FButton <> nil then
        begin
        Result := FButton.Layout;
        end;
end;

procedure TsesCustomEditor.SetButtonLayout(Value:TButtonLayout);
begin

FButton.Layout := Value;
Invalidate;
end;



function TsesCustomEditor.GetButtonGlyph:TBitMap;
begin
  Result := FButton.Glyph;
end;

procedure TsesCustomEditor.SetButtonGlyph(Value:TBitMap);
begin
  if Value <> nil then
    FButton.Glyph := Value
  else
  begin
    FButton.Glyph.Handle := LoadBitmap(HInstance, 'EMPTY');
    FButton.NumGlyphs := 1;
    FButton.Invalidate;
  end;
end;



function TsesCustomEditor.GetButtonNumGlyphs: TNumGlyphs;
begin
  Result := FButton.NumGlyphs;
end;

procedure TsesCustomEditor.SetButtonNumGlyphs(Value: TNumGlyphs);
begin
  FButton.NumGlyphs := Value;
end;
}
procedure TsesCustomEditor.KeyUp(var Key: Word; Shift: TShiftState);
begin
If Assigned(TsesCustomEdit(Self.Parent).OnKeyUp) then
   begin
   TsesCustomEdit(Self.Parent).OnKeyUp(TsesCustomEdit(Self.Parent),Key,Shift);
   If Key = 0 then Exit;
   end;
inherited KeyUp(Key, Shift);
end;
procedure TsesCustomEditor.KeyDown(var Key: Word; Shift: TShiftState);
var
F:TCustomForm;
aKey:Word;
aShift:TShiftState;
Edit:TsesCustomEdit;
begin
// VK_DECIMAL(110) is the case, when user pressed down Decimal key on the numeric keypad
If (TsesCustomEdit(Self.Parent).FDecimalKeyAsSeparator) and (Key = 110) then
                                begin
                                FSeparatorPressed := True;
                                end;


If Assigned(TsesCustomEdit(Self.Parent).OnKeyDown) then
   begin
   TsesCustomEdit(Self.Parent).OnKeyDown(TsesCustomEdit(Self.Parent),Key,Shift);
   If Key = 0 then Exit;
   end;


ShortCutToKey(FButtonHotKey,aKey,aShift);
If (aKey = Key) and (aShift = Shift) then
   begin
   Key:=0;
   end;
Case Key Of
VK_UP,VK_DOWN:
              begin
              //
              //OnInternalButtonClick(Self);
              Edit:=TsesCustomEdit(Parent);
              Edit.KeyDown(Key,Shift);
              Key:=0;
              end;
VK_RETURN:
              begin
              If FEnterAsTab then
                begin
                Key:=0;
                F:=GetParentForm(Self);
                F.Perform(WM_NEXTDLGCTL, 0, 0);
                end; // EnterAsTab
            end; // Enter
else
inherited KeyDown(Key, Shift);
end; // Case Of Key
end;

procedure TsesCustomEditor.KeyPress(var Key: Char);
begin
// VK_DECIMAL(110) is the case, when user pressed down Decimal key on the numeric keypad
If FSeparatorPressed then
                                begin
                                FSeparatorPressed := False;
// try to translate Decimal Separator(numeric keypad) to universal Separator
Case FClassicDataType Of
cdtFloat: Key := DecimalSeparator;
cdtDate:  Key := DateSeparator;
cdtTime:  Key := TimeSeparator;
end;// Case FClassicDataType Of
                                end;

If Assigned(TsesCustomEdit(Self.Parent).OnKeyPress) then
   begin
   TsesCustomEdit(Self.Parent).OnKeyPress(TsesCustomEdit(Self.Parent),Key);
   If Key = #0 then Exit;

   end;

If (Key = #27) then
    begin
    If TsesCustomEdit(Self.Parent).Cancel then
       begin
       GetParentForm(Self).Close;
       Key := #0;
       end;
    Exit;
    end
    else
If (Key = #13) Or (Key = #9) then
    begin
    Key := #0;
    Exit;
    end
    else
    begin
    If TsesCustomEdit(Self.Parent).ReadOnly then
       begin
       Key:=#0;
       Exit;
       end;
    TsesCustomEdit(Self.Parent).sesState := ssEdit;
    TsesCustomEdit(Self.Parent).Invalidate;

  if not IsValidChar(Key) then
          begin
          ActivateCharCheckPopUpForm(Key);
          Key := #0;
          end
          else
          begin
          TsesCustomEdit(Self.Parent).sesState := ssEdit;
          TsesCustomEdit(Self.Parent).Invalidate;
          end;
  if Key <> #0 then inherited KeyPress(Key);
    end;

end;

procedure TsesCustomEditor.DoChangeText(Sender:TObject);
begin
If Assigned(TsesCustomEdit(Self.Parent).OnChange) then
    TsesCustomEdit(Self.Parent).OnChange(TsesCustomEdit(Self.Parent));
end;

function TsesCustomEditor.IsValidChar(Key: Char): Boolean;
begin
Result := False;
Case FClassicDataType Of
cdtString: Result:=True;
cdtDate:  Result := (Key in DateSymbolSet);
cdtBoolean: Result:=True;
cdtCardinal: Result := (Key in CardinalSymbolSet);
cdtInteger: Result := (Key in IntegerSymbolSet);
cdtFloat: Result := (Key in FloatSymbolSet);
cdtTime:  Result := (Key in TimeSymbolSet);
end; // Case Of FClassicDataType


  Result := Result or (Key < #32);
  if not Enabled and Result and ((Key >= #32) or
      (Key = Char(VK_BACK)) or (Key = Char(VK_DELETE))) then
    Result := False;
end;

function TsesCustomEditor.IsStringBoolean:Boolean;
begin
If (Text=FRangeMinOrFalse) or (Text=FRangeMaxOrTrue) then
   Result := True
   else
   Result:=False;
end;

procedure TsesCustomEditor.CMEnter(var Message: TCMGotFocus);
begin
If AutoSelect then
SelectAll;
inherited;
end;

// check up the Text Value for format errors...
function TsesCustomEditor.CheckText:TEditErrorType;
var
Check:Boolean;
Edit:TsesCustomEdit;
ParsedDate:TParsedDate;
ParsedTime:TParsedTime;
begin
Edit:=TsesCustomEdit(Self.Parent);

//1. Control level 1 "Required Control" tests the Text value is empty
Result:=ceOK;
If Edit.Required then
   begin
   If Length(Text) = 0 then
      begin
      Result := ceRequired;
      end;
   end;
If not (Result = ceOK) then
       begin
       Edit.sesState:=ssError;
       Edit.Invalidate;
       Exit;
       end;


//2. Control level 2 "Format Control" - tests the Text  depending on Data Type
Case FClassicDataType Of
cdtString: Result := ceOK;
cdtDate:
           begin
           Check := IsStringDate(Self.Text,ParsedDate) = 0;
           If Check then
              Result := ceOK
              else
              Result := ceDate;
           end;
cdtTime:
           begin
           Check := IsStringTime(Self.Text,ParsedTime) = 0;
           If Check then
              Result := ceOK
              else
              Result := ceTime;
           end;

cdtCardinal:
           begin
           Check := IsStringCardinal(Self.Text) = 0;
           If Check then
              Result := ceOK
              else
              Result := ceNumber;
           end;
cdtInteger:
           begin
           Check := IsStringInteger(Self.Text) = 0;
           If Check then
              Result := ceOK
              else
              Result := ceNumber;
           end;

cdtBoolean:
           begin
           Check := IsStringBoolean;
           If Check then
              Result := ceOK
              else
              Result := ceBoolean;
           end;
cdtFloat:
           begin
           Check := IsStringFloat(Self.Text) = 0;
           If Check then
              Result := ceOK
              else
              Result := ceNumber;
           end;

end; // Case Of FClassicDataType
If not (Result = ceOK) then
       begin
       Edit.sesState:=ssError;
       Edit.Invalidate;
       Exit;
       end;

//3. Control level 3 "Range Control" - tests the Text  depending on gived Range
If FRanged then
      begin
Case FClassicDataType Of
cdtString: Result := ceOK;
cdtCardinal,cdtInteger:
    If IsIntegerInRange(Self.Text, FRangeMinOrFalse, FRangeMaxOrTrue) then
       Result := ceOK
       else
       Result := ceRange;
cdtFloat:
    If IsFloatInRange(Self.Text, FRangeMinOrFalse, FRangeMaxOrTrue) then
       Result := ceOK
       else
       Result := ceRange;

end; // Case Of FClassicDataType
      end;
If not (Result = ceOK) then
       begin
       Edit.sesState:=ssError;
       Edit.Invalidate;
       Exit;
       end;

//4. Control level 4 "Logical Control" - tests the Text  depending on "to belong to PickList"
If Edit.PickListKind <> plNone then
   begin
   If sesAdviseForm.IsValueFounded(Edit) > -1 then
      Result := ceOK
      else
      Result := cePickList;
   end
   else
   Result := ceOK;

If not (Result = ceOK) then
       begin
       Edit.sesState:=ssError;
       Edit.Invalidate;
       Exit;
       end;

//5. The last level 5 "User Control" by OnCheckText event handler
   If Edit.DoCheckText then
      Result := ceOK
      else
      Result := ceUser;

If not (Result = ceOK) then
       begin
       Edit.sesState:=ssError;
       Edit.Invalidate;
       Exit;
       end;

// logs the right result into a history jornal, if logger is presented
   Edit.sesState:=ssComplete;
   Edit.PostLog;
   Edit.Invalidate;


end;


procedure TsesCustomEditor.UserExitAction;
begin
  TsesCustomEdit(Self).DoExit;
end;

procedure TsesCustomEditor.CMExit(var Message: TCMExit);
var
EditCompleted:Boolean;
begin
If LoseFocusAllowed then
   begin
   LoseFocusAllowed:=False;
   UserExitAction;
   Exit;
   end;

If (TsesCustomEdit(Self.Parent).FErrorExit = eeIfInit) and
   (TsesCustomEdit(Self.Parent).sesState = ssInit) then
      begin
      UserExitAction;
      inherited DoExit;
      Exit;
      end;


// Has edit completed with the right result?
CheckTextResult := CheckText;
EditCompleted := CheckTextResult = ceOK;
If EditCompleted then
   TsesCustomEdit(Self.Parent).sesState := ssComplete
   else
   TsesCustomEdit(Self.Parent).sesState := ssError;
TsesCustomEdit(Self.Parent).Invalidate;
// can Edit lose focus?
If FCanLoseFocus or EditCompleted then
   begin
   // yes, it can. Then triiger this flag to False for the future...
   FCanLoseFocus := False;
   // a OnExit procedure will be executed, if assigned...
   UserExitAction;
   Exit;
   end
   else
   begin
   If not EditCompleted then ActivateErrorWindow;
   If FCanLoseFocus then
      begin
      FCanLoseFocus := False;
      // a OnExit procedure will be executed, if assigned...
      UserExitAction;
      Exit;
      end
      else
      SetFocus;
   end;


end;



procedure TsesCustomEditor.ActivateErrorWindow;
var
Key:Char;
Edit:TsesCustomEdit;
Res:TModalResult;
begin
Res:=0;

TsesCustomEdit(Self.Parent).sesSayMessage(Self.FCheckTextResult);
If TsesCustomEdit(Self.Parent).ErrorDialog = edShow then
                     begin

Key:='9';
If sesAdviseForm = nil then
   begin
   sesAdviseForm:=TsesAdviseForm.Create(Application);
   end;
Edit:=TsesCustomEdit(Self.Parent);
Case Self.FCheckTextResult Of
ceBoolean,ceNumber,ceRange:
                            begin
                            sesAdviseForm.ShowEditError(Self, Key, CheckTextResult);
                            Case sesAdviseForm.ModalResult Of
                                  mrRetry,mrCancel: FCanLoseFocus:=False;
                                  mrAbort: FCanLoseFocus:=True;
                            end; // Case CEAdviseForm.ModalResult Of
                            Exit;
                            end;
cePickList:
           begin
           sesAdviseForm.ShowPickList(Edit);
           Res := sesAdviseForm.ShowModal;
           end;
ceDate:
           begin
           sesAdviseForm.ShowDatePicker(Edit);
           Res := sesAdviseForm.ShowModal;
           end;
ceTime:
           begin
           sesAdviseForm.ShowTimePicker(Edit);
           Res := sesAdviseForm.ShowModal;
           end;

ceUser:
           begin
           Res := Edit.DoUserErrorProcess;
           end;

ceRequired:
           begin
           sesAdviseForm.ShowErrorMessage(Edit,ce_Required_Caption,
           ce_Required_Text + Edit.Caption);
           Res := sesAdviseForm.ShowModal;
           end;

end; // Case Of FCheckTextResult
     Case Res Of
     mrRetry,mrCancel: FCanLoseFocus:=False;
     mrAbort: FCanLoseFocus:=True;
     end; // Case CEAdviseForm.ModalResult Of


                     end;
end;


procedure TsesCustomEditor.CEOperation(var Message: TWMKey);
begin
Case Message.CharCode Of
VK_IsEditCompleted:
  begin
  CheckTextResult := CheckText;
  If not (CheckTextResult = ceOK) then
     begin
     Message.Result := 1; // break broadcasting chain
     LastEdit:=TsesCustomEdit(Self.Parent); // pointer to last operational Edit
     end;

  end;
VK_SetFocusIfNotEditCompleted:
  begin
  If not (CheckText = ceOK) then
     begin
     If Focused then
                begin
                Message.Result := 1;
                LastEdit := TsesCustomEdit(Self.Parent); // pointer to last operational Edit
                end
                else
                begin
                If CanFocus then
                   begin
                   Message.Result := 1;
                   SetFocus;
                   LastEdit:=TsesCustomEdit(Self.Parent); // pointer to last operational Edit
                   end;
                end;
     end;
  end;
end; // Case Of Message.Unused
inherited;

end;


procedure TsesCustomEditor.ActivateCharCheckPopUpForm(Key:Char);
begin
TsesCustomEdit(Self.Parent).sesSayMessage(ceChar);
If TsesCustomEdit(Self.Parent).ErrorDialog = edShow then
                     begin
If sesAdviseForm = nil then
   begin
   sesAdviseForm := TsesAdviseForm.Create(Application);
   end;
  sesAdviseForm.ShowEditError(Self,Key,ceChar);
  Case sesAdviseForm.ModalResult Of
  mrRetry,mrCancel: FCanLoseFocus:=False;
  mrAbort: FCanLoseFocus:=True;
  end; // Case CEAdviseForm.ModalResult Of

end;                 end;


procedure TsesCustomEditor.DeActivateErrorWindow;
begin
end;

{***TsesCustomEditor implementation}

procedure InitLocals; far;
var
  I: Integer;
begin
  for I := Low(sesButtonGlyphs) to High(sesButtonGlyphs) do
    If sesButtonGlyphs[I] <> nil then
       sesButtonGlyphs[I].Free;
sesAdviseForm := TsesAdviseForm.Create(Application);
FastList1:=TWorkList.Create;
FastList2:=TWorkList.Create;
end;

procedure DestroyLocals; far;
var
  I: Integer;
begin
  for I := Low(sesButtonGlyphs) to High(sesButtonGlyphs) do
       sesButtonGlyphs[I] := nil;
sesAdviseForm.Free;
FastList1.Free;
FastList2.Free;
end;

initialization
InitLocals;
finalization
  DestroyLocals;


end.
