unit sesRegisterUnit;

interface
uses
  sesSystem,sesEditUnit,sesDataSetUnit,FieldDescDialogUnit, DataStringsDialogUnit,
  DsgnIntf,TypInfo,
  Windows, Messages, SysUtils, Classes, Graphics, stdctrls,Controls,
  Buttons,extctrls, Forms, Dialogs;

type

// edits FileName property by OpenDialog
TComponentFileNameProperty = class(TPropertyEditor)
  public
  function GetAttributes: TPropertyAttributes;override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  procedure Edit;override;
  end;

// sesFields property editor: sesDataset structure
TComponentsesFieldsProperty = class(TPropertyEditor)
  public
  function GetAttributes: TPropertyAttributes;override;
  procedure Edit;override;
  end;

// DataStrings property editor: sesDataset memory data
TComponentDataStringsProperty = class(TPropertyEditor)
  public
  function GetAttributes: TPropertyAttributes;override;
  procedure Edit;override;
  end;

procedure Register;

implementation
{$R *.DCR}

{Property editors' types}

{TComponentFileNameProperty}
function TComponentFileNameProperty.GetAttributes: TPropertyAttributes;
begin
 Result := [paDialog];
end;

procedure TComponentFileNameProperty.Edit;
var
  OpenDialog: TOpenDialog;
begin
  OpenDialog:= TOpenDialog.Create(Application);                { construct the editor }
  try
    OpenDialog.Filter :=
    'Text file(*.txt)|*.txt|Any file(*.*)|*.*';
    OpenDialog.FileName := GetValue;                           { use the existing value }
    if OpenDialog.Execute then                          { if the user OKs the dialog... }
      SetValue(OpenDialog.FileName);                   { ...use the result to set value }
  finally
    OpenDialog.Free;                                               { destroy the editor }

  end;
end;

function TComponentFileNameProperty.GetValue: string;
begin
Result := GetStrValue;
end;

procedure TComponentFileNameProperty.SetValue(const Value:string);
begin
SetStrValue(Value);
end;

{TComponentsesFieldsProperty}
function TComponentsesFieldsProperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paDialog] - [paSubProperties];
end;

procedure TComponentsesFieldsProperty.Edit;
var
Comp:TsesDataSet;
begin
  FieldDescDialog:= TFieldDescDialog.Create(Application);
  Comp := TsesDataSet(GetComponent(0));
  FieldDescDialog.sesDataSet := Comp;
  FieldDescDialog.Caption := 'Restructure -> '+Comp.Name;
  try

  FieldDescDialog.Memo.Lines := TStrings(GetOrdValue);
    case FieldDescDialog.ShowModal of
      mrOk: SetOrdValue(Longint(FieldDescDialog.Memo.Lines));
    end; // case

  finally
    FieldDescDialog.Free;

  end;
end;

{TComponentDataStringsProperty}
function TComponentDataStringsProperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paDialog] - [paSubProperties];
end;

procedure TComponentDataStringsProperty.Edit;
var
Comp:TsesDataSet;
begin
  DataStringsDialog:= TDataStringsDialog.Create(Application);
  Comp := TsesDataSet(GetComponent(0));
  DataStringsDialog.sesDataSet := Comp;
  DataStringsDialog.Structure :=
  DataStringsDialog.sesDataSet.StructureToString;
  If  DataStringsDialog.Structure = '' then
      DataStringsDialog.Structure :=sesFieldDefSeparator+sesFieldDefSeparator+sesFieldDefSeparator +
      'Line;1;Line of text;80;240';
  DataStringsDialog.Caption := 'DataStrings -> '+Comp.Name;
  try

  DataStringsDialog.Memo.Lines := TStrings(GetOrdValue);
  If not DataStringsDialog.MemoToDataSet then Exit;
    case DataStringsDialog.ShowModal of
      mrOk: SetOrdValue(Longint(DataStringsDialog.Memo.Lines));
    end; // case

  finally
    DataStringsDialog.Free;

  end;
end;

procedure Register;
begin
{Register components}
  RegisterComponents('sesLib', [TsesDataSet,TsesLogger,TsesEdit]);

{Register Property editors}
  RegisterPropertyEditor(TypeInfo(TsesFileName), TComponent, 'FileName',
      TComponentFileNameProperty);
  RegisterPropertyEditor(TypeInfo(TStrings), TComponent, 'sesFields',
      TComponentsesFieldsProperty);
  RegisterPropertyEditor(TypeInfo(TStrings), TComponent, 'DataStrings',
      TComponentDataStringsProperty);

end;

end.
